# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.6 `fit` Chapter.

``fit`` - Curve and Surface Fitting

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

The main aim of this module is to assist you in finding a function which
approximates a set of data points.
Typically the data contain random errors, as of experimental
measurement, which need to be smoothed out.
To seek an approximation to the data, it is first necessary to specify
for the approximating function a mathematical form (a polynomial, for
example) which contains a number of unspecified coefficients: the
appropriate fitting function then derives for the coefficients the
values which provide the best fit of that particular form.
The module deals mainly with curve and surface fitting (i.e., fitting
with functions of one and of two variables) when a polynomial or a cubic
spline is used as the fitting function, since these cover the most
common needs.
However, fitting with other functions and/or more variables can be
undertaken by means of general linear or nonlinear functions (some of
which are contained in other modules) depending on whether the
coefficients in the function occur linearly or nonlinearly.
Cases where a graph rather than a set of data points is given can be
treated simply by first reading a suitable set of points from the graph.

The module also contains functions for evaluating, differentiating and
integrating polynomial and spline curves and surfaces, once the
numerical values of their coefficients have been determined.

There is also a function for computing a Padé approximant of a
mathematical function (see [Background to the Problems]).

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02intro.html
"""

# NAG Copyright 2017-2022.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import _nag_e_consts
from ..base import utils

def dim1_cheb_arb(m, kplus1, x, y, w, a, s):
    u"""
    Least squares curve fit, by polynomials, arbitrary data points.

    ``dim1_cheb_arb`` computes weighted least squares polynomial
    approximations to an arbitrary set of data points.

    For full information please refer to the NAG Library document for
    e02ad

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02adf.html

    Parameters
    ----------
    m : int
        The number m of data points.

    kplus1 : int
        k+1, where k is the maximum degree required.

    x : float, shape (`m`)
        The values x_r of the independent variable, for r = 1,2,...,m.

    y : float, shape (`m`)
        The values y_r of the dependent variable, for r = 1,2,...,m.

    w : float, shape (`m`)
        The set of weights, w_r, for r = 1,2,...,m. For advice on the
        choice of weights, see [the E02 Introduction].

    a : float, shape (`kplus1`, `kplus1`), modified in place
        `On exit`: the coefficients of T_j(x_) in the approximating
        polynomial of degree i. `a`[i,j] contains the coefficient
        a_{i+1,j+1}, for j = 0,1,...,i, for i = 0,1,...,k.

    s : float, shape (`kplus1`), modified in place
        `On exit`: `s`[i] contains the root-mean-square residual s_i,
        for i = 0,1,...,k, as described in [Notes]. For the
        interpretation of the values of the s_i and their use in
        selecting an appropriate degree, see [the E02 Introduction].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, i = *<value>* and `w`[i-1] = *<value>*.

            Constraint: `w`[i-1] > 0.0.

        (`errno` 2)
            On entry, i = *<value>*, `x`[i-1] = *<value>* and `x`[i-2] =
            *<value>*.

            Constraint: `x`[i-1] >= `x`[i-2].

        (`errno` 3)
            On entry, all `x`[I-1] have the same value: `x`[0] =
            *<value>*.

        (`errno` 4)
            On entry, `kplus1` = *<value>* and mdist = *<value>*.

            Constraint: `kplus1` <= mdist, where mdist is the number of
            distinct `x` values.

        (`errno` 4)
            On entry, `kplus1` = *<value>* and `m` = *<value>*.

            Constraint: `kplus1` <= `m`.

        (`errno` 4)
            On entry, `kplus1` = *<value>*.

            Constraint: `kplus1` >= 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_cheb_arb`` determines least squares polynomial approximations
    of degrees 0,1,...,k to the set of data points (x_r,y_r) with
    weights w_r, for r = 1,2,...,m.

    The approximation of degree i has the property that it minimizes
    sigma_i the sum of squares of the weighted residuals epsilon_r,
    where

        epsilon_r = w_r(y_r-f_r)

    and f_r is the value of the polynomial of degree i at the rth data
    point.

    Each polynomial is represented in Chebyshev series form with
    normalized argument x_.
    This argument lies in the range -1 to +1 and is related to the
    original variable x by the linear transformation

        x_= ((2x-x_max-x_min))/((x_max-x_min)).

    Here x_max and x_min are respectively the largest and smallest
    values of x_r.
    The polynomial approximation of degree i is represented as

1/2a_{i+1,1}T_0(x_)+a_{i+1,2}T_1(x_)+a_{i+1,3}T_2(x_)+...+a_{i+1,i+1}T_i(x_),

    where T_j(x_), for j = 0,1,...,i, are the Chebyshev polynomials of
    the first kind of degree j with argument (x_).

    For i = 0,1,...,k, the function produces the values of a_{i+1,j+1},
    for j = 0,1,...,i, together with the value of the root-mean-square
    residual s_i = sqrt(sigma_i/(m-i-1)).
    In the case m = i+1 the function sets the value of s_i to zero.

    The method employed is due to Forsythe (1957) and is based on the
    generation of a set of polynomials orthogonal with respect to
    summation over the normalized dataset.
    The extensions due to Clenshaw (1960) to represent these polynomials
    as well as the approximating polynomials in their Chebyshev series
    forms are incorporated.
    The modifications suggested by Reinsch and Gentleman (see Gentleman
    (1969)) to the method originally employed by Clenshaw for evaluating
    the orthogonal polynomials from their Chebyshev series
    representations are used to give greater numerical stability.

    For further details of the algorithm and its use see Cox (1974) and
    Cox and Hayes (1973).

    Subsequent evaluation of the Chebyshev series representations of the
    polynomial approximations should be carried out using
    ``dim1_cheb_eval``.

    References
    ----------
    Clenshaw, C W, 1960, `Curve fitting with a digital computer`,
    Comput. J. (2), 170--173

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Forsythe, G E, 1957, `Generation and use of orthogonal polynomials
    for data fitting with a digital computer`, J. Soc. Indust. Appl.
    Math. (5), 74--88

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_arb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, kplus1.
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineFloat64ArrayType.c_type, # Output, a[lda, kplus1].
        utils._EngineFloat64ArrayType.c_type, # Output, s[kplus1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _kplus1 = utils._EngineIntScalarType(
        kplus1, {'fun_name': fun_name, 'entity': 'kplus1'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _kplus1_py = utils._EngineIntScalarType.to_py(_kplus1)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_kplus1_py, _kplus1_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_kplus1_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _kplus1,
        _lda,
        _x,
        _y,
        _w,
        _work1,
        _work2,
        _a,
        _s,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and w[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w[i-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, x[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[i-1] >= x[i-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all x[I-1] have the same value:',
                    'x[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kplus1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'mdist = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kplus1 <= mdist, where mdist is the number of distinct x values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kplus1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kplus1 <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kplus1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kplus1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _s_cls.refresh_py(_s)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_cheb_eval(nplus1, a, xcap):
    u"""
    Evaluation of fitted polynomial in one variable from Chebyshev
    series form (simplified parameter list).

    ``dim1_cheb_eval`` evaluates a polynomial from its Chebyshev series
    representation.

    For full information please refer to the NAG Library document for
    e02ae

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02aef.html

    Parameters
    ----------
    nplus1 : int
        The number n+1 of terms in the series (i.e., one greater than
        the degree of the polynomial).

    a : float, shape (`nplus1`)
        `a`[i-1] must be set to the value of the ith coefficient in the
        series, for i = 1,2,...,n+1.

    xcap : float
        x_, the argument at which the polynomial is to be evaluated. It
        should lie in the range -1 to +1, but a value just outside this
        range is permitted (see [Raises/Warns]) to allow for possible
        rounding errors committed in the transformation from x to x_
        discussed in [Notes]. Provided the recommended form of the
        transformation is used, a successful exit is thus assured
        whenever the value of x lies in the range x_min to x_max.

    Returns
    -------
    p : float
        The value of the polynomial.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xcap` = *<value>* and EPS = *<value>*.

            Constraint: abs(`xcap`) <= 1+4*EPS, where EPS is machine
            precision.

        (`errno` 2)
            On entry, `nplus1` = *<value>*.

            Constraint: `nplus1` >= 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_cheb_eval`` evaluates the polynomial

        1/2a_1 T_0(x_)+a_2 T_1(x_)+a_3 T_2(x_)+...+a_{n+1}T_n(x_)

    for any value of x_ satisfying -1 <= x_<= 1.
    Here T_j(x_) denotes the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    The value of n is prescribed by you.

    In practice, the variable x_ will usually have been obtained from an
    original variable x, where x_min <= x <= x_max and

        x_= ((x-x_min)-(x_max-x))/((x_max-x_min))

    Note that this form of the transformation should be used
    computationally rather than the mathematical equivalent

        x_= ((2x-x_min-x_max))/((x_max-x_min))

    since the former guarantees that the computed value of x_ differs
    from its true value by at most 4 epsilon, where epsilon is the
    machine precision, whereas the latter has no such guarantee.

    The method employed is based on the three-term recurrence relation
    due to Clenshaw (1955), with modifications to give greater numerical
    stability due to Reinsch and Gentleman (see Gentleman (1969)).

    For further details of the algorithm and its use see Cox (1974) and
    Cox and Hayes (1973).

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nplus1.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nplus1].
        utils._BYREF_FLOAT64, # Input, xcap.
        utils._BYREF_FLOAT64, # Output, p.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nplus1 = utils._EngineIntScalarType(
        nplus1, {'fun_name': fun_name, 'entity': 'nplus1'},
    ).to_ctypes()
    _nplus1_py = utils._EngineIntScalarType.to_py(_nplus1)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nplus1_py,),
    ).to_ctypes()
    _xcap = utils._EngineFloat64ScalarType(
        xcap, {'fun_name': fun_name, 'entity': 'xcap'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nplus1,
        _a,
        _xcap,
        _p,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xcap = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'EPS = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(xcap) <= 1+4*EPS, where EPS is machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nplus1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nplus1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        return_data = _p_py
    _err_class.manage()
    return return_data

def dim1_cheb_glp(nplus1, f, a):
    u"""
    Least squares polynomial fit, special data points (including
    interpolation).

    ``dim1_cheb_glp`` computes the coefficients of a polynomial, in its
    Chebyshev series form, which interpolates (passes exactly through)
    data at a special set of points.
    Least squares polynomial approximations can also be obtained.

    For full information please refer to the NAG Library document for
    e02af

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02aff.html

    Parameters
    ----------
    nplus1 : int
        The number n+1 of data points (one greater than the degree n of
        the interpolating polynomial).

    f : float, shape (`nplus1`)
        For r = 1,2,...,n+1, `f`[r-1] must contain f_r the value of the
        dependent variable (ordinate) corresponding to the value

            x__r = cos(pi(r-1)/n)

        of the independent variable (abscissa) x_, or equivalently to
        the value

            x(r) = 1/2(x_max-x_min)cos(pi(r-1)/n)+1/2(x_max+x_min)

        of your original variable x. Here x_max and x_min are
        respectively the upper and lower ends of the range over which
        you wish to interpolate.

    a : float, shape (`nplus1`), modified in place
        `On exit`: `a`[j-1] is the coefficient a_j in the interpolating
        polynomial, for j = 1,2,...,n+1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nplus1` = *<value>*.

            Constraint: `nplus1` >= 2.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_cheb_glp`` computes the coefficients a_j, for j =
    1,2,...,n+1, in the Chebyshev series

        1/2a_1 T_0(x_)+a_2 T_1(x_)+a_3 T_2(x_)+...+a_{n+1}T_n(x_),

    which interpolates the data f_r at the points

        x__r = cos((r-1)pi/n), r = 1,2,...,n+1.

    Here T_j(x_) denotes the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    The use of these points minimizes the risk of unwanted fluctuations
    in the polynomial and is recommended when the data abscissae can be
    chosen by you, e.g., when the data is given as a graph.
    For further advantages of this choice of points, see Clenshaw
    (1962).

    In terms of your original variables, x say, the values of x at which
    the data f_r are to be provided are

        x_r = 1/2(x_max-x_min)cos(pi(r-1)/n)+1/2(x_max+x_min), r =
        1,2,...,n+1

    where x_max and x_min are respectively the upper and lower ends of
    the range of x over which you wish to interpolate.

    Truncation of the resulting series after the term involving a_{i+1},
    say, yields a least squares approximation to the data.
    This approximation, p(x_), say, is the polynomial of degree i which
    minimizes

        1/2 epsilon_1^2+epsilon_2^2+epsilon_3^2+...+epsilon_n^2+1/2
        epsilon_n+1^2,

    where the residual epsilon_r = p(x__r)-f_r, for r = 1,2,...,n+1.

    The method employed is based on the application of the three-term
    recurrence relation due to Clenshaw (1955) for the evaluation of the
    defining expression for the Chebyshev coefficients (see, for
    example, Clenshaw (1962)).
    The modifications to this recurrence relation suggested by Reinsch
    and Gentleman (see Gentleman (1969)) are used to give greater
    numerical stability.

    For further details of the algorithm and its use see Cox (1974) and
    Cox and Hayes (1973).

    Subsequent evaluation of the computed polynomial, perhaps truncated
    after an appropriate number of terms, should be carried out using
    ``dim1_cheb_eval``.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_glp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nplus1.
        utils._EngineFloat64ArrayType.c_type, # Input, f[nplus1].
        utils._EngineFloat64ArrayType.c_type, # Output, a[nplus1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nplus1 = utils._EngineIntScalarType(
        nplus1, {'fun_name': fun_name, 'entity': 'nplus1'},
    ).to_ctypes()
    _nplus1_py = utils._EngineIntScalarType.to_py(_nplus1)
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_nplus1_py,),
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nplus1_py,),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nplus1,
        _f,
        _a,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nplus1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nplus1 >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_cheb_con(m, k, xmin, xmax, x, y, w, mf, xf, yf, lyf, ip, a, s, resid):
    u"""
    Least squares polynomial fit, values and derivatives may be
    constrained, arbitrary data points.

    ``dim1_cheb_con`` computes constrained weighted least squares
    polynomial approximations in Chebyshev series form to an arbitrary
    set of data points.
    The values of the approximations and any number of their derivatives
    can be specified at selected points.

    For full information please refer to the NAG Library document for
    e02ag

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02agf.html

    Parameters
    ----------
    m : int
        m, the number of data points to be fitted.

    k : int
        k, the maximum degree required.

    xmin : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. Unless there are specific reasons to the
        contrary, it is recommended that `xmin` and `xmax` be set
        respectively to the lowest and highest value among the x_r and
        xf_r. This avoids the danger of extrapolation provided there is
        a constraint point or data point with nonzero weight at each end
        point.

    xmax : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. Unless there are specific reasons to the
        contrary, it is recommended that `xmin` and `xmax` be set
        respectively to the lowest and highest value among the x_r and
        xf_r. This avoids the danger of extrapolation provided there is
        a constraint point or data point with nonzero weight at each end
        point.

    x : float, shape (`m`)
        `x`[r-1] must contain the value x_r of the independent variable
        at the rth data point, for r = 1,2,...,m.

    y : float, shape (`m`)
        `y`[r-1] must contain y_r, the value of the dependent variable
        at the rth data point, for r = 1,2,...,m.

    w : float, shape (`m`)
        `w`[r-1] must contain the weight w_r to be applied to the data
        point x_r, for r = 1,2,...,m. For advice on the choice of
        weights see [the E02 Introduction]. Negative weights are treated
        as positive. A zero weight causes the corresponding data point
        to be ignored. Zero weight should be given to any data point
        whose x and y values both coincide with those of a constraint
        (otherwise the denominators involved in the root mean square
        residuals S_i will be slightly in error).

    mf : int
        mf, the number of values of the independent variable at which a
        constraint is specified.

    xf : float, shape (`mf`)
        `xf`[r-1] must contain xf_r, the value of the independent
        variable at which a constraint is specified, for r =
        1,2,...,`mf`.

    yf : float, shape (`lyf`)
        The values which the approximating polynomials and their
        derivatives are required to take at the points specified in
        `xf`. For each value of `xf`[r-1], `yf` contains in successive
        elements the required value of the approximation, its first
        derivative, second derivative, ...,p_rth derivative, for r =
        1,2,...,mf. Thus the value, yf_s, which the kth derivative of
        each approximation (k = 0 referring to the approximation itself)
        is required to take at the point `xf`[r-1] must be contained in
        `yf`[s-1], where

            s = r+k+p_1+p_2+...+p_{r-1},

        where k = 0,1,...,p_r and r = 1,2,...,mf. The derivatives are
        with respect to the independent variable x.

    lyf : int
        The dimension of the array `yf`.

    ip : int, shape (`mf`)
        `ip`[r-1] must contain p_r, the order of the highest-order
        derivative specified at `xf`[r-1], for r = 1,2,...,mf. p_r = 0
        implies that the value of the approximation at `xf`[r-1] is
        specified, but not that of any derivative.

    a : float, shape (`k`+1, `k`+1), modified in place
        `On exit`: `a`[i,j] contains the coefficient a_{ij} in the
        approximating polynomial of degree i, for j = 0,1,...,i, for i =
        n,...,k.

    s : float, shape (`k`+1), modified in place
        `On exit`: `s`[i] contains S_i, for i = n,...,k, the root mean
        square residual corresponding to the approximating polynomial of
        degree i. In the case where the number of data points with
        nonzero weight is equal to k+1-n, S_i is indeterminate: the
        function sets it to zero. For the interpretation of the values
        of S_i and their use in selecting an appropriate degree, see
        [the E02 Introduction].

    resid : float, shape (`m`), modified in place
        `On exit`: contains weighted residuals of the highest degree of
        fit determined (k). The residual at x_r is in element
        `resid`[r-1], for r = 1,2,...,m.

    Returns
    -------
    n : int
        Contains the total number of constraint conditions imposed: `n`
        = `mf`+p_1+p_2+...+p_`mf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `k` >= `n`.

        (`errno` 1)
            On entry, `lyf` = *<value>* and N = *<value>*.

            Constraint: `lyf` >= N, where N is the total number of
            constraints.

        (`errno` 1)
            On entry, `mf` = *<value>*.

            Constraint: `mf` >= 1.

        (`errno` 2)
            On entry, I = *<value>* and `ip`[I-1] = *<value>*.

            Constraint: `ip`[I-1] >= 0.

        (`errno` 3)
            On entry, I = *<value>*, `xf`[I-1] = *<value>*, J =
            *<value>* and `xf`[J-1] = *<value>*.

            Constraint: `xf`[I-1] != `xf`[J-1].

        (`errno` 3)
            On entry, `xf`[I-1] lies outside interval [`xmin`,`xmax`]: I
            = *<value>*, `xf`[I-1] = *<value>*, `xmin` = *<value>* and
            `xmax` = *<value>*.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `x`[I-1] lies outside interval [`xmin`,`xmax`] for
            some I.

        (`errno` 4)
            On entry, `x`[I-1] lies outside interval [`xmin`,`xmax`]: I
            = *<value>*, `x`[I-1] = *<value>*, `xmin` = *<value>* and
            `xmax` = *<value>*.

        (`errno` 5)
            On entry, i = *<value>*, `x`[i-1] = *<value>* and `x`[i-2] =
            *<value>*.

            Constraint: `x`[i-1] >= `x`[i-2].

        (`errno` 6)
            On entry, `k`+1 > m''+`n`, where m'' is the number of data
            points with nonzero weight and distinct abscissae different
            from all `xf`, and `n` is the total number of constraints:
            `k`+1 = *<value>*, m'' = *<value>* and `n` = *<value>*.

        (`errno` 7)
            The polynomials mu(x) and/or nu(x) cannot be found. The
            problem is too ill-conditioned.

    Notes
    -----
    ``dim1_cheb_con`` determines least squares polynomial approximations
    of degrees up to k to the set of data points (x_r,y_r) with weights
    w_r, for r = 1,2,...,m.
    The value of k, the maximum degree required, is to be prescribed by
    you.
    At each of the values xf_r, for r = 1,2,...,mf, of the independent
    variable x, the approximations and their derivatives up to order p_r
    are constrained to have one of the values yf_s, for s = 1,2,...,n,
    specified by you, where n = mf+sum_r = 0^mfp_r.

    The approximation of degree i has the property that, subject to the
    imposed constraints, it minimizes sigma_i, the sum of the squares of
    the weighted residuals epsilon_r, for r = 1,2,...,m, where

        epsilon_r = w_r(y_r-f_i(x_r))

    and f_i(x_r) is the value of the polynomial approximation of degree
    i at the rth data point.

    Each polynomial is represented in Chebyshev series form with
    normalized argument x_.
    This argument lies in the range -1 to +1 and is related to the
    original variable x by the linear transformation

        x_= (2x-(x_max+x_min))/((x_max-x_min))

    where x_min and x_max, specified by you, are respectively the lower
    and upper end points of the interval of x over which the polynomials
    are to be defined.

    The polynomial approximation of degree i can be written as

        1/2a_{i,0}+a_{i,1}T_1(x_)+...+a_{ij}T_j(x_)+...+a_{ii}T_i(x_)

    where T_j(x_) is the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    For i = n,n+1,...,k, the function produces the values of the
    coefficients a_{ij}, for j = 0,1,...,i, together with the value of
    the root mean square residual,

        S_i = sqrt(sigma_i/((m'+n-i-1))),

    where m' is the number of data points with nonzero weight.

    Values of the approximations may subsequently be computed using
    ``dim1_cheb_eval`` or ``dim1_cheb_eval2``.

    First ``dim1_cheb_con`` determines a polynomial mu(x_), of degree
    n-1, which satisfies the given constraints, and a polynomial nu(x_),
    of degree n, which has value (or derivative) zero wherever a
    constrained value (or derivative) is specified.
    It then fits y_r-mu(x_r), for r = 1,2,...,m, with polynomials of the
    required degree in x_ each with factor nu(x_).
    Finally the coefficients of mu(x_) are added to the coefficients of
    these fits to give the coefficients of the constrained polynomial
    approximations to the data points (x_r,y_r), for r = 1,2,...,m.
    The method employed is given in Hayes (1970): it is an extension of
    Forsythe's orthogonal polynomials method (see Forsythe (1957)) as
    modified by Clenshaw (see Clenshaw (1960)).

    References
    ----------
    Clenshaw, C W, 1960, `Curve fitting with a digital computer`,
    Comput. J. (2), 170--173

    Forsythe, G E, 1957, `Generation and use of orthogonal polynomials
    for data fitting with a digital computer`, J. Soc. Indust. Appl.
    Math. (5), 74--88

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_con'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, kplus1.
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._BYREF_INT, # Input, mf.
        utils._EngineFloat64ArrayType.c_type, # Input, xf[mf].
        utils._EngineFloat64ArrayType.c_type, # Input, yf[lyf].
        utils._BYREF_INT, # Input, lyf.
        utils._EngineIntArrayType.c_type, # Input, ip[mf].
        utils._EngineFloat64ArrayType.c_type, # Output, a[lda, k+1].
        utils._EngineFloat64ArrayType.c_type, # Output, s[k+1].
        utils._BYREF_INT, # Output, np1.
        utils._EngineFloat64ArrayType.c_type, # Output, wrk[m].
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _kplus1_py = utils._EngineFloat64ArrayType.get_shape(s, {'fun_name': fun_name, 'entity': 's'}, exp_rank=1)[0]
    _kplus1 = utils._EngineIntScalarType(
        _kplus1_py, {'fun_name': fun_name, 'entity': '_kplus1'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _mf = utils._EngineIntScalarType(
        mf, {'fun_name': fun_name, 'entity': 'mf'},
    ).to_ctypes()
    _mf_py = utils._EngineIntScalarType.to_py(_mf)
    _xf = utils._EngineFloat64ArrayType(
        xf, {'fun_name': fun_name, 'entity': 'xf'},
        exp_shape=(_mf_py,),
    ).to_ctypes()
    _lyf = utils._EngineIntScalarType(
        lyf, {'fun_name': fun_name, 'entity': 'lyf'},
    ).to_ctypes()
    _ip_cls = utils._EngineIntArrayType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
        exp_shape=(_mf_py,),
    )
    _ip = _ip_cls.to_ctypes()
    _yf = utils._EngineFloat64ArrayType(
        yf, {'fun_name': fun_name, 'entity': 'yf'},
        exp_shape=(_mf_py+_ip_cls.get_sum(0, _mf_py-1),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _k_py+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k_py+1,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    _np1 = utils._EngineIntScalarType.empty_ctypes()
    utils._EngineFloat64ArrayType(
        resid, {'fun_name': fun_name, 'entity': 'resid'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(_m_py)
    _lwrk = utils._EngineIntScalarType.c_type(0)
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _liwrk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _m,
        _kplus1,
        _lda,
        _xmin,
        _xmax,
        _x,
        _y,
        _w,
        _mf,
        _xf,
        _yf,
        _lyf,
        _ip,
        _a,
        _s,
        _np1,
        _wrk,
        _lwrk,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lyf = {:s} and N = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lyf >= N, where N is the total number of constraints.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mf >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ip[I-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, xf[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'J = {:s} and xf[J-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xf[I-1] != xf[J-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xf[I-1] lies outside interval [xmin,xmax]:',
                    'I = {:s}, xf[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmin = {:s} and xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[I-1] lies outside interval [xmin,xmax] for some I.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[I-1] lies outside interval [xmin,xmax]:',
                    'I = {:s}, x[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmin = {:s} and xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, x[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[i-1] >= x[i-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k+1 > m\'\'+n, where m\'\' is the number of data',
                    'points with nonzero weight and distinct abscissae different from',
                    'all xf, and n is the total number of constraints:',
                    'k+1 = {:s}, m\'\' = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The polynomials mu(x) and/or nu(x) cannot be found.',
                    'The problem is too ill-conditioned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _s_cls.refresh_py(_s)
        resid[:] = utils._EngineFloat64ArrayType.from_pointer(_wrk, (_m_py,))
        utils._unlink_pointer(_wrk)
        _np1_py = utils._EngineIntScalarType.to_py(_np1)
        _n_py = _np1_py-1
        return_data = _n_py
    _err_class.manage()
    return return_data

def dim1_cheb_deriv(n, xmin, xmax, a, ia1, la, adif, iadif1, ladif):
    u"""
    Derivative of fitted polynomial in Chebyshev series form.

    ``dim1_cheb_deriv`` determines the coefficients in the Chebyshev
    series representation of the derivative of a polynomial given in
    Chebyshev series form.

    For full information please refer to the NAG Library document for
    e02ah

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02ahf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, shape (`la`)
        The Chebyshev coefficients of the polynomial p(x). Specifically,
        element i*`ia1` of `a` must contain the coefficient a_i, for i =
        0,1,...,n. Only these n+1 elements will be accessed.

        Unchanged on exit, but see `adif`, below.

    ia1 : int
        The index increment of `a`. Most frequently the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3. See also
        [Further Comments].

    la : int
        The dimension of the array `a`.

    adif : float, shape (`ladif`), modified in place
        `On exit`: the Chebyshev coefficients of the derived polynomial
        q(x). (The differentiation is with respect to the variable x.)
        Specifically, element i*`iadif1` of `adif` contains the
        coefficient a__i, for i = 0,1,...,n-1. Additionally, element
        n*`iadif1` is set to zero. A call of the function may have the
        array name `adif` the same as `a`, provided that note is taken
        of the order in which elements are overwritten, when choosing
        the starting elements and increments `ia1` and `iadif1`, i.e.,
        the coefficients a_0,a_1,...,a_{i-1} must be intact after
        coefficient a__i is stored. In particular, it is possible to
        overwrite the a_i completely by having `ia1` = `iadif1`, and the
        actual arrays for `a` and `adif` identical.

    iadif1 : int
        The index increment of `adif`. Most frequently the Chebyshev
        coefficients are required in adjacent elements of `adif`, and
        `iadif1` must be set to 1. However, if, for example, they are to
        be stored in `adif`[0],`adif`[3],`adif`[6],..., the value of
        `iadif1` must be 3. See [Further Comments].

    ladif : int
        The dimension of the array `adif`.

    Returns
    -------
    patm1 : float
        The value of p(x_min). If this value is passed to the
        integration function ``dim1_cheb_integ`` with the coefficients
        of q(x), the original polynomial p(x) is recovered, including
        its constant coefficient.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ladif` = *<value>*, `n` = *<value>* and `iadif1`
            = *<value>*.

            Constraint: `ladif` > `n`*`iadif1`.

        (`errno` 1)
            On entry, `iadif1` = *<value>*.

            Constraint: `iadif1` >= 1.

        (`errno` 1)
            On entry, `la` = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: `la` > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``dim1_cheb_deriv`` forms the polynomial which is the derivative of
    a given polynomial.
    Both the original polynomial and its derivative are represented in
    Chebyshev series form.
    Given the coefficients a_i, for i = 0,1,...,n, of a polynomial p(x)
    of degree n, where

        p(x) = 1/2a_0+a_1 T_1(x_)+...+a_n T_n(x_)

    the function returns the coefficients a__i, for i = 0,1,...,n-1, of
    the polynomial q(x) of degree n-1, where

        q(x) = dp(x)/dx = 1/2a__0+a__1 T_1(x_)+...+a__{n-1}T_{n-1}(x_).

    Here T_j(x_) denotes the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    It is assumed that the normalized variable x_ in the interval
    [-1,+1] was obtained from your original variable x in the interval
    [x_min,x_max] by the linear transformation

        x_= (2x-(x_max+x_min))/(x_max-x_min)

    and that you require the derivative to be with respect to the
    variable x.
    If the derivative with respect to x_ is required, set x_max = 1 and
    x_min = -1.

    Values of the derivative can subsequently be computed, from the
    coefficients obtained, by using ``dim1_cheb_eval2``.

    The method employed is that of Chebyshev series (see Module 8 of
    Modern Computing Methods (1961)), modified to obtain the derivative
    with respect to x.
    Initially setting a__{n+1} = a__n = 0, the function forms
    successively

        a__{i-1} = a__{i+1}+2/(x_max-x_min)2ia_i, i = n,n-1,...,1.

    References
    ----------
    Modern Computing Methods, 1961, `Chebyshev-series`, NPL Notes on
    Applied Science (16), (2nd Edition), HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, ia1.
        utils._BYREF_INT, # Input, la.
        utils._BYREF_FLOAT64, # Output, patm1.
        utils._EngineFloat64ArrayType.c_type, # Output, adif[ladif].
        utils._BYREF_INT, # Input, iadif1.
        utils._BYREF_INT, # Input, ladif.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ia1 = utils._EngineIntScalarType(
        ia1, {'fun_name': fun_name, 'entity': 'ia1'},
    ).to_ctypes()
    _ia1_py = utils._EngineIntScalarType.to_py(_ia1)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(1+_n_py*_ia1_py,),
    ).to_ctypes()
    _patm1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _ladif = utils._EngineIntScalarType(
        ladif, {'fun_name': fun_name, 'entity': 'ladif'},
    ).to_ctypes()
    _iadif1 = utils._EngineIntScalarType(
        iadif1, {'fun_name': fun_name, 'entity': 'iadif1'},
    ).to_ctypes()
    _iadif1_py = utils._EngineIntScalarType.to_py(_iadif1)
    _adif_cls = utils._EngineFloat64ArrayType(
        adif, {'fun_name': fun_name, 'entity': 'adif'},
        exp_shape=(1+_n_py*_iadif1_py,),
        check_contents=False,
    )
    _adif = _adif_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmin,
        _xmax,
        _a,
        _ia1,
        _la,
        _patm1,
        _adif,
        _iadif1,
        _ladif,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ladif = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'iadif1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ladif > n*iadif1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iadif1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iadif1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: la > n*ia1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ia1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmax > xmin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _adif_cls.refresh_py(_adif)
        _patm1_py = utils._EngineFloat64ScalarType.to_py(_patm1)
        return_data = _patm1_py
    _err_class.manage()
    return return_data

def dim1_cheb_integ(n, xmin, xmax, a, ia1, la, qatm1, aintc, iaint1, laint):
    u"""
    Integral of fitted polynomial in Chebyshev series form.

    ``dim1_cheb_integ`` determines the coefficients in the Chebyshev
    series representation of the indefinite integral of a polynomial
    given in Chebyshev series form.

    For full information please refer to the NAG Library document for
    e02aj

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02ajf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, shape (`la`)
        The Chebyshev coefficients of the polynomial p(x). Specifically,
        element i*`ia1` of `a` must contain the coefficient a_i, for i =
        0,1,...,n. Only these n+1 elements will be accessed.

        Unchanged on exit, but see `aintc`, below.

    ia1 : int
        The index increment of `a`. Most frequently the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3. See also
        [Further Comments].

    la : int
        The dimension of the array `a`.

    qatm1 : float
        The value that the integrated polynomial is required to have at
        the lower end point of its interval of definition, i.e., at x_=
        -1 which corresponds to x = x_min. Thus, `qatm1` is a constant
        of integration and will normally be set to zero by you.

    aintc : float, shape (`laint`), modified in place
        `On exit`: the Chebyshev coefficients of the integral q(x). (The
        integration is with respect to the variable x, and the constant
        coefficient is chosen so that q(x_min) equals `qatm1`).
        Specifically, element i*`iaint1` of `aintc` contains the
        coefficient a'_i, for i = 0,1,...,n+1. A call of the function
        may have the array name `aintc` the same as `a`, provided that
        note is taken of the order in which elements are overwritten
        when choosing starting elements and increments `ia1` and
        `iaint1`: i.e., the coefficients, a_0,a_1,...,a_{i-2} must be
        intact after coefficient a'_i is stored. In particular it is
        possible to overwrite the a_i entirely by having `ia1` =
        `iaint1`, and the actual array for `a` and `aintc` identical.

    iaint1 : int
        The index increment of `aintc`. Most frequently the Chebyshev
        coefficients are required in adjacent elements of `aintc`, and
        `iaint1` must be set to 1. However, if, for example, they are to
        be stored in `aintc`[0],`aintc`[3],`aintc`[6],..., the value of
        `iaint1` must be 3. See also [Further Comments].

    laint : int
        The dimension of the array `aintc`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `laint` = *<value>*, `n` = *<value>* and `iaint1`
            = *<value>*.

            Constraint: `laint` > (`n`+1)*`iaint1`.

        (`errno` 1)
            On entry, `iaint1` = *<value>*.

            Constraint: `iaint1` >= 1.

        (`errno` 1)
            On entry, `la` = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: `la` > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``dim1_cheb_integ`` forms the polynomial which is the indefinite
    integral of a given polynomial.
    Both the original polynomial and its integral are represented in
    Chebyshev series form.
    If supplied with the coefficients a_i, for i = 0,1,...,n, of a
    polynomial p(x) of degree n, where

        p(x) = 1/2a_0+a_1 T_1(x_)+...+a_n T_n(x_),

    the function returns the coefficients a'_i, for i = 0,1,...,n+1, of
    the polynomial q(x) of degree n+1, where

        q(x) = 1/2a'_0+a'_1T_1(x_)+...+a'_n+1T_{n+1}(x_),

    and

        q(x) = intp(x)dx.

    Here T_j(x_) denotes the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    It is assumed that the normalized variable x_ in the interval
    [-1,+1] was obtained from your original variable x in the interval
    [x_min,x_max] by the linear transformation

        x_= (2x-(x_max+x_min))/(x_max-x_min)

    and that you require the integral to be with respect to the variable
    x.
    If the integral with respect to x_ is required, set x_max = 1 and
    x_min = -1.

    Values of the integral can subsequently be computed, from the
    coefficients obtained, by using ``dim1_cheb_eval2``.

    The method employed is that of Chebyshev series (see Module 8 of
    Modern Computing Methods (1961)), modified for integrating with
    respect to x.
    Initially taking a_{n+1} = a_{n+2} = 0, the function forms
    successively

        a'_i = (a_{i-1}-a_{i+1})/2i*(x_max-x_min)/2, i = n+1,n,...,1.

    The constant coefficient a'_0 is chosen so that q(x) is equal to a
    specified value, `qatm1`, at the lower end point of the interval on
    which it is defined, i.e., x_= -1, which corresponds to x = x_min.

    References
    ----------
    Modern Computing Methods, 1961, `Chebyshev-series`, NPL Notes on
    Applied Science (16), (2nd Edition), HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_integ'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, ia1.
        utils._BYREF_INT, # Input, la.
        utils._BYREF_FLOAT64, # Input, qatm1.
        utils._EngineFloat64ArrayType.c_type, # Output, aintc[laint].
        utils._BYREF_INT, # Input, iaint1.
        utils._BYREF_INT, # Input, laint.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ia1 = utils._EngineIntScalarType(
        ia1, {'fun_name': fun_name, 'entity': 'ia1'},
    ).to_ctypes()
    _ia1_py = utils._EngineIntScalarType.to_py(_ia1)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(1+_n_py*_ia1_py,),
    ).to_ctypes()
    _qatm1 = utils._EngineFloat64ScalarType(
        qatm1, {'fun_name': fun_name, 'entity': 'qatm1'},
    ).to_ctypes()
    _laint = utils._EngineIntScalarType(
        laint, {'fun_name': fun_name, 'entity': 'laint'},
    ).to_ctypes()
    _iaint1 = utils._EngineIntScalarType(
        iaint1, {'fun_name': fun_name, 'entity': 'iaint1'},
    ).to_ctypes()
    _iaint1_py = utils._EngineIntScalarType.to_py(_iaint1)
    _aintc_cls = utils._EngineFloat64ArrayType(
        aintc, {'fun_name': fun_name, 'entity': 'aintc'},
        exp_shape=(1+(_n_py+1)*_iaint1_py,),
        check_contents=False,
    )
    _aintc = _aintc_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmin,
        _xmax,
        _a,
        _ia1,
        _la,
        _qatm1,
        _aintc,
        _iaint1,
        _laint,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, laint = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'iaint1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: laint > (n+1)*iaint1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iaint1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iaint1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: la > n*ia1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ia1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmax > xmin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _aintc_cls.refresh_py(_aintc)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_cheb_eval2(n, xmin, xmax, a, ia1, la, x):
    u"""
    Evaluation of fitted polynomial in one variable from Chebyshev
    series form.

    ``dim1_cheb_eval2`` evaluates a polynomial from its Chebyshev series
    representation, allowing an arbitrary index increment for accessing
    the array of coefficients.

    For full information please refer to the NAG Library document for
    e02ak

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02akf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x_).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, shape (`la`)
        The Chebyshev coefficients of the polynomial p(x_).
        Specifically, element i*`ia1` must contain the coefficient a_i,
        for i = 0,1,...,n. Only these n+1 elements will be accessed.

    ia1 : int
        The index increment of `a`. Most frequently, the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if, for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3.

    la : int
        The dimension of the array `a`.

    x : float
        The argument x at which the polynomial is to be evaluated.

    Returns
    -------
    result : float
        The value of the polynomial p(x_).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `la` = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: `la` > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `n`+1 = *<value>*>.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `x` does not lie in [`xmin`,`xmax`]: `x` =
            *<value>*, `xmin` = *<value>* and `xmax` = *<value>*.

    Notes
    -----
    If supplied with the coefficients a_i, for i = 0,1,...,n, of a
    polynomial p(x_) of degree n, where

        p(x_) = 1/2a_0+a_1 T_1(x_)+...+a_n T_n(x_),

    ``dim1_cheb_eval2`` returns the value of p(x_) at a user-specified
    value of the variable x.
    Here T_j(x_) denotes the Chebyshev polynomial of the first kind of
    degree j with argument x_.
    It is assumed that the independent variable x_ in the interval
    [-1,+1] was obtained from your original variable x in the interval
    [x_min,x_max] by the linear transformation

        x_= (2x-(x_max+x_min))/(x_max-x_min).

    The coefficients a_i may be supplied in the array `a`, with any
    increment between the indices of array elements which contain
    successive coefficients.
    This enables the function to be used in surface fitting and other
    applications, in which the array might have two or more dimensions.

    The method employed is based on the three-term recurrence relation
    due to Clenshaw (see Clenshaw (1955)), with modifications due to
    Reinsch and Gentleman (see Gentleman (1969)).
    For further details of the algorithm and its use see Cox (1973) and
    Cox and Hayes (1973).

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Cox, M G, 1973, `A data-fitting package for the non-specialist
    user`, NPL Report NAC 40, National Physical Laboratory

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_cheb_eval2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, ia1.
        utils._BYREF_INT, # Input, la.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, result.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ia1 = utils._EngineIntScalarType(
        ia1, {'fun_name': fun_name, 'entity': 'ia1'},
    ).to_ctypes()
    _ia1_py = utils._EngineIntScalarType.to_py(_ia1)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py*_ia1_py+1,),
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _result = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmin,
        _xmax,
        _a,
        _ia1,
        _la,
        _x,
        _result,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmax > xmin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: la > n*ia1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ia1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ia1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n+1 = {:s}>.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x does not lie in [xmin,xmax]:',
                    'x = {:s}, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _result_py = utils._EngineFloat64ScalarType.to_py(_result)
        return_data = _result_py
    _err_class.manage()
    return return_data

def dim1_minimax_polynomial(n, x, y, m, a):
    u"""
    Minimax curve fit by polynomials.

    ``dim1_minimax_polynomial`` calculates a minimax polynomial fit to a
    set of data points.

    For full information please refer to the NAG Library document for
    e02al

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02alf.html

    Parameters
    ----------
    n : int
        n, the number of data points.

    x : float, shape (`n`)
        The values of the x coordinates, x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        The values of the y coordinates, y_i, for i = 1,2,...,n.

    m : int
        m, where m is the degree of the polynomial to be found.

    a : float, shape (`m`+1), modified in place
        `On exit`: the coefficients a_i of the minimax polynomial, for i
        = 0,1,...,m.

    Returns
    -------
    ref : float
        The final reference deviation, i.e., the maximum deviation of
        the computed polynomial evaluated at x_i from the reference
        values y_i, for i = 1,2,...,n. `ref` may return a negative value
        which indicates that the algorithm started to cycle due to
        round-off errors.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` < 100.

        (`errno` 2)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` < `n`-1.

        (`errno` 3)
            On entry, i = *<value>*, `x`[i] = *<value>* and `x`[i-1] =
            *<value>*.

            Constraint: `x`[i] > `x`[i-1].

    Notes
    -----
    Given a set of data points (x_i,y_i), for i = 1,2,...,n,
    ``dim1_minimax_polynomial`` uses the exchange algorithm to compute
    an mth-degree polynomial

        P(x) = a_0+a_1 x+a_2 x^2+...+a_m x^m

    such that max_i|P(x_i)-y_i| is a minimum.

    The function also returns a number whose absolute value is the final
    reference deviation (see [Parameters]).
    The function is an adaptation of Boothroyd (1967).

    References
    ----------
    Boothroyd, J B, 1967, `Algorithm 318`, Comm. ACM (10), 801

    Stieffel, E, 1959, `Numerical methods of Tchebycheff approximation`,
    On Numerical Approximation, (ed R E Langer), 217--232, University of
    Wisconsin Press
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_minimax_polynomial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02alft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Output, a[m+1].
        utils._BYREF_FLOAT64, # Output, ref.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py+1,),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _ref = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _m,
        _a,
        _ref,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m < 100.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m < n-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, x[i] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[i] > x[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _ref_py = utils._EngineFloat64ScalarType.to_py(_ref)
        return_data = _ref_py
    _err_class.manage()
    return return_data

def dim1_spline_knots(m, ncap7, x, y, w, lamda, c):
    u"""
    Least squares curve cubic spline fit (including interpolation).

    ``dim1_spline_knots`` computes a weighted least squares
    approximation to an arbitrary set of data points by a cubic spline
    with knots prescribed by you.
    Cubic spline interpolation can also be carried out.

    For full information please refer to the NAG Library document for
    e02ba

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02baf.html

    Parameters
    ----------
    m : int
        The number m of data points.

    ncap7 : int
        n_+7, where n_ is the number of intervals of the spline (which
        is one greater than the number of interior knots, i.e., the
        knots strictly within the range x_1 to x_m) over which the
        spline is defined.

    x : float, shape (`m`)
        The values x_r of the independent variable (abscissa), for r =
        1,2,...,m. The values must satisfy the Schoenberg--Whitney
        conditions (see [Further Comments]).

    y : float, shape (`m`)
        The values y_r of the dependent variable (ordinate), for r =
        1,2,...,m.

    w : float, shape (`m`)
        The values w_r of the weights, for r = 1,2,...,m. For advice on
        the choice of weights, see [the E02 Introduction].

    lamda : float, shape (`ncap7`), modified in place
        `On entry`: `lamda`[i-1] must be set to the (i-4)th (interior)
        knot, lambda_i, for i = 5,6,...,n_+3.

        `On exit`: the input values are unchanged, and `lamda`[i-1], for
        i = 1,2,...,4, `ncap7`-3, `ncap7`-2, `ncap7`-1, `ncap7` contains
        the additional (exterior) knots introduced by the function. For
        advice on the choice of knots, see [the E02 Introduction].

    c : float, shape (`ncap7`), modified in place
        `On exit`: the coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not used.

    Returns
    -------
    ss : float
        The residual sum of squares, theta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, J = *<value>*, `lamda`[J-1] = *<value>* and
            `lamda`[J] = *<value>*.

            Constraint: `lamda`[J-1] <= `lamda`[J].

        (`errno` 1)
            On entry, `ncap7` = *<value>*, `lamda`[`ncap7`-5] =
            *<value>*, `m` = *<value>* and `x`[`m`-1] = *<value>*.

            Constraint: `lamda`[`ncap7`-5] < `x`[`m`-1].

        (`errno` 1)
            On entry, `lamda`[4] = *<value>* and `x`[0] = *<value>*.

            Constraint: `lamda`[4] > `x`[0].

        (`errno` 2)
            On entry, i = *<value>* and `w`[i-1] = *<value>*.

            Constraint: `w`[i-1] > 0.0.

        (`errno` 3)
            On entry, the `x` values are not in nondecreasing order. I =
            *<value>*, `x`[I-1] = *<value>*, J = *<value>* and
            xdist[J-1] = *<value>*.

            Constraint: `x`[I-1] >= xdist[J-1], where xdist is the set
            of distinct x-values.

        (`errno` 4)
            On entry, `ncap7` = *<value>* and mdist = *<value>*.

            Constraint: `ncap7` <= mdist+4, where mdist is the number of
            distinct x-values.

        (`errno` 4)
            On entry, `ncap7` = *<value>* and `m` = *<value>*.

            Constraint: `ncap7` <= `m`+4.

        (`errno` 4)
            On entry, `ncap7` = *<value>*.

            Constraint: `ncap7` >= 8.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. I =
            *<value>*, xdist[I-1] = *<value>*, J = *<value>* and
            `lamda`[J-1] = *<value>*.

            Constraint: xdist[I-1] < `lamda`[J-1], where xdist is the
            set of distinct x-values.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. L =
            *<value>*, xdist[L-1] = *<value>*, I = *<value>* and
            `lamda`[I-1] = *<value>*.

            Constraint: xdist[L-1] > `lamda`[I-1], where xdist is the
            set of distinct x-values.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. J =
            *<value>*, xdist[J-1] = *<value>*, J+4 = *<value>* and
            `lamda`[J+3] = *<value>*.

            Constraint: xdist[J-1] < `lamda`[J+3], where xdist is the
            set of distinct x-values.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline_knots`` determines a least squares cubic spline
    approximation s(x) to the set of data points (x_r,y_r) with weights
    w_r, for r = 1,2,...,m.
    The value of `ncap7` = n_+7, where n_ is the number of intervals of
    the spline (one greater than the number of interior knots), and the
    values of the knots lambda_5,lambda_6,...,lambda_{n_+3}, interior to
    the data interval, are prescribed by you.

    s(x) has the property that it minimizes theta, the sum of squares of
    the weighted residuals epsilon_r, for r = 1,2,...,m, where

        epsilon_r = w_r(y_r-s(x_r)).

    The function produces this minimizing value of theta and the
    coefficients c_1,c_2,...,c_q, where q = n_+3, in the B-spline
    representation

        s(x) = sum_i = 1^qc_i N_i(x).

    Here N_i(x) denotes the normalized B-spline of degree 3 defined upon
    the knots lambda_i,lambda_{i+1},...,lambda_{i+4}.

    In order to define the full set of B-splines required, eight
    additional knots lambda_1,lambda_2,lambda_3,lambda_4 and
    lambda_{n_+4},lambda_{n_+5},lambda_{n_+6},lambda_{n_+7} are inserted
    automatically by the function.
    The first four of these are set equal to the smallest x_r and the
    last four to the largest x_r.

    The representation of s(x) in terms of B-splines is the most compact
    form possible in that only n_+3 coefficients, in addition to the
    n_+7 knots, fully define s(x).

    The method employed involves forming and then computing the least
    squares solution of a set of m linear equations in the coefficients
    c_i, for i = 1,2,...,n_+3.
    The equations are formed using a recurrence relation for B-splines
    that is unconditionally stable (see Cox (1972) and de Boor (1972)),
    even for multiple (coincident) knots.
    The least squares solution is also obtained in a stable manner by
    using orthogonal transformations, viz. a variant of Givens rotations
    (see Gentleman (1974) and Gentleman (1973)).
    This requires only one equation to be stored at a time.
    Full advantage is taken of the structure of the equations, there
    being at most four nonzero values of N_i(x) for any value of x and
    hence at most four coefficients in each equation.

    For further details of the algorithm and its use see Cox (1974), Cox
    (1975) and Cox and Hayes (1973).

    Subsequent evaluation of s(x) from its B-spline representation may
    be carried out using ``dim1_spline_eval``.
    If derivatives of s(x) are also required, ``dim1_spline_deriv`` may
    be used. ``dim1_spline_integ`` can be used to compute the definite
    integral of s(x).

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G, 1975, `Numerical methods for the interpolation and
    approximation of data by spline functions`, PhD Thesis, City
    University, London

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Gentleman, W M, 1973, `Least squares computations by Givens
    transformations without square roots`, J. Inst. Math. Applic. (12),
    329--336

    Gentleman, W M, 1974, `Algorithm AS 75. Basic procedures for large
    sparse or weighted linear least squares problems`, Appl. Statist.
    (23), 448--454

    Schoenberg, I J and Whitney, A, 1953, `On Polya frequency functions
    III`, Trans. Amer. Math. Soc. (74), 246--259
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_knots'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ncap7.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, lamda[ncap7].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ncap7].
        utils._BYREF_FLOAT64, # Output, ss.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ncap7 = utils._EngineIntScalarType(
        ncap7, {'fun_name': fun_name, 'entity': 'ncap7'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ncap7_py = utils._EngineIntScalarType.to_py(_ncap7)
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_ncap7_py,),
    )
    _lamda = _lamda_cls.to_ctypes()
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncap7_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _ss = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _ncap7,
        _x,
        _y,
        _w,
        _lamda,
        _work1,
        _work2,
        _c,
        _ss,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, J = {:s}, lamda[J-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and lamda[J] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[J-1] <= lamda[J].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[ncap7-5] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'x[m-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: lamda[ncap7-5] < x[m-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[4] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'x[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lamda[4] > x[0].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and w[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w[i-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the x values are not in nondecreasing order.',
                    'I = {:s}, x[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'J = {:s} and xdist[J-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[I-1] >= xdist[J-1], where xdist is the set of',
                    'distinct x-values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'mdist = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ncap7 <= mdist+4, where mdist is the number of',
                    'distinct x-values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ncap7 <= m+4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncap7 >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the Schoenberg--Whitney conditions fail to hold for at least one',
                    'subset of the distinct data abscissae. I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'xdist[I-1] = {:s}, J = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'lamda[J-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xdist[I-1] < lamda[J-1], where xdist is the set of',
                    'distinct x-values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the Schoenberg--Whitney conditions fail to hold for at least one',
                    'subset of the distinct data abscissae.',
                    'L = {:s}, xdist[L-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'I = {:s} and lamda[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xdist[L-1] > lamda[I-1], where xdist is the set of',
                    'distinct x-values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the Schoenberg--Whitney conditions fail to hold for at least one',
                    'subset of the distinct data abscissae. J = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'xdist[J-1] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'J+4 = {:s} and lamda[J+3] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xdist[J-1] < lamda[J+3], where xdist is the set of',
                    'distinct x-values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lamda_cls.refresh_py(_lamda)
        _c_cls.refresh_py(_c)
        _ss_py = utils._EngineFloat64ScalarType.to_py(_ss)
        return_data = _ss_py
    _err_class.manage()
    return return_data

def dim1_spline_eval(ncap7, lamda, c, x):
    u"""
    Evaluation of fitted cubic spline, function only.

    ``dim1_spline_eval`` evaluates a cubic spline from its B-spline
    representation.

    For full information please refer to the NAG Library document for
    e02bb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02bbf.html

    Parameters
    ----------
    ncap7 : int
        n_+7, where n_ is the number of intervals (one greater than the
        number of interior knots, i.e., the knots strictly within the
        range lambda_4 to lambda_{n_+4}) over which the spline is
        defined.

    lamda : float, shape (`ncap7`)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, shape (`ncap7`)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    x : float
        The argument x at which the cubic spline is to be evaluated.

    Returns
    -------
    s : float
        The value of the spline, s(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*, `ncap7` = *<value>* and
            `lamda`[`ncap7`-4] = *<value>*.

            Constraint: `x` <= `lamda`[`ncap7`-4].

        (`errno` 1)
            On entry, `x` = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `x` >= `lamda`[3].

        (`errno` 2)
            On entry, `ncap7` = *<value>*.

            Constraint: `ncap7` >= 8.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline_eval`` evaluates the cubic spline s(x) at a prescribed
    argument x from its augmented knot set lambda_i, for i =
    1,2,...,n+7, (see ``dim1_spline_knots``) and from the coefficients
    c_i, for i = 1,2,...,q in its B-spline representation

        s(x) = sum_i = 1^qc_i N_i(x).

    Here q = n_+3, where n_ is the number of intervals of the spline,
    and N_i(x) denotes the normalized B-spline of degree 3 defined upon
    the knots lambda_i,lambda_{i+1},...,lambda_{i+4}.
    The prescribed argument x must satisfy lambda_4 <= x <=
    lambda_{n_+4}.

    It is assumed that lambda_j >= lambda_{j-1}, for j = 2,3,...,n_+7,
    and lambda_{n_+4} > lambda_4.

    If x is a point at which 4 knots coincide, s(x) is discontinuous at
    x; in this case, `s` contains the value defined as x is approached
    from the right.

    The method employed is that of evaluation by taking convex
    combinations due to de Boor (1972).
    For further details of the algorithm and its use see Cox (1972) and
    Cox and Hayes (1973).

    It is expected that a common use of ``dim1_spline_eval`` will be the
    evaluation of the cubic spline approximations produced by
    ``dim1_spline_knots``.
    A generalization of ``dim1_spline_eval`` which also forms the
    derivative of s(x) is ``dim1_spline_deriv``. ``dim1_spline_deriv``
    takes about 50% longer than ``dim1_spline_eval``.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ncap7.
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[ncap7].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ncap7].
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, s.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ncap7 = utils._EngineIntScalarType(
        ncap7, {'fun_name': fun_name, 'entity': 'ncap7'},
    ).to_ctypes()
    _ncap7_py = utils._EngineIntScalarType.to_py(_ncap7)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ncap7,
        _lamda,
        _c,
        _x,
        _s,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, ncap7 = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and lamda[ncap7-4] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x <= lamda[ncap7-4].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[3] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x >= lamda[3].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncap7 >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        return_data = _s_py
    _err_class.manage()
    return return_data

def dim1_spline_deriv(ncap7, lamda, c, x, left, s):
    u"""
    Evaluation of fitted cubic spline, function and derivatives.

    ``dim1_spline_deriv`` evaluates a cubic spline and its first three
    derivatives from its B-spline representation.

    For full information please refer to the NAG Library document for
    e02bc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02bcf.html

    Parameters
    ----------
    ncap7 : int
        n_+7, where n_ is the number of intervals of the spline (which
        is one greater than the number of interior knots, i.e., the
        knots strictly within the range lambda_4 to lambda_{n_+4} over
        which the spline is defined).

    lamda : float, shape (`ncap7`)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, shape (`ncap7`)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    x : float
        The argument x at which the cubic spline and its derivatives are
        to be evaluated.

    left : int
        Specifies whether left- or right-hand values of the spline and
        its derivatives are to be computed (see [Notes]). Left - or
        right-hand values are formed according to whether `left` is
        equal or not equal to 1.

        If x does not coincide with a knot, the value of `left` is
        immaterial.

        If x = `lamda`[3], right-hand values are computed.

        If x = `lamda`[`ncap7`-4], left-hand values are formed,
        regardless of the value of `left`.

    s : float, shape (4), modified in place
        `On exit`: `s`[j-1] contains the value of the (j-1)th derivative
        of the spline at the argument x, for j = 1,2,...,4. Note that
        `s`[0] contains the value of the spline.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ncap7` = *<value>*.

            Constraint: `ncap7` >= 8.

        (`errno` 2)
            On entry, `x` = *<value>*, `ncap7` = *<value>* and
            `lamda`[`ncap7`-4] = *<value>*.

            Constraint: `x` <= `lamda`[`ncap7`-4].

        (`errno` 2)
            On entry, `x` = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `x` >= `lamda`[3].

        (`errno` 2)
            On entry, `lamda`[3] = *<value>*, `ncap7` = *<value>* and
            `lamda`[`ncap7`-4] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[`ncap7`-4].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline_deriv`` evaluates the cubic spline s(x) and its first
    three derivatives at a prescribed argument x.
    It is assumed that s(x) is represented in terms of its B-spline
    coefficients c_i, for i = 1,2,...,n_+3 and (augmented) ordered knot
    set lambda_i, for i = 1,2,...,n_+7, (see ``dim1_spline_knots``),
    i.e.,

        s(x) = sum_i = 1^qc_i N_i(x).

    Here q = n_+3, n_ is the number of intervals of the spline and
    N_i(x) denotes the normalized B-spline of degree 3 (order 4) defined
    upon the knots lambda_i,lambda_{i+1},...,lambda_{i+4}.
    The prescribed argument x must satisfy

        lambda_4 <= x <= lambda_{n_+4}.

    At a simple knot lambda_i (i.e., one satisfying lambda_{i-1} <
    lambda_i < lambda_{i+1}), the third derivative of the spline is in
    general discontinuous.
    At a multiple knot (i.e., two or more knots with the same value),
    lower derivatives, and even the spline itself, may be discontinuous.
    Specifically, at a point x = u where (exactly) r knots coincide
    (such a point is termed a knot of multiplicity r), the values of the
    derivatives of order 4-j, for j = 1,2,...,r, are in general
    discontinuous. (Here 1 <= r <= 4; r > 4 is not meaningful.) You must
    specify whether the value at such a point is required to be the
    left- or right-hand derivative.

    The method employed is based upon:

    (i) carrying out a binary search for the knot interval containing
        the argument x (see Cox (1978)),

    (#) evaluating the nonzero B-splines of orders 1, 2, 3 and 4 by
        recurrence (see Cox (1972) and Cox (1978)),

    (#) computing all derivatives of the B-splines of order 4 by
        applying a second recurrence to these computed B-spline values
        (see de Boor (1972)),

    (#) multiplying the fourth-order B-spline values and their
        derivative by the appropriate B-spline coefficients, and
        summing, to yield the values of s(x) and its derivatives.

    ``dim1_spline_deriv`` can be used to compute the values and
    derivatives of cubic spline fits and interpolants produced by
    ``dim1_spline_knots``.

    If only values and not derivatives are required,
    ``dim1_spline_eval`` may be used instead of ``dim1_spline_deriv``,
    which takes about 50% longer than ``dim1_spline_eval``.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ncap7.
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[ncap7].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ncap7].
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Input, left.
        utils._EngineFloat64ArrayType.c_type, # Output, s[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ncap7 = utils._EngineIntScalarType(
        ncap7, {'fun_name': fun_name, 'entity': 'ncap7'},
    ).to_ctypes()
    _ncap7_py = utils._EngineIntScalarType.to_py(_ncap7)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _left = utils._EngineIntScalarType(
        left, {'fun_name': fun_name, 'entity': 'left'},
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(4,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ncap7,
        _lamda,
        _c,
        _x,
        _left,
        _s,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncap7 >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, ncap7 = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and lamda[ncap7-4] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x <= lamda[ncap7-4].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[3] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x >= lamda[3].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[3] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ncap7 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[ncap7-4] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[3] < lamda[ncap7-4].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _s_cls.refresh_py(_s)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_spline_integ(ncap7, lamda, c):
    u"""
    Evaluation of fitted cubic spline, definite integral.

    ``dim1_spline_integ`` computes the definite integral of a cubic
    spline from its B-spline representation.

    For full information please refer to the NAG Library document for
    e02bd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02bdf.html

    Parameters
    ----------
    ncap7 : int
        n_+7, where n_ is the number of intervals of the spline (which
        is one greater than the number of interior knots, i.e., the
        knots strictly within the range a to b) over which the spline is
        defined.

    lamda : float, shape (`ncap7`)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, shape (`ncap7`)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    Returns
    -------
    dint : float
        The value of the definite integral of s(x) between the limits x
        = a and x = b, where a = lambda_4 and b = lambda_{n_+4}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ncap7` = *<value>*.

            Constraint: `ncap7` >= 8.

        (`errno` 2)
            On entry, J = *<value>*, `lamda`[J-1] = *<value>* and
            `lamda`[J-2] = *<value>*.

            Constraint: `lamda`[J-1] >= `lamda`[J-2].

        (`errno` 2)
            On entry, `ncap7` = *<value>*, `lamda`[`ncap7`-2] =
            *<value>* and `lamda`[`ncap7`-1] = *<value>*.

            Constraint: `lamda`[`ncap7`-2] = `lamda`[`ncap7`-1].

        (`errno` 2)
            On entry, `ncap7` = *<value>*, `lamda`[`ncap7`-3] =
            *<value>* and `lamda`[`ncap7`-2] = *<value>*.

            Constraint: `lamda`[`ncap7`-3] = `lamda`[`ncap7`-2].

        (`errno` 2)
            On entry, `ncap7` = *<value>*, `lamda`[`ncap7`-4] =
            *<value>* and `lamda`[`ncap7`-4] = *<value>*.

            Constraint: `lamda`[`ncap7`-4] = `lamda`[`ncap7`-3].

        (`errno` 2)
            On entry, `lamda`[2] = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `lamda`[2] = `lamda`[3].

        (`errno` 2)
            On entry, `lamda`[1] = *<value>* and `lamda`[2] = *<value>*.

            Constraint: `lamda`[1] = `lamda`[2].

        (`errno` 2)
            On entry, `lamda`[0] = *<value>* and `lamda`[1] = *<value>*.

            Constraint: `lamda`[0] = `lamda`[1].

        (`errno` 2)
            On entry, `lamda`[3] = *<value>*, NCAP4 = *<value>* and
            `lamda`[NCAP4-1] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[NCAP4-1].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline_integ`` computes the definite integral of the cubic
    spline s(x) between the limits x = a and x = b, where a and b are
    respectively the lower and upper limits of the range over which s(x)
    is defined.
    It is assumed that s(x) is represented in terms of its B-spline
    coefficients c_i, for i = 1,2,...,n_+3 and (augmented) ordered knot
    set lambda_i, for i = 1,2,...,n_+7, with lambda_i = a, for i =
    1,2,...,4 and lambda_i = b, for i = n_+4,...,n_+7, (see
    ``dim1_spline_knots``), i.e.,

        s(x) = sum_i = 1^qc_i N_i(x).

    Here q = n_+3, n_ is the number of intervals of the spline and
    N_i(x) denotes the normalized B-spline of degree 3 (order 4) defined
    upon the knots lambda_i,lambda_{i+1},...,lambda_{i+4}.

    The method employed uses the formula given in Section 3 of Cox
    (1975).

    ``dim1_spline_integ`` can be used to determine the definite
    integrals of cubic spline fits and interpolants produced by
    ``dim1_spline_knots``.

    References
    ----------
    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_integ'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ncap7.
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[ncap7].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ncap7].
        utils._BYREF_FLOAT64, # Output, dint.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ncap7 = utils._EngineIntScalarType(
        ncap7, {'fun_name': fun_name, 'entity': 'ncap7'},
    ).to_ctypes()
    _ncap7_py = utils._EngineIntScalarType.to_py(_ncap7)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _dint = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ncap7,
        _lamda,
        _c,
        _dint,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncap7 >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, J = {:s}, lamda[J-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and lamda[J-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[J-1] >= lamda[J-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[ncap7-2] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[ncap7-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[ncap7-2] = lamda[ncap7-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[ncap7-3] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[ncap7-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[ncap7-3] = lamda[ncap7-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[ncap7-4] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[ncap7-4] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[ncap7-4] = lamda[ncap7-3].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[2] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[3] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lamda[2] = lamda[3].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[1] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lamda[1] = lamda[2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[0] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lamda[0] = lamda[1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[3] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'NCAP4 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[NCAP4-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[3] < lamda[NCAP4-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _dint_py = utils._EngineFloat64ScalarType.to_py(_dint)
        return_data = _dint_py
    _err_class.manage()
    return return_data

def dim1_spline_auto(start, m, x, y, w, s, nest, n, lamda, c, comm):
    u"""
    Least squares cubic spline curve fit, automatic knot placement.

    ``dim1_spline_auto`` computes a cubic spline approximation to an
    arbitrary set of data points.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02be

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02bef.html

    Parameters
    ----------
    start : str, length 1
        Must be set to 'C' or 'W'.

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `n`, `lamda`, `comm`['wrk'] or `comm`['iwrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `n`, `lamda`, `comm`['wrk'], and
            `comm`['iwrk'] must be unchanged from that previous call.
            This warm start can save much time in searching for a
            satisfactory value of `s`.

    m : int
        m, the number of data points.

    x : float, shape (`m`)
        The values x_r of the independent variable (abscissa) x, for r =
        1,2,...,m.

    y : float, shape (`m`)
        The values y_r of the dependent variable (ordinate) y, for r =
        1,2,...,m.

    w : float, shape (`m`)
        The values w_r of the weights, for r = 1,2,...,m. For advice on
        the choice of weights, see [the E02 Introduction].

    s : float
        The smoothing factor, S.

        If S = 0.0, the function returns an interpolating spline.

        If S is smaller than machine precision, it is assumed equal to
        zero.

        For advice on the choice of S, see [Notes] and [Further
        Comments].

    nest : int
        An overestimate for the number, n, of knots required.

    n : int
        If the warm start option is used, the value of `n` must be left
        unchanged from the previous call.

    lamda : float, shape (`nest`), modified in place
        `On entry`: if the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[ `n` -1] must be left
        unchanged from the previous call.

        `On exit`: the knots of the spline, i.e., the positions of the
        interior knots `lamda`[4], `lamda`[5],...,`lamda`[ `n`-5] as
        well as the positions of the additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = x_1

        and

            `lamda`[ `n`-4] = `lamda`[ `n`-3] = `lamda`[ `n`-2] =
            `lamda`[ `n` -1] = x_m

        needed for the B-spline representation.

    c : float, shape (`nest`), modified in place
        `On exit`: the coefficient c_i of the B-spline N_i(x) in the
        spline approximation s(x), for i = 1,2,...,n-4.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    n : int
        The total number, n, of knots of the computed spline.

    fp : float
        The sum of the squared weighted residuals, theta, of the
        computed spline approximation. If `fp` = 0.0, this is an
        interpolating spline. `fp` should equal `s` within a relative
        tolerance of 0.001 unless n = 8 when the spline has no interior
        knots and so is simply a cubic polynomial. For knots to be
        inserted, `s` must be set to a value below the value of `fp`
        produced in this case.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, r = *<value>* and `w`[r-1] = *<value>*.

            Constraint: `w`[r-1] > 0.0 for all r.

        (`errno` 1)
            On entry, `nest` = *<value>*.

            Constraint: `nest` >= 8.

        (`errno` 1)
            On entry, `nest` = *<value>*.

            Constraint: `nest` >= *<value>* when `s` = 0.0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 4.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 2)
            On entry, r = *<value>* and `w`[r-1] = *<value>*.

            Constraint: `w`[r-1] > 0.0 for all r.

        (`errno` 3)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 4)
            The number of knots needed is greater than `nest`: `nest` =
            *<value>*. Possibly `s` is too small: `s` = *<value>*.

        (`errno` 5)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline_auto`` determines a smooth cubic spline approximation
    s(x) to the set of data points (x_r,y_r), with weights w_r, for r =
    1,2,...,m.

    The spline is given in the B-spline representation

        s(x) = sum_i = 1^n-4c_i N_i(x),

    where N_i(x) denotes the normalized cubic B-spline defined upon the
    knots lambda_i,lambda_{i+1},...,lambda_{i+4}.

    The total number n of these knots and their values
    lambda_1,...,lambda_n are chosen automatically by the function.
    The knots lambda_5,...,lambda_{n-4} are the interior knots; they
    divide the approximation interval [x_1,x_m] into n-7 sub-intervals.
    The coefficients c_1,c_2,...,c_{n-4} are then determined as the
    solution of the following constrained minimization problem:

    minimize

        eta = sum_i = 5^n-4 delta_i^2

    subject to the constraint

        theta = sum_r = 1^m epsilon_r^2 <= S,

    [table omitted]

    The quantity eta can be seen as a measure of the (lack of)
    smoothness of s(x), while closeness of fit is measured through
    theta.
    By means of the argument S, 'the smoothing factor', you can then
    control the balance between these two (usually conflicting)
    properties.
    If S is too large, the spline will be too smooth and signal will be
    lost (underfit); if S is too small, the spline will pick up too much
    noise (overfit).
    In the extreme cases the function will return an interpolating
    spline (theta = 0) if S is set to zero, and the weighted least
    squares cubic polynomial (eta = 0) if S is set very large.
    Experimenting with S values between these two extremes should result
    in a good compromise. (See [Choice of S] for advice on choice of S.)

    The method employed is outlined in [Outline of Method Used] and
    fully described in Dierckx (1975), Dierckx (1981) and Dierckx
    (1982).
    It involves an adaptive strategy for locating the knots of the cubic
    spline (depending on the function underlying the data and on the
    value of S), and an iterative method for solving the constrained
    minimization problem once the knots have been determined.

    Values of the computed spline, or of its derivatives or definite
    integral, can subsequently be computed by calling
    ``dim1_spline_eval``, ``dim1_spline_deriv`` or
    ``dim1_spline_integ``, as described in [Evaluation of Computed
    Spline].

    References
    ----------
    Dierckx, P, 1975, `An algorithm for smoothing, differentiating and
    integration of experimental data using spline functions`, J. Comput.
    Appl. Math. (1), 165--184

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_auto'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, start(len=1).
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_INT, # Input, nest.
        utils._BYREF_INT, # Input/Output, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, lamda[nest].
        utils._EngineFloat64ArrayType.c_type, # Output, c[nest].
        utils._BYREF_FLOAT64, # Output, fp.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, wrk[lwrk].
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Input/Output, iwrk[nest].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of start.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _start = utils._EngineCharScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _nest = utils._EngineIntScalarType(
        nest, {'fun_name': fun_name, 'entity': 'nest'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nest_py = utils._EngineIntScalarType.to_py(_nest)
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_nest_py,),
    )
    _lamda = _lamda_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_nest_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _fp = utils._EngineFloat64ScalarType.empty_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('wrk', 'iwrk'),
    )
    if _comm_needs_init:
        _wrk = utils._EngineFloat64ArrayType.empty_ctypes(4*_m_py+16*_nest_py+41)
        comm['wrk'] = _wrk
    else:
        _wrk = comm['wrk']
    _lwrk_py = utils._EngineFloat64ArrayType.get_shape(comm['wrk'], {'fun_name': fun_name, 'entity': 'comm[\'wrk\']'}, exp_rank=1, validate=False)[0]
    _lwrk = utils._EngineIntScalarType(
        _lwrk_py, {'fun_name': fun_name, 'entity': '_lwrk'},
    ).to_ctypes()
    if _comm_needs_init:
        _iwrk = utils._EngineIntArrayType.empty_ctypes(_nest_py)
        comm['iwrk'] = _iwrk
    else:
        _iwrk = comm['iwrk']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _start,
        _m,
        _x,
        _y,
        _w,
        _s,
        _nest,
        _n,
        _lamda,
        _c,
        _fp,
        _wrk,
        _lwrk,
        _iwrk,
        _errbuf,
        _ierr,
        _start._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s} and w[r-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w[r-1] > 0.0 for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nest >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nest >= {:s} when s = 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start != \'W\' or \'C\': start = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s} and w[r-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w[r-1] > 0.0 for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, x[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[r-2] < x[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of knots needed is greater than nest:',
                    'nest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative process has failed to converge.',
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _lamda_cls.refresh_py(_lamda)
        _c_cls.refresh_py(_c)
        _fp_py = utils._EngineFloat64ScalarType.to_py(_fp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['n', 'fp'],
        )(
            _n_py,
            _fp_py,
        )
    _err_class.manage()
    return return_data

def dim1_spline_deriv_vector(start, ncap7, lamda, c, deriv, xord, x, ixloc, nx, s, iwrk, liwrk):
    u"""
    Evaluation of fitted cubic spline, function and optionally
    derivatives at a vector of points.

    ``dim1_spline_deriv_vector`` evaluates a cubic spline and up to its
    first three derivatives from its B-spline representation at a vector
    of points. ``dim1_spline_deriv_vector`` can be used to compute the
    values and derivatives of cubic spline fits and interpolants
    produced by reference to ``interp.dim1_spline``,
    ``dim1_spline_knots`` and ``dim1_spline_auto``.

    For full information please refer to the NAG Library document for
    e02bf

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02bff.html

    Parameters
    ----------
    start : int
        Indicates the completion state of the first phase of the
        algorithm.

        `start` = 0

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have not been determined, and you wish to
            use the sorted mode of vectorization.

        `start` = 1

            The enclosing interval numbers ixloc_j have been determined
            and are provided in `ixloc`, however the required
            permutation and interval related information has not been
            determined and you wish to use the sorted mode of
            vectorization.

        `start` = 2

            You wish to use the sorted mode of vectorization, and the
            entire first phase has been completed, with the enclosing
            interval numbers supplied in `ixloc`, and the required
            permutation and interval related information provided in
            `iwrk` (from a previous call to
            ``dim1_spline_deriv_vector``).

        `start` = 10

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have not been determined, and you wish to
            use the unsorted mode of vectorization.

        `start` = 11

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have been supplied in `ixloc`, and you wish
            to use the unsorted mode of vectorization.

        `Additional`: `start` = 0 or 10 should be used unless you are
        sure that the knot set is unchanged between calls.

    ncap7 : int
        n_+7, where n_ is the number of intervals of the spline (which
        is one greater than the number of interior knots, i.e., the
        knots strictly within the range lambda_4 to lambda_{n_+4} over
        which the spline is defined). Note that if ``dim1_spline_auto``
        was used to generate the knots and spline coefficients then
        `ncap7` should contain the same value as returned in n by
        ``dim1_spline_auto``.

    lamda : float, shape (`ncap7`)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, shape (`ncap7`)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    deriv : int
        Determines the maximum order of derivatives required, D_ord.

        If `deriv` < 0 left derivatives are calculated, otherwise right
        derivatives are calculated. For abscissae satisfying x_j =
        lambda_4 or x_j = lambda_{n_+4} only right-handed or left-handed
        computation will be used respectively.

        For abscissae which do not coincide exactly with a knot, the
        handedness of the computation is immaterial.

        `deriv` = 0

            No derivatives required.

        `deriv` = +/-1

            Only s(x) and its first derivative are required.

        `deriv` = +/-2

            Only s(x) and its first and second derivatives are required.

        `deriv` = +/-3

            s(x) and its first, second and third derivatives are
            required.

        Note: if |`deriv`| is greater than 3 only the derivatives up to
        and including 3 will be returned.

    xord : int
        Indicates whether `x` is supplied in a sufficiently ordered
        manner. If `x` is sufficiently ordered
        ``dim1_spline_deriv_vector`` will complete faster.

        `xord` != 0

            The abscissae in `x` are ordered at least by ascending
            interval, in that any two abscissae contained in the same
            interval are only separated by abscissae in the same
            interval, and the intervals are arranged in ascending order.
            For example, x_j < x_{j+1}, for j = 1,2,...,`nx`-1.

        `xord` = 0

            The abscissae in `x` are not sufficiently ordered.

    x : float, shape (`nx`)
        The abscissae x_j, for j = 1,2,...,n_x. If `start` = 0 or 10
        then evaluations will only be performed for these x_j satisfying
        lambda_4 <= x_j <= lambda_{n_+4}. Otherwise evaluation will be
        performed unless the corresponding element of `ixloc` contains
        an invalid interval number. Please note that if the `ixloc`[j]
        is a valid interval number then no check is made that `x`[j]
        actually lies in that interval.

    ixloc : int, shape (`nx`), modified in place
        `On entry`: if `start` = 1, 2 or 11, if you wish x_j to be
        evaluated, `ixloc`[j-1] must be the enclosing interval number
        ixloc_j of the abscissae x_j (see [equation]). If you do not
        wish x_j to be evaluated, you may set the interval number to be
        either less than 4 or greater than n_+4.

        Otherwise, `ixloc` need not be set.

        `On exit`: if `start` = 1, 2 or 11, `ixloc` is unchanged on
        exit.

        Otherwise, `ixloc`[j-1], contains the enclosing interval number
        ixloc_j, for the abscissa supplied in `x`[j-1], for j =
        1,2,...,n_x.

        Evaluations will only be performed for abscissae x_j satisfying
        lambda_4 <= x_j <= lambda_{n_+4}.

        If evaluation is not performed `ixloc`[j-1] is set to 0 if x_j <
        lambda_4 or n_+7 if x_j > lambda_{n_+4}.

    nx : int
        n_x, the total number of abscissae contained in `x`, including
        any that will not be evaluated.

    s : float, shape (`nx`, min(|`deriv`|, 3)+1), modified in place
        `On exit`: if x_j is valid, `s`[j-1,d-1] will contain the
        (d-1)th derivative of s(x), for j = 1,2,...,n_x, for d =
        1,2,...,D_ord+1. In particular, `s`[j-1,0] will contain the
        approximation of s(x_j) for all legal values in `x`.

    iwrk : None or int, shape (`liwrk`), modified in place
        `Optionally, on entry`: if `start` = 2, `iwrk` must be unchanged
        from a previous call to ``dim1_spline_deriv_vector`` with
        `start` = 0 or 1.

        Otherwise, `iwrk` need not be set.

        `On exit`, if not None on entry: if `start` = 10 or 11, `iwrk`
        is unchanged on exit.

        Otherwise, `iwrk` contains the required permutation of elements
        of `x`, if any, and information related to the division of the
        abscissae x_j between the intervals derived from `lamda`.

    liwrk : int
        The dimension of the array `iwrk`.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, all elements of `x` had enclosing interval numbers
            in `ixloc` outside the domain allowed by the provided
            spline.

            *<value>* entries of `x` were indexed below the lower bound
            *<value>*.

            *<value>* entries of `x` were indexed above the upper bound
            *<value>*.

        (`errno` 11)
            On entry, `start` = *<value>*.

            Constraint: `start` = 0, 1, 2, 10 or 11.

        (`errno` 12)
            On entry, `start` = 2 and `nx` is not consistent with the
            previous call to ``dim1_spline_deriv_vector``.

            On entry, `nx` = *<value>*.

            Constraint: `nx` = *<value>*.

        (`errno` 21)
            On entry, `ncap7` = *<value>*.

            Constraint: `ncap7` >= 8.

        (`errno` 31)
            On entry, `lamda`[3] = *<value>*, `ncap7` = *<value>* and
            `lamda`[`ncap7`-4] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[`ncap7`-4].

        (`errno` 91)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 131)
            On entry, `liwrk` = *<value>*.

            Constraint: `liwrk` >= 3*`nx`+3 = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one element of `x` has an enclosing
            interval number in `ixloc` outside the set allowed by the
            provided spline. The spline has been evaluated for all `x`
            with enclosing interval numbers inside the allowable set.

            *<value>* entries of `x` were indexed below the lower bound
            *<value>*.

            *<value>* entries of `x` were indexed above the upper bound
            *<value>*.

    Notes
    -----
    ``dim1_spline_deriv_vector`` evaluates the cubic spline s(x) and
    optionally derivatives up to order 3 for a vector of points x_j, for
    j = 1,2,...,n_x.
    It is assumed that s(x) is represented in terms of its B-spline
    coefficients c_i, for i = 1,2,...,n_+3, and (augmented) ordered knot
    set lambda_i, for i = 1,2,...,n_+7, (see ``dim1_spline_knots`` and
    ``dim1_spline_auto``), i.e.,

        s(x) = sum_i = 1^qc_i N_i(x).

    Here q = n_+3, n_ is the number of intervals of the spline and
    N_i(x) denotes the normalized B-spline of degree 3 (order 4) defined
    upon the knots lambda_i,lambda_{i+1},...,lambda_{i+4}.
    The knots lambda_5,lambda_6,...,lambda_{n_+3} are the interior
    knots.
    The remaining knots, lambda_1, lambda_2, lambda_3, lambda_4 and
    lambda_{n_+4}, lambda_{n_+5}, lambda_{n_+6}, lambda_n+7_ are the
    exterior knots.
    The knots lambda_4 and lambda_{n_+4} are the boundaries of the
    spline.

    Only abscissae satisfying,

        lambda_4 <= x_j <= lambda_{n_+4},

    will be evaluated.
    At a simple knot lambda_i (i.e., one satisfying lambda_{i-1} <
    lambda_i < lambda_{i+1}), the third derivative of the spline is, in
    general, discontinuous.
    At a multiple knot (i.e., two or more knots with the same value),
    lower derivatives, and even the spline itself, may be discontinuous.
    Specifically, at a point x = u where (exactly) r knots coincide
    (such a point is termed a knot of multiplicity r), the values of the
    derivatives of order 4-j, for j = 1,2,...,r, are, in general,
    discontinuous. (Here 1 <= r <= 4; r > 4 is not meaningful.) The
    maximum order of the derivatives to be evaluated D_ord, and the
    left- or right-handedness of the computation when an abscissa
    corresponds exactly to an interior knot, are determined by the value
    of `deriv`.

    Each abscissa (point at which the spline is to be evaluated) x_j
    contained in `x` has an associated enclosing interval number,
    ixloc_j either supplied or returned in `ixloc` (see argument
    `start`).
    A simple call to ``dim1_spline_deriv_vector`` would set `start` = 0
    and the contents of `ixloc` need never be set nor referenced, and
    the following description on modes of operation can be ignored.
    However, where efficiency is an important consideration, the
    following description will help to choose the appropriate mode of
    operation.

    The interval numbers are used to determine which B-splines must be
    evaluated for a given abscissa, and are defined as

        ixloc_j = [table omitted]

    The algorithm has two modes of vectorization, termed here sorted and
    unsorted, which are selectable by the argument `start`.

    Furthermore, if the supplied abscissae are sufficiently ordered, as
    indicated by the argument `xord`, the algorithm will take advantage
    of significantly faster methods for the determination of both the
    interval numbers and the subsequent spline evaluations.

    The sorted mode has two phases, a sorting phase and an evaluation
    phase.
    This mode is recommended if there are many abscissae to evaluate
    relative to the number of intervals of the spline, or the abscissae
    are distributed relatively densely over a subsection of the spline.
    In the first phase, ixloc_j is determined for each x_j and a
    permutation is calculated to sort the x_j by interval number.
    The first phase may be either partially or completely by-passed
    using the argument `start` if the enclosing segments and/or the
    subsequent ordering are already known `a priori`, for example if
    multiple spline coefficients `c` are to be evaluated over the same
    set of knots `lamda`.

    In the second phase of the sorted mode, spline approximations are
    evaluated by segment, so that non-abscissa dependent calculations
    over a segment may be reused in the evaluation for all abscissae
    belonging to a specific segment.
    For example, all third derivatives of all abscissae in the same
    segment will be identical.

    In the unsorted mode of vectorization, no `a priori` segment sorting
    is performed, and if the abscissae are not sufficiently ordered, the
    evaluation at an abscissa will be independent of evaluations at
    other abscissae; also non-abscissa dependent calculations over a
    segment will be repeated for each abscissa in a segment.
    This may be quicker if the number of abscissa is small in comparison
    to the number of knots in the spline, and they are distributed
    sparsely throughout the domain of the spline.
    This is effectively a direct vectorization of ``dim1_spline_eval``
    and ``dim1_spline_deriv``, although if the enclosing interval
    numbers ixloc_j are known, these may again be provided.

    If the abscissae are sufficiently ordered, then once the first
    abscissa in a segment is known, an efficient algorithm will be used
    to determine the location of the final abscissa in this segment.
    The spline will subsequently be evaluated in a vectorized manner for
    all the abscissae indexed between the first and last of the current
    segment.

    If no derivatives are required, the spline evaluation is calculated
    by taking convex combinations due to de Boor (1972).
    Otherwise, the calculation of s(x) and its derivatives is based
    upon,

    (i) evaluating the nonzero B-splines of orders 1, 2, 3 and 4 by
        recurrence (see Cox (1972) and Cox (1978)),

    (#) computing all derivatives of the B-splines of order 4 by
        applying a second recurrence to these computed B-spline values
        (see de Boor (1972)),

    (#) multiplying the fourth-order B-spline values and their
        derivative by the appropriate B-spline coefficients, and
        summing, to yield the values of s(x) and its derivatives.

    The method of convex combinations is significantly faster than the
    recurrence based method.
    If higher derivatives of order 2 or 3 are not required, as much
    computation as possible is avoided.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim1_spline_deriv_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, start.
        utils._BYREF_INT, # Input, ncap7.
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[ncap7].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ncap7].
        utils._BYREF_INT, # Input, deriv.
        utils._BYREF_INT, # Input, xord.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._EngineIntArrayType.c_type, # Input/Output, ixloc[nx].
        utils._BYREF_INT, # Input, nx.
        utils._EngineFloat64ArrayType.c_type, # Output, s[lds, min(abs(deriv), 3)+1].
        utils._BYREF_INT, # Input, lds.
        utils._EngineIntArrayType.c_type, # Input/Output, iwrk[liwrk].
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((s, 's'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _start = utils._EngineIntScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'},
    ).to_ctypes()
    _ncap7 = utils._EngineIntScalarType(
        ncap7, {'fun_name': fun_name, 'entity': 'ncap7'},
    ).to_ctypes()
    _ncap7_py = utils._EngineIntScalarType.to_py(_ncap7)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncap7_py,),
    ).to_ctypes()
    _deriv = utils._EngineIntScalarType(
        deriv, {'fun_name': fun_name, 'entity': 'deriv'},
    ).to_ctypes()
    _xord = utils._EngineIntScalarType(
        xord, {'fun_name': fun_name, 'entity': 'xord'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _ixloc_cls = utils._EngineIntArrayType(
        ixloc, {'fun_name': fun_name, 'entity': 'ixloc'},
        exp_shape=(_nx_py,),
    )
    _ixloc = _ixloc_cls.to_ctypes()
    _lds_py = utils._EngineFloat64ArrayType.get_nd(s, {'fun_name': fun_name, 'entity': 's'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lds = utils._EngineIntScalarType(
        _lds_py, {'fun_name': fun_name, 'entity': '_lds'},
    ).to_ctypes()
    _deriv_py = utils._EngineIntScalarType.to_py(_deriv)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_nx_py, min(abs(_deriv_py), 3)+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    _liwrk = utils._EngineIntScalarType(
        liwrk, {'fun_name': fun_name, 'entity': 'liwrk'},
    ).to_ctypes()
    _start_py = utils._EngineIntScalarType.to_py(_start)
    if _start_py == 2:
        if iwrk is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'iwrk'},
                'must not be None when start = 2',
            )
    if _start_py not in (10, 11):
        if iwrk is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'iwrk'},
                'must not be None when start not in (10, 11)',
            )
    _liwrk_py = utils._EngineIntScalarType.to_py(_liwrk)
    _iwrk_cls = utils._EngineIntArrayType(
        iwrk, {'fun_name': fun_name, 'entity': 'iwrk'},
        exp_shape=(_liwrk_py,),
        check_contents=_start_py == 2,
        allow_none=True,
    )
    _iwrk = _iwrk_cls.to_ctypes(can_reference_contents=_start_py == 2)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _start,
        _ncap7,
        _lamda,
        _c,
        _deriv,
        _xord,
        _x,
        _ixloc,
        _nx,
        _s,
        _lds,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_e_consts.E02BF_IERR_NTINSP and _ifmt == _nag_e_consts.E02BF_IFMT_NTINSP:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of x has an enclosing interval number in',
                    'ixloc outside the set allowed by the provided spline. The spline has been',
                    'evaluated for all x with enclosing interval numbers inside the allowable',
                    'set.',
                    '{:s} entries of x were indexed below the lower bound'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    '{:s} entries of x were indexed above the upper bound'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_NOVALID and _ifmt == _nag_e_consts.E02BF_IFMT_NOVALID:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all elements of x had enclosing interval numbers in ixloc outside',
                    'the domain allowed by the provided spline.',
                    '{:s} entries of x were indexed below the lower bound'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    '{:s} entries of x were indexed above the upper bound'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_START and _ifmt == _nag_e_consts.E02BF_IFMT_START:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: start = 0, 1, 2, 10 or 11.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_START_NX and _ifmt == _nag_e_consts.E02BF_IFMT_START_NX:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start = 2 and nx is not consistent with',
                    'the previous call to dim1_spline_deriv_vector.',
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_NCAP7 and _ifmt == _nag_e_consts.E02BF_IFMT_NCAP7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncap7 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncap7 >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_LAMDA and _ifmt == _nag_e_consts.E02BF_IFMT_LAMDA:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[3] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ncap7 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'lamda[ncap7-4] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[3] < lamda[ncap7-4].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_NX and _ifmt == _nag_e_consts.E02BF_IFMT_NX:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.E02BF_IERR_LIWRK and _ifmt == _nag_e_consts.E02BF_IFMT_LIWRK:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liwrk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: liwrk >= 3*nx+3 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ixloc_cls.refresh_py(_ixloc)
        _s_cls.refresh_py(_s)
        if iwrk is not None:
            _iwrk_cls.refresh_py(_iwrk)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim2_cheb_lines(m, n, k, l, x, y, f, w, mtot, a, na, xmin, xmax, nux, inuxp1, nuy, inuyp1):
    u"""
    Least squares surface fit by polynomials, data on lines parallel to
    one independent coordinate axis.

    ``dim2_cheb_lines`` forms an approximation to the weighted, least
    squares Chebyshev series surface fit to data arbitrarily distributed
    on lines parallel to one independent coordinate axis.

    For full information please refer to the NAG Library document for
    e02ca

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02caf.html

    Parameters
    ----------
    m : int, shape (`n`)
        `m`[s-1] must be set to m_s, the number of data x values on the
        line y = y_s, for s = 1,2,...,n.

    n : int
        The number of lines y = constant on which data points are given.

    k : int
        k, the required degree of x in the fit.

    l : int
        l, the required degree of y in the fit.

    x : float, shape (`mtot`)
        The x values of the data points. The sequence must be

            all points on y = y_1, followed by

            all points on y = y_2, followed by

            ...

            all points on y = y_n.

    y : float, shape (`n`)
        `y`[s-1] must contain the y value of line y = y_s, for s =
        1,2,...,n, on which data is given.

    f : float, shape (`mtot`)
        f, the data values of the dependent variable in the same
        sequence as the x values.

    w : float, shape (`mtot`)
        The weights to be assigned to the data points, in the same
        sequence as the x values. These weights should be calculated
        from estimates of the absolute accuracies of the f_r, expressed
        as standard deviations, probable errors or some other measure
        which is of the same dimensions as f_r. Specifically, each w_r
        should be inversely proportional to the accuracy estimate of
        f_r. Often weights all equal to unity will be satisfactory. If a
        particular weight is zero, the corresponding data point is
        omitted from the fit.

    mtot : int
        The dimension of the arrays `x`, `f` and `w`.

    a : float, shape (`na`), modified in place
        `On exit`: contains the Chebyshev coefficients of the fit.
        `a`[i*(`l`+1)+j-1] is the coefficient a_{ij} of [Notes] defined
        according to the standard convention. These coefficients are
        used by ``dim2_cheb_eval`` to calculate values of the fitted
        function.

    na : int
        The dimension of the array `a`.

    xmin : float, shape (`n`)
        `xmin`[s-1] must contain x_min^(s), the lower end of the range
        of x on the line y = y_s, for s = 1,2,...,n. It must not be
        greater than the lowest data value of x on the line. Each
        x_min^(s) is scaled to -1.0 in the fit. (See also [Further
        Comments].)

    xmax : float, shape (`n`)
        `xmax`[s-1] must contain x_max^(s), the upper end of the range
        of x on the line y = y_s, for s = 1,2,...,n. It must not be less
        than the highest data value of x on the line. Each x_max^(s) is
        scaled to +1.0 in the fit. (See also [Further Comments].)

    nux : float, shape (`inuxp1`)
        `nux`[i-1] must contain the coefficient of the Chebyshev
        polynomial of degree (i-1) in x_, in the Chebyshev series
        representation of the polynomial factor in x_ which you require
        the fit to contain, for i = 1,2,...,`inuxp1`. These coefficients
        are defined according to the standard convention of [Notes].

    inuxp1 : int
        INUX+1, where INUX is the degree of a polynomial factor in x_
        which you require the fit to contain. (See [Notes], last
        paragraph.)

        If this option is not required, `inuxp1` should be set equal to
        1.

    nuy : float, shape (`inuyp1`)
        `nuy`[i-1] must contain the coefficient of the Chebyshev
        polynomial of degree (i-1) in y_, in the Chebyshev series
        representation of the polynomial factor which you require the
        fit to contain, for i = 1,2,...,`inuyp1`. These coefficients are
        defined according to the standard convention of [Notes].

    inuyp1 : int
        INUY+1, where INUY is the degree of a polynomial factor in y_
        which you require the fit to contain. (See [Notes], last
        paragraph.) If this option is not required, `inuyp1` should be
        set equal to 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `na` is too small. `na` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, `mtot` is too small. `mtot` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, I = *<value>*, `m`[I-1] = *<value>*, `k` =
            *<value>* and `inuxp1` = *<value>*.

            Constraint: `m`[I-1] >= `k`-`inuxp1`+2.

        (`errno` 1)
            On entry, `n` = *<value>*, `l` = *<value>* and `inuyp1` =
            *<value>*.

            Constraint: `n` >= `l`-`inuyp1`+2.

        (`errno` 1)
            On entry, `inuyp1` = *<value>* and `l` = *<value>*.

            Constraint: `inuyp1` <= `l`+1.

        (`errno` 1)
            On entry, `inuxp1` = *<value>* and `k` = *<value>*.

            Constraint: `inuxp1` <= `k`+1.

        (`errno` 1)
            On entry, `inuyp1` = *<value>*.

            Constraint: `inuyp1` >= 1.

        (`errno` 1)
            On entry, `inuxp1` = *<value>*.

            Constraint: `inuxp1` >= 1.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            On entry, `xmin`[I-1] and `xmax`[I-1] do not span the data
            `x` values on `y` = `y`[I-1]: I = *<value>*, `xmin`[I-1] =
            *<value>*, `xmax`[I-1] = *<value>* and `y`[I-1] = *<value>*.

        (`errno` 3)
            On entry, the data `x` values are not nondecreasing for `y`
            = `y`[I-1]: I = *<value>* and `y`[I-1] = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `y`[I-1] = *<value>* and `y`[I-2] =
            *<value>*.

            Constraint: `y`[I-1] > `y`[I-2].

        (`errno` 4)
            On entry, the number of distinct `x` values with nonzero
            weight on `y` = `y`[I-1] is less than `k`-`inuxp1`+2: I =
            *<value>*, `y`[I-1] = *<value>*, `k` = *<value>* and
            `inuxp1` = *<value>*.

        (`errno` 5)
            On entry, `inuxp1` = *<value>*, `nux`[`inuxp1`-1] =
            *<value>*, `inuyp1` = *<value>* and `nuy`[`inuyp1`-1] =
            *<value>*.

            Constraint: if `nux`[`inuxp1`-1] = 0.0, `inuxp1` = 1; if
            `nuy`[`inuyp1`-1] = 0.0, `inuyp1` = 1.

    Notes
    -----
    ``dim2_cheb_lines`` determines a bivariate polynomial approximation
    of degree k in x and l in y to the set of data points
    (x_{r,s},y_s,f_{r,s}), with weights w_{r,s}, for r = 1,2,...,m_s,
    for s = 1,2,...,n.
    That is, the data points are on lines y = y_s, but the x values may
    be different on each line.
    The values of k and l are prescribed by you (for guidance on their
    choice, see [Further Comments]).
    The function is based on the method described in Sections 5 and 6 of
    Clenshaw and Hayes (1965).

    The polynomial is represented in double Chebyshev series form with
    arguments x_ and y_.
    The arguments lie in the range -1 to +1 and are related to the
    original variables x and y by the transformations

        x_= (2x-(x_max+x_min))/((x_max-x_min)) and y_=
        (2y-(y_max+y_min))/((y_max-y_min)).

    Here y_max and y_min are set by the function to, respectively, the
    largest and smallest value of y_s, but x_max and x_min are functions
    of y prescribed by you (see [Further Comments]).
    For this function, only their values x_max^(s) and x_min^(s) at each
    y = y_s are required.
    For each s = 1,2,...,n, x_max^(s) must not be less than the largest
    x_{r,s} on the line y = y_s, and, similarly, x_min^(s) must not be
    greater than the smallest x_{r,s}.

    The double Chebyshev series can be written as

        sum_i = 0^ksum_j = 0^la_{ij}T_i(x_)T_j(y_)

    where T_i(x_) is the Chebyshev polynomial of the first kind of
    degree i with argument x_, and T_j(y) is similarly defined.
    However, the standard convention, followed in this function, is that
    coefficients in the above expression which have either i or j zero
    are written as 1/2a_{ij}, instead of simply a_{ij}, and the
    coefficient with both i and j equal to zero is written as
    1/4a_{0,0}.
    The series with coefficients output by the function should be summed
    using this convention. ``dim2_cheb_eval`` is available to compute
    values of the fitted function from these coefficients.

    The function first obtains Chebyshev series coefficients c_{s,i},
    for i = 0,1,...,k, of the weighted least squares polynomial curve
    fit of degree k in x_ to the data on each line y = y_s, for s =
    1,2,...,n, in turn, using an auxiliary function.
    The same function is then called k+1 times to fit c_{s,i}, for s =
    1,2,...,n, by a polynomial of degree l in y_, for each i =
    0,1,...,k.
    The resulting coefficients are the required a_{ij}.

    You can force the fit to contain a given polynomial factor.
    This allows for the surface fit to be constrained to have specified
    values and derivatives along the boundaries x = x_min, x = x_max, y
    = y_min and y = y_max or indeed along any lines x_= constant or y_=
    constant (see Section 8 of Clenshaw and Hayes (1965)).

    References
    ----------
    Clenshaw, C W and Hayes, J G, 1965, `Curve and surface fitting`, J.
    Inst. Math. Appl. (1), 164--183

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_cheb_lines'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineIntArrayType.c_type, # Input, m[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, l.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mtot].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[mtot].
        utils._EngineFloat64ArrayType.c_type, # Input, w[mtot].
        utils._BYREF_INT, # Input, mtot.
        utils._EngineFloat64ArrayType.c_type, # Output, a[na].
        utils._BYREF_INT, # Input, na.
        utils._EngineFloat64ArrayType.c_type, # Input, xmin[n].
        utils._EngineFloat64ArrayType.c_type, # Input, xmax[n].
        utils._EngineFloat64ArrayType.c_type, # Input, nux[inuxp1].
        utils._BYREF_INT, # Input, inuxp1.
        utils._EngineFloat64ArrayType.c_type, # Input, nuy[inuyp1].
        utils._BYREF_INT, # Input, inuyp1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, nwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_cls = utils._EngineIntArrayType(
        m, {'fun_name': fun_name, 'entity': 'm'},
        exp_shape=(_n_py,),
    )
    _m = _m_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _mtot = utils._EngineIntScalarType(
        mtot, {'fun_name': fun_name, 'entity': 'mtot'},
    ).to_ctypes()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_cls.get_sum(0, _n_py-1),),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_cls.get_sum(0, _n_py-1),),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_cls.get_sum(0, _n_py-1),),
    ).to_ctypes()
    _na = utils._EngineIntScalarType(
        na, {'fun_name': fun_name, 'entity': 'na'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=((_k_py+1)*(_l_py+1),),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _xmin = utils._EngineFloat64ArrayType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xmax = utils._EngineFloat64ArrayType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _inuxp1 = utils._EngineIntScalarType(
        inuxp1, {'fun_name': fun_name, 'entity': 'inuxp1'},
    ).to_ctypes()
    _inuxp1_py = utils._EngineIntScalarType.to_py(_inuxp1)
    _nux = utils._EngineFloat64ArrayType(
        nux, {'fun_name': fun_name, 'entity': 'nux'},
        exp_shape=(_inuxp1_py,),
    ).to_ctypes()
    _inuyp1 = utils._EngineIntScalarType(
        inuyp1, {'fun_name': fun_name, 'entity': 'inuyp1'},
    ).to_ctypes()
    _inuyp1_py = utils._EngineIntScalarType.to_py(_inuyp1)
    _nuy = utils._EngineFloat64ArrayType(
        nuy, {'fun_name': fun_name, 'entity': 'nuy'},
        exp_shape=(_inuyp1_py,),
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _m,
        _n,
        _k,
        _l,
        _x,
        _y,
        _f,
        _w,
        _mtot,
        _a,
        _na,
        _xmin,
        _xmax,
        _nux,
        _inuxp1,
        _nuy,
        _inuyp1,
        _work,
        _nwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, na is too small.',
                    'na = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Minimum possible dimension: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtot is too small.',
                    'mtot = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Minimum possible dimension: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, m[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'k = {:s} and inuxp1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: m[I-1] >= k-inuxp1+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, l = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'inuyp1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: n >= l-inuyp1+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inuyp1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: inuyp1 <= l+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inuxp1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: inuxp1 <= k+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inuyp1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inuyp1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inuxp1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inuxp1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin[I-1] and xmax[I-1] do not span the data x',
                    'values on y = y[I-1]:',
                    'I = {:s}, xmin[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmax[I-1] = {:s} and y[I-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the data x values are not nondecreasing for',
                    'y = y[I-1]:',
                    'I = {:s} and y[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, y[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'y[I-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: y[I-1] > y[I-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the number of distinct x values with nonzero weight',
                    'on y = y[I-1] is less than k-inuxp1+2:',
                    'I = {:s}, y[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'k = {:s} and inuxp1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inuxp1 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nux[inuxp1-1] = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'inuyp1 = {:s} and nuy[inuyp1-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if nux[inuxp1-1] = 0.0, inuxp1 = 1;',
                    'if nuy[inuyp1-1] = 0.0, inuyp1 = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_cheb_eval(mfirst, mlast, k, l, x, xmin, xmax, y, ymin, ymax, ff, a, na):
    u"""
    Evaluation of fitted polynomial in two variables.

    ``dim2_cheb_eval`` evaluates a bivariate polynomial from the
    rectangular array of coefficients in its double Chebyshev series
    representation.

    For full information please refer to the NAG Library document for
    e02cb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02cbf.html

    Parameters
    ----------
    mfirst : int
        The index of the first and last x value in the array x at which
        the evaluation is required respectively (see [Further
        Comments]).

    mlast : int
        The index of the first and last x value in the array x at which
        the evaluation is required respectively (see [Further
        Comments]).

    k : int
        The degree k of x and l of y, respectively, in the polynomial.

    l : int
        The degree k of x and l of y, respectively, in the polynomial.

    x : float, shape (`mlast`)
        `x`[i-1], for i = `mfirst`,...,`mlast`, must contain the x
        values at which the evaluation is required.

    xmin : float
        The lower and upper ends, x_min and x_max, of the range of the
        variable x (see [Notes]).

        The values of `xmin` and `xmax` may depend on the value of y
        (e.g., when the polynomial has been derived using
        ``dim2_cheb_lines``).

    xmax : float
        The lower and upper ends, x_min and x_max, of the range of the
        variable x (see [Notes]).

        The values of `xmin` and `xmax` may depend on the value of y
        (e.g., when the polynomial has been derived using
        ``dim2_cheb_lines``).

    y : float
        The value of the y coordinate of all the points at which the
        evaluation is required.

    ymin : float
        The lower and upper ends, y_min and y_max, of the range of the
        variable y (see [Notes]).

    ymax : float
        The lower and upper ends, y_min and y_max, of the range of the
        variable y (see [Notes]).

    ff : float, shape (`mlast`), modified in place
        `On exit`: `ff`[i-1] gives the value of the polynomial at the
        point (x_i,y), for i = `mfirst`,...,`mlast`.

    a : float, shape (`na`)
        The Chebyshev coefficients of the polynomial. The coefficient
        a_{ij} defined according to the standard convention (see
        [Notes]) must be in `a`[i*(l+1)+j].

    na : int
        The dimension of the array `a`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `na` = *<value>*, `k` = *<value>* and `l` =
            *<value>*.

            Constraint: `na` >= (`k`+1)*(`l`+1).

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 1)
            On entry, `mfirst` = *<value>* and `mlast` = *<value>*.

            Constraint: `mfirst` <= `mlast`.

        (`errno` 2)
            On entry, `y` = *<value>* and `ymax` = *<value>*.

            Constraint: `y` <= `ymax`.

        (`errno` 2)
            On entry, `y` = *<value>* and `ymin` = *<value>*.

            Constraint: `y` >= `ymin`.

        (`errno` 2)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `xmax` =
            *<value>*.

            Constraint: `x`[I-1] <= `xmax`.

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `xmin` =
            *<value>*.

            Constraint: `x`[I-1] >= `xmin`.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            Unexpected internal failure when evaluating the polynomial.

    Notes
    -----
    This function evaluates a bivariate polynomial (represented in
    double Chebyshev form) of degree k in one variable, x_, and degree l
    in the other, y_.
    The range of both variables is -1 to +1.
    However, these normalized variables will usually have been derived
    (as when the polynomial has been computed by ``dim2_cheb_lines``,
    for example) from your original variables x and y by the
    transformations

        x_= (2x-(x_max+x_min))/((x_max-x_min)) and y_=
        (2y-(y_max+y_min))/((y_max-y_min)).

    (Here x_min and x_max are the ends of the range of x which has been
    transformed to the range -1 to +1 of x_. y_min and y_max are
    correspondingly for y.
    See [Further Comments]).
    For this reason, the function has been designed to accept values of
    x and y rather than x_ and y_, and so requires values of x_min, etc.
    to be supplied by you.
    In fact, for the sake of efficiency in appropriate cases, the
    function evaluates the polynomial for a sequence of values of x, all
    associated with the same value of y.

    The double Chebyshev series can be written as

        sum_i = 0^ksum_j = 0^la_{ij}T_i(x_)T_j(y_),

    where T_i(x_) is the Chebyshev polynomial of the first kind of
    degree i and argument x_, and T_j(y_) is similarly defined.
    However the standard convention, followed in this function, is that
    coefficients in the above expression which have either i or j zero
    are written 1/2a_{ij}, instead of simply a_{ij}, and the coefficient
    with both i and j zero is written 1/4a_{0,0}.

    The function first forms c_i = sum_j = 0^la_{ij}T_j(y_), with
    a_{i,0} replaced by 1/2a_{i,0}, for each of i = 0,1,...,k.
    The value of the double series is then obtained for each value of x,
    by summing c_i*T_i(x_), with c_0 replaced by 1/2c_0, over i =
    0,1,...,k.
    The Clenshaw three term recurrence (see Clenshaw (1955)) with
    modifications due to Reinsch and Gentleman (1969) is used to form
    the sums.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_cheb_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, mfirst.
        utils._BYREF_INT, # Input, mlast.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, l.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mlast].
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Input, ymin.
        utils._BYREF_FLOAT64, # Input, ymax.
        utils._EngineFloat64ArrayType.c_type, # Output, ff[mlast].
        utils._EngineFloat64ArrayType.c_type, # Input, a[na].
        utils._BYREF_INT, # Input, na.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, nwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _mfirst = utils._EngineIntScalarType(
        mfirst, {'fun_name': fun_name, 'entity': 'mfirst'},
    ).to_ctypes()
    _mlast = utils._EngineIntScalarType(
        mlast, {'fun_name': fun_name, 'entity': 'mlast'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _mlast_py = utils._EngineIntScalarType.to_py(_mlast)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mlast_py,),
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _ymin = utils._EngineFloat64ScalarType(
        ymin, {'fun_name': fun_name, 'entity': 'ymin'},
    ).to_ctypes()
    _ymax = utils._EngineFloat64ScalarType(
        ymax, {'fun_name': fun_name, 'entity': 'ymax'},
    ).to_ctypes()
    _ff_cls = utils._EngineFloat64ArrayType(
        ff, {'fun_name': fun_name, 'entity': 'ff'},
        exp_shape=(_mlast_py,),
        check_contents=False,
    )
    _ff = _ff_cls.to_ctypes(can_reference_contents=False)
    _na = utils._EngineIntScalarType(
        na, {'fun_name': fun_name, 'entity': 'na'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=((_k_py+1)*(_l_py+1),),
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _mfirst,
        _mlast,
        _k,
        _l,
        _x,
        _xmin,
        _xmax,
        _y,
        _ymin,
        _ymax,
        _ff,
        _a,
        _na,
        _work,
        _nwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, na = {:s}, k = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: na >= (k+1)*(l+1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mfirst = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'mlast = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mfirst <= mlast.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ymax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: y <= ymax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ymin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: y >= ymin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ymin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ymax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ymin < ymax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, x[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[I-1] <= xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, x[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[I-1] >= xmin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unexpected internal failure when evaluating the polynomial.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ff_cls.refresh_py(_ff)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_panel(m, px, py, x, y, f, w, lamda, mu, point, npoint, dl, c, nc, eps):
    u"""
    Least squares surface fit, bicubic splines.

    ``dim2_spline_panel`` forms a minimal, weighted least squares
    bicubic spline surface fit with prescribed knots to a given set of
    data points.

    For full information please refer to the NAG Library document for
    e02da

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02daf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    px : int
        The total number of knots lambda and mu associated with the
        variables x and y, respectively.

    py : int
        The total number of knots lambda and mu associated with the
        variables x and y, respectively.

    x : float, shape (`m`)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    y : float, shape (`m`)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    f : float, shape (`m`)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    w : float, shape (`m`)
        The weight w_r of the rth data point. It is important to note
        the definition of weight implied by the equation (1) in [Notes],
        since it is also common usage to define weight as the square of
        this weight. In this function, each w_r should be chosen
        inversely proportional to the (absolute) accuracy of the
        corresponding f_r, as expressed, for example, by the standard
        deviation or probable error of the f_r. When the f_r are all of
        the same accuracy, all the w_r may be set equal to 1.0.

    lamda : float, shape (`px`), modified in place
        `On entry`: `lamda`[i+3] must contain the ith interior knot
        lambda_{i+4} associated with the variable x, for i =
        1,2,...,`px`-8. The knots must be in nondecreasing order and lie
        strictly within the range covered by the data values of x. A
        knot is a value of x at which the spline is allowed to be
        discontinuous in the third derivative with respect to x, though
        continuous up to the second derivative. This degree of
        continuity can be reduced, if you require, by the use of
        coincident knots, provided that no more than four knots are
        chosen to coincide at any point. Two, or three, coincident knots
        allow loss of continuity in, respectively, the second and first
        derivative with respect to x at the value of x at which they
        coincide. Four coincident knots split the spline surface into
        two independent parts. For choice of knots see [Further
        Comments].

        `On exit`: the interior knots `lamda`[4] to `lamda`[`px`-5] are
        unchanged, and the segments `lamda`[0:4] and
        `lamda`[`px`-4:`px`] contain additional (exterior) knots
        introduced by the function in order to define the full set of
        B-splines required. The four knots in the first segment are all
        set equal to the lowest data value of x and the other four
        additional knots are all set equal to the highest value: there
        is experimental evidence that coincident end-knots are best for
        numerical accuracy. The complete array must be left undisturbed
        if ``dim2_spline_evalv`` or ``dim2_spline_evalm`` is to be used
        subsequently.

    mu : float, shape (`py`), modified in place
        `On entry`: `mu`[i+3] must contain the ith interior knot
        mu_{i+4} associated with the variable y, for i = 1,2,...,`py`-8.

        `On exit`: the same remarks apply to `mu` as to `lamda` above,
        with `y` replacing `x`, and y replacing x.

    point : int, shape (`npoint`)
        Indexing information usually provided by ``dim2_spline_sort``
        which enables the data points to be accessed in the order which
        produces the advantageous matrix structure mentioned in [Notes].
        This order is such that, if the (x,y) plane is thought of as
        being divided into rectangular panels by the two sets of knots,
        all data in a panel occur before data in succeeding panels,
        where the panels are numbered from bottom to top and then left
        to right with the usual arrangement of axes, as indicated in
        Figure [label omitted].

        [figure omitted]

        A data point lying exactly on one or more panel sides is
        considered to be in the highest numbered panel adjacent to the
        point. ``dim2_spline_sort`` should be called to obtain the array
        `point`, unless it is provided by other means.

    npoint : int
        The dimension of the array `point`.

    dl : float, shape (`nc`), modified in place
        `On exit`: gives the squares of the diagonal elements of the
        reduced triangular matrix, divided by the mean squared weight.
        It includes those elements, less than epsilon, which are treated
        as zero (see [Notes]).

    c : float, shape (`nc`), modified in place
        `On exit`: gives the coefficients of the fit. `c`[ (`py`-4)*
        (i-1) +j -1] is the coefficient c_{ij} of [Notes] and [Further
        Comments], for j = 1,2,...,`py`-4, for i = 1,2,...,`px`-4. These
        coefficients are used by ``dim2_spline_evalv`` or
        ``dim2_spline_evalm`` to calculate values of the fitted
        function.

    nc : int
        The value (`px`-4)*(`py`-4).

    eps : float
        A threshold epsilon for determining the effective rank of the
        system of linear equations. The rank is determined as the number
        of elements of the array `dl` which are nonzero. An element of
        `dl` is regarded as zero if it is less than epsilon. Machine
        precision is a suitable value for epsilon in most practical
        applications which have only 2 or 3 decimals accurate in data.
        If some coefficients of the fit prove to be very large compared
        with the data ordinates, this suggests that epsilon should be
        increased so as to decrease the rank. The array `dl` will give a
        guide to appropriate values of epsilon to achieve this, as well
        as to the choice of epsilon in other cases where some
        experimentation may be needed to determine a value which leads
        to a satisfactory fit.

    Returns
    -------
    sigma : float
        Sigma, the weighted sum of squares of residuals. This is not
        computed from the individual residuals but from the right-hand
        sides of the orthogonally-transformed linear equations. For
        further details see page 97 of Hayes and Halliday (1974). The
        two methods of computation are theoretically equivalent, but the
        results may differ because of rounding error.

    rank : int
        The rank of the system as determined by the value of the
        threshold epsilon.

        `rank` = `nc`

            The least squares solution is unique.

        `rank` != `nc`

            The minimal least squares solution is computed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            At least one set of knots is not in nondecreasing order.

        (`errno` 2)
            More than four knots coincide at a single point.

        (`errno` 3)
            Array `point` does not indicate the data points in panel
            order.

        (`errno` 4)
            On entry, `npoint` = *<value>*, `m` = *<value>*, `px` =
            *<value>* and `py` = *<value>*.

            Constraint: `npoint` >= `m`+(`px`-7)*(`py`-7).

        (`errno` 4)
            On entry, `nc` = *<value>*, `px` = *<value>* and `py` =
            *<value>*.

            Constraint: `nc` = (`px`-4)*(`py`-4).

        (`errno` 4)
            On entry, `py` = *<value>*.

            Constraint: `py` >= 8.

        (`errno` 4)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 8.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` > 1.

        (`errno` 5)
            All the weights are zero, or rank determined as zero.

    Notes
    -----
    ``dim2_spline_panel`` determines a bicubic spline fit s(x,y) to the
    set of data points (x_r,y_r,f_r) with weights w_r, for r =
    1,2,...,m.
    The two sets of internal knots of the spline, {lambda} and {mu},
    associated with the variables x and y respectively, are prescribed
    by you.
    These knots can be thought of as dividing the data region of the
    (x,y) plane into panels (see Figure [label omitted] in
    [Parameters]).
    A bicubic spline consists of a separate bicubic polynomial in each
    panel, the polynomials joining together with continuity up to the
    second derivative across the panel boundaries.

    s(x,y) has the property that Sigma, the sum of squares of its
    weighted residuals rho_r, for r = 1,2,...,m, where

        rho_r = w_r(s(x_r,y_r)-f_r)

    is as small as possible for a bicubic spline with the given knot
    sets.
    The function produces this minimized value of Sigma and the
    coefficients c_{ij} in the B-spline representation of s(x,y) -- see
    [Further Comments]. ``dim2_spline_evalv``, ``dim2_spline_evalm`` and
    ``dim2_spline_derivm`` are available to compute values and
    derivatives of the fitted spline from the coefficients c_{ij}.

    The least squares criterion is not always sufficient to determine
    the bicubic spline uniquely: there may be a whole family of splines
    which have the same minimum sum of squares.
    In these cases, the function selects from this family the spline for
    which the sum of squares of the coefficients c_{ij} is smallest: in
    other words, the minimal least squares solution.
    This choice, although arbitrary, reduces the risk of unwanted
    fluctuations in the spline fit.
    The method employed involves forming a system of m linear equations
    in the coefficients c_{ij} and then computing its least squares
    solution, which will be the minimal least squares solution when
    appropriate.
    The basis of the method is described in Hayes and Halliday (1974).
    The matrix of the equation is formed using a recurrence relation for
    B-splines which is numerically stable (see Cox (1972) and de Boor
    (1972) -- the former contains the more elementary derivation but,
    unlike de Boor (1972), does not cover the case of coincident knots).
    The least squares solution is also obtained in a stable manner by
    using orthogonal transformations, viz. a variant of Givens rotation
    (see Gentleman (1973)).
    This requires only one row of the matrix to be stored at a time.
    Advantage is taken of the stepped-band structure which the matrix
    possesses when the data points are suitably ordered, there being at
    most sixteen nonzero elements in any row because of the definition
    of B-splines.
    First the matrix is reduced to upper triangular form and then the
    diagonal elements of this triangle are examined in turn.
    When an element is encountered whose square, divided by the mean
    squared weight, is less than a threshold epsilon, it is replaced by
    zero and the rest of the elements in its row are reduced to zero by
    rotations with the remaining rows.
    The rank of the system is taken to be the number of nonzero diagonal
    elements in the final triangle, and the nonzero rows of this
    triangle are used to compute the minimal least squares solution.
    If all the diagonal elements are nonzero, the rank is equal to the
    number of coefficients c_{ij} and the solution obtained is the
    ordinary least squares solution, which is unique in this case.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Gentleman, W M, 1973, `Least squares computations by Givens
    transformations without square roots`, J. Inst. Math. Applic. (12),
    329--336

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_panel'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, px.
        utils._BYREF_INT, # Input, py.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, lamda[px].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, mu[py].
        utils._EngineIntArrayType.c_type, # Input, point[npoint].
        utils._BYREF_INT, # Input, npoint.
        utils._EngineFloat64ArrayType.c_type, # Output, dl[nc].
        utils._EngineFloat64ArrayType.c_type, # Output, c[nc].
        utils._BYREF_INT, # Input, nc.
        utils._EngineFloat64ArrayType.c_type, # Workspace, ws.
        utils._BYREF_INT, # Input, nws.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._BYREF_FLOAT64, # Output, sigma.
        utils._BYREF_INT, # Output, rank.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _px = utils._EngineIntScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineIntScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _px_py = utils._EngineIntScalarType.to_py(_px)
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_px_py,),
    )
    _lamda = _lamda_cls.to_ctypes()
    _py_py = utils._EngineIntScalarType.to_py(_py)
    _mu_cls = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_py_py,),
    )
    _mu = _mu_cls.to_ctypes()
    _npoint = utils._EngineIntScalarType(
        npoint, {'fun_name': fun_name, 'entity': 'npoint'},
    ).to_ctypes()
    _npoint_py = utils._EngineIntScalarType.to_py(_npoint)
    _point = utils._EngineIntArrayType(
        point, {'fun_name': fun_name, 'entity': 'point'},
        exp_shape=(_npoint_py,),
    ).to_ctypes()
    _nc = utils._EngineIntScalarType(
        nc, {'fun_name': fun_name, 'entity': 'nc'},
    ).to_ctypes()
    _dl_cls = utils._EngineFloat64ArrayType(
        dl, {'fun_name': fun_name, 'entity': 'dl'},
        exp_shape=((_px_py-4)*(_py_py-4),),
        check_contents=False,
    )
    _dl = _dl_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_px_py-4)*(_py_py-4),),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _ws = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nws = utils._EngineIntScalarType.c_type(0)
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType.empty_ctypes()
    _rank = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _m,
        _px,
        _py,
        _x,
        _y,
        _f,
        _w,
        _lamda,
        _mu,
        _point,
        _npoint,
        _dl,
        _c,
        _nc,
        _ws,
        _nws,
        _eps,
        _sigma,
        _rank,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one set of knots is not in nondecreasing order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'More than four knots coincide at a single point.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Array point does not indicate the data points in panel order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npoint = {:s}, m = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'px = {:s} and py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: npoint >= m+(px-7)*(py-7).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s}, px = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: nc = (px-4)*(py-4).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: py >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All the weights are zero, or rank determined as zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lamda_cls.refresh_py(_lamda)
        _mu_cls.refresh_py(_mu)
        _dl_cls.refresh_py(_dl)
        _c_cls.refresh_py(_c)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _rank_py = utils._EngineIntScalarType.to_py(_rank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['sigma', 'rank'],
        )(
            _sigma_py,
            _rank_py,
        )
    _err_class.manage()
    return return_data

def dim2_spline_grid(start, mx, x, my, y, f, s, nxest, nyest, nx, lamda, ny, mu, c, comm):
    u"""
    Least squares surface fit by bicubic splines with automatic knot
    placement, data on rectangular grid.

    ``dim2_spline_grid`` computes a bicubic spline approximation to a
    set of data values, given on a rectangular grid in the x-y plane.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02dc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02dcf.html

    Parameters
    ----------
    start : str, length 1
        Determines whether calculations are to be performed afresh (Cold
        Start) or whether knots found in previous calls are to be used
        as an initial estimate of knot placement (Warm Start).

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `nx`, `ny`, `lamda`, `mu`, `comm`['wrk'] or `comm`['iwrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `nx`, `ny`, `lamda`, `mu`, `comm`['wrk']
            and `comm`['iwrk'] must be unchanged from that previous
            call. This warm start can save much time in determining a
            satisfactory set of knots for the given value of `s`. This
            is particularly useful when different smoothing factors are
            used for the same dataset.

    mx : int
        m_x, the number of grid points along the x axis.

    x : float, shape (`mx`)
        `x`[q-1] must be set to x_q, the x coordinate of the qth grid
        point along the x axis, for q = 1,2,...,m_x.

    my : int
        m_y, the number of grid points along the y axis.

    y : float, shape (`my`)
        `y`[r-1] must be set to y_r, the y coordinate of the rth grid
        point along the y axis, for r = 1,2,...,m_y.

    f : float, shape (`mx`*`my`)
        `f`[m_y*(q-1)+r-1] must contain the data value f_{q,r}, for r =
        1,2,...,m_y, for q = 1,2,...,m_x.

    s : float
        The smoothing factor, S.

        If `s` = 0.0, the function returns an interpolating spline.

        If `s` is smaller than machine precision, it is assumed equal to
        zero.

        For advice on the choice of `s`, see [Notes] and [Further
        Comments].

    nxest : int
        An upper bound for the number of knots n_x and n_y required in
        the x- and y-directions respectively.

        In most practical situations, `nxest` = m_x/2 and `nyest` =
        m_y/2 is sufficient. `nxest` and `nyest` never need to be larger
        than m_x+4 and m_y+4 respectively, the numbers of knots needed
        for interpolation (`s` = 0.0).

        See also [Further Comments].

    nyest : int
        An upper bound for the number of knots n_x and n_y required in
        the x- and y-directions respectively.

        In most practical situations, `nxest` = m_x/2 and `nyest` =
        m_y/2 is sufficient. `nxest` and `nyest` never need to be larger
        than m_x+4 and m_y+4 respectively, the numbers of knots needed
        for interpolation (`s` = 0.0).

        See also [Further Comments].

    nx : int
        If the warm start option is used, the value of `nx` must be left
        unchanged from the previous call.

    lamda : float, shape (`nxest`), modified in place
        `On entry`: if the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[`nx`-1] must be left unchanged
        from the previous call.

        `On exit`: contains the complete set of knots lambda_i
        associated with the x variable, i.e., the interior knots
        `lamda`[4],`lamda`[5],...,`lamda`[`nx`-5] as well as the
        additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = `x`[0]

        and

            `lamda`[`nx`-4] = `lamda`[`nx`-3] = `lamda`[`nx`-2] =
            `lamda`[`nx`-1] = `x`[`mx`-1]

        needed for the B-spline representation.

    ny : int
        If the warm start option is used, the value of `ny` must be left
        unchanged from the previous call.

    mu : float, shape (`nyest`), modified in place
        `On entry`: if the warm start option is used, the values
        `mu`[0],`mu`[1],...,`mu`[`ny`-1] must be left unchanged from the
        previous call.

        `On exit`: contains the complete set of knots mu_i associated
        with the y variable, i.e., the interior knots
        `mu`[4],`mu`[5],...,`mu`[`ny`-5] as well as the additional knots

            `mu`[0] = `mu`[1] = `mu`[2] = `mu`[3] = `y`[0]

        and

            `mu`[`ny`-4] = `mu`[`ny`-3] = `mu`[`ny`-2] = `mu`[`ny`-1] =
            `y`[`my`-1]

        needed for the B-spline representation.

    c : float, shape ((`nxest`-4)*(`nyest`-4)), modified in place
        `On exit`: the coefficients of the spline approximation.
        `c`[(n_y-4) * (i-1) +j -1] is the coefficient c_{ij} defined in
        [Notes].

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    nx : int
        The total number of knots, n_x, of the computed spline with
        respect to the x variable.

    ny : int
        The total number of knots, n_y, of the computed spline with
        respect to the y variable.

    fp : float
        The sum of squared residuals, theta, of the computed spline
        approximation. If `fp` = 0.0, this is an interpolating spline.
        `fp` should equal `s` within a relative tolerance of 0.001
        unless `nx` = `ny` = 8, when the spline has no interior knots
        and so is simply a bicubic polynomial. For knots to be inserted,
        `s` must be set to a value below the value of `fp` produced in
        this case.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nxest` = *<value>*, `mx` = *<value>*, `nyest` =
            *<value>* and `my` = *<value>*.

            Constraint: if `s` = 0.0, `nxest` >= `mx`+4 and `nyest` >=
            `my`+4.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, `nyest` = *<value>*.

            Constraint: `nyest` >= *<value>*.

        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= *<value>*.

        (`errno` 1)
            On entry, `nxest` = *<value>*.

            Constraint: `nxest` >= *<value>*.

        (`errno` 1)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= *<value>*.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 2)
            On entry, q = *<value>*, `x`[q-2] = *<value>* and `x`[q-1] =
            *<value>*.

            Constraint: `x`[q-2] < `x`[q-1] for all q.

        (`errno` 3)
            On entry, r = *<value>*, `y`[r-2] = *<value>* and `y`[r-1] =
            *<value>*.

            Constraint: `y`[r-2] < `y`[r-1] for all r.

        (`errno` 4)
            The number of knots needed in one direction is greater than
            `nxest` or `nyest`: `nxest`, `nyest` = *<value>*,*<value>*.
            Possibly `s` is too small: `s` = *<value>*.

        (`errno` 5)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_spline_grid`` determines a smooth bicubic spline
    approximation s(x,y) to the set of data points (x_q,y_r,f_{q,r}),
    for r = 1,2,...,m_y, for q = 1,2,...,m_x.

    The spline is given in the B-spline representation

        s(x,y) = sum_i = 1^n_x-4sum_j = 1^n_y-4c_{ij}M_i(x)N_j(y),

    where M_i(x) and N_j(y) denote normalized cubic B-splines, the
    former defined on the knots lambda_i to lambda_{i+4} and the latter
    on the knots mu_j to mu_{j+4}.
    For further details, see Hayes and Halliday (1974) for bicubic
    splines and de Boor (1972) for normalized B-splines.

    The total numbers n_x and n_y of these knots and their values
    lambda_1,...,lambda_n_x and mu_1,...,mu_n_y are chosen automatically
    by the function.
    The knots lambda_5,...,lambda_{n_x-4} and mu_5,...,mu_{n_y-4} are
    the interior knots; they divide the approximation domain
    [x_1,x_m_x]*[y_1,y_m_y] into (n_x-7)*(n_y-7) subpanels
    [lambda_i,lambda_{i+1}]*[mu_j,mu_{j+1}], for j = 4,5,...,n_y-4, for
    i = 4,5,...,n_x-4.
    Then, much as in the curve case (see ``dim1_spline_auto``), the
    coefficients c_{ij} are determined as the solution of the following
    constrained minimization problem:

        minimize(eta),

    subject to the constraint

        theta = sum_q = 1^m_xsum_r = 1^m_y epsilon_q,r^2 <= S,

    [table omitted]

    By means of the argument S, 'the smoothing factor', you will then
    control the balance between smoothness and closeness of fit, as
    measured by the sum of squares of residuals in (3).
    If S is too large, the spline will be too smooth and signal will be
    lost (underfit); if S is too small, the spline will pick up too much
    noise (overfit).
    In the extreme cases the function will return an interpolating
    spline (theta = 0) if S is set to zero, and the least squares
    bicubic polynomial (eta = 0) if S is set very large.
    Experimenting with S-values between these two extremes should result
    in a good compromise. (See [Choice of s] for advice on choice of S.)

    The method employed is outlined in [Further Comments] and fully
    described in Dierckx (1981) and Dierckx (1982).
    It involves an adaptive strategy for locating the knots of the
    bicubic spline (depending on the function underlying the data and on
    the value of S), and an iterative method for solving the constrained
    minimization problem once the knots have been determined.

    Values and derivatives of the computed spline can subsequently be
    computed by calling ``dim2_spline_evalv``, ``dim2_spline_evalm`` or
    ``dim2_spline_derivm`` as described in [Further Comments].

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_grid'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, start(len=1).
        utils._BYREF_INT, # Input, mx.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mx].
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, f[mx*my].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_INT, # Input, nxest.
        utils._BYREF_INT, # Input, nyest.
        utils._BYREF_INT, # Input/Output, nx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, lamda[nxest].
        utils._BYREF_INT, # Input/Output, ny.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, mu[nyest].
        utils._EngineFloat64ArrayType.c_type, # Output, c[(nxest-4)*(nyest-4)].
        utils._BYREF_FLOAT64, # Output, fp.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, wrk[lwrk].
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Input/Output, iwrk[liwrk].
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of start.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _start = utils._EngineCharScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'}, exp_chrlen=1,
    ).to_ctypes()
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_mx_py*_my_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _nxest = utils._EngineIntScalarType(
        nxest, {'fun_name': fun_name, 'entity': 'nxest'},
    ).to_ctypes()
    _nyest = utils._EngineIntScalarType(
        nyest, {'fun_name': fun_name, 'entity': 'nyest'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nxest_py = utils._EngineIntScalarType.to_py(_nxest)
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_nxest_py,),
    )
    _lamda = _lamda_cls.to_ctypes()
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _nyest_py = utils._EngineIntScalarType.to_py(_nyest)
    _mu_cls = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_nyest_py,),
    )
    _mu = _mu_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_nxest_py-4)*(_nyest_py-4),),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _fp = utils._EngineFloat64ScalarType.empty_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('wrk', 'iwrk'),
    )
    if _comm_needs_init:
        _wrk = utils._EngineFloat64ArrayType.empty_ctypes(4*(_mx_py+_my_py)+11*(_nxest_py+_nyest_py)+_nxest_py*_my_py+max(_my_py, _nxest_py)+54)
        comm['wrk'] = _wrk
    else:
        _wrk = comm['wrk']
    _lwrk_py = utils._EngineFloat64ArrayType.get_shape(comm['wrk'], {'fun_name': fun_name, 'entity': 'comm[\'wrk\']'}, exp_rank=1, validate=False)[0]
    _lwrk = utils._EngineIntScalarType(
        _lwrk_py, {'fun_name': fun_name, 'entity': '_lwrk'},
    ).to_ctypes()
    if _comm_needs_init:
        _iwrk = utils._EngineIntArrayType.empty_ctypes(3+_mx_py+_my_py+_nxest_py+_nyest_py)
        comm['iwrk'] = _iwrk
    else:
        _iwrk = comm['iwrk']
    _liwrk_py = utils._EngineIntArrayType.get_shape(comm['iwrk'], {'fun_name': fun_name, 'entity': 'comm[\'iwrk\']'}, exp_rank=1, validate=False)[0]
    _liwrk = utils._EngineIntScalarType(
        _liwrk_py, {'fun_name': fun_name, 'entity': '_liwrk'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _start,
        _mx,
        _x,
        _my,
        _y,
        _f,
        _s,
        _nxest,
        _nyest,
        _nx,
        _lamda,
        _ny,
        _mu,
        _c,
        _fp,
        _wrk,
        _lwrk,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _start._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxest = {:s}, mx = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nyest = {:s} and my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if s = 0.0, nxest >= mx+4 and nyest >= my+4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nyest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nyest >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: my >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nxest >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mx >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start != \'W\' or \'C\': start = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}, x[q-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[q-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[q-2] < x[q-1] for all q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, y[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and y[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: y[r-2] < y[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of knots needed in one direction is greater than nxest or nyest:',
                    'nxest, nyest = {:s},{:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative process has failed to converge.',
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nx_py = utils._EngineIntScalarType.to_py(_nx)
        _lamda_cls.refresh_py(_lamda)
        _ny_py = utils._EngineIntScalarType.to_py(_ny)
        _mu_cls.refresh_py(_mu)
        _c_cls.refresh_py(_c)
        _fp_py = utils._EngineFloat64ScalarType.to_py(_fp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nx', 'ny', 'fp'],
        )(
            _nx_py,
            _ny_py,
            _fp_py,
        )
    _err_class.manage()
    return return_data

def dim2_spline_sctr(start, m, x, y, f, w, s, nxest, nyest, nx, lamda, ny, mu, c, comm):
    u"""
    Least squares surface fit by bicubic splines with automatic knot
    placement, scattered data.

    ``dim2_spline_sctr`` computes a bicubic spline approximation to a
    set of scattered data.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02dd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02ddf.html

    Parameters
    ----------
    start : str, length 1
        Determines whether calculations are to be performed afresh (Cold
        Start) or whether knots found in previous calls are to be used
        as an initial estimate of knot placement (Warm Start).

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `nx`, `ny`, `lamda`, `mu` or `comm`['wrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `nx`, `ny`, `lamda`, `mu` and
            `comm`['wrk'] must be unchanged from that previous call.
            This warm start can save much time in determining a
            satisfactory set of knots for the given value of `s`. This
            is particularly useful when different smoothing factors are
            used for the same dataset.

    m : int
        m, the number of data points.

        The number of data points with nonzero weight (see `w`) must be
        at least 16.

    x : float, shape (`m`), modified in place
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    y : float, shape (`m`), modified in place
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    f : float, shape (`m`)
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    w : float, shape (`m`)
        `w`[r-1] must be set to w_r, the rth value in the set of
        weights, for r = 1,2,...,m. Zero weights are permitted and the
        corresponding points are ignored, except when determining x_min,
        x_max, y_min and y_max (see [Further Comments]). For advice on
        the choice of weights, see [the E02 Introduction].

    s : float
        The smoothing factor, `s`.

        For advice on the choice of `s`, see [Notes] and [Further
        Comments].

    nxest : int
        An upper bound for the number of knots n_x and n_y required in
        the x- and y-directions respectively.

        In most practical situations, `nxest` = `nyest` = 4+sqrt(m/2) is
        sufficient.

        See also [Choice of nxest and nyest].

    nyest : int
        An upper bound for the number of knots n_x and n_y required in
        the x- and y-directions respectively.

        In most practical situations, `nxest` = `nyest` = 4+sqrt(m/2) is
        sufficient.

        See also [Choice of nxest and nyest].

    nx : int
        If the warm start option is used, the value of `nx` must be left
        unchanged from the previous call.

    lamda : float, shape (`nxest`), modified in place
        `On entry`: if the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[`nx`-1] must be left unchanged
        from the previous call.

        `On exit`: contains the complete set of knots lambda_i
        associated with the x variable, i.e., the interior knots
        `lamda`[4],`lamda`[5],...,`lamda`[`nx`-5] as well as the
        additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = x_min

        and

            `lamda`[`nx`-4] = `lamda`[`nx`-3] = `lamda`[`nx`-2] =
            `lamda`[`nx`-1] = x_max

        needed for the B-spline representation (where x_min and x_max
        are as described in [Notes]).

    ny : int
        If the warm start option is used, the value of `ny` must be left
        unchanged from the previous call.

    mu : float, shape (`nyest`), modified in place
        `On entry`: if the warm start option is used, the values
        `mu`[0],`mu`[1],...,`mu`[`ny`-1] must be left unchanged from the
        previous call.

        `On exit`: contains the complete set of knots mu_i associated
        with the y variable, i.e., the interior knots
        `mu`[4],`mu`[5],...,`mu`[`ny`-5] as well as the additional knots

            `mu`[0] = `mu`[1] = `mu`[2] = `mu`[3] = y_min

        and

            `mu`[`ny`-4] = `mu`[`ny`-3] = `mu`[`ny`-2] = `mu`[`ny`-1] =
            y_max

        needed for the B-spline representation (where y_min and y_max
        are as described in [Notes]).

    c : float, shape ((`nxest`-4)*(`nyest`-4)), modified in place
        `On exit`: the coefficients of the spline approximation.
        `c`[(n_y-4) * (i-1) +j -1] is the coefficient c_{ij} defined in
        [Notes].

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    nx : int
        The total number of knots, n_x, of the computed spline with
        respect to the x variable.

    ny : int
        The total number of knots, n_y, of the computed spline with
        respect to the y variable.

    fp : float
        The weighted sum of squared residuals, theta, of the computed
        spline approximation. `fp` should equal `s` within a relative
        tolerance of 0.001 unless `nx` = `ny` = 8, when the spline has
        no interior knots and so is simply a bicubic polynomial. For
        knots to be inserted, `s` must be set to a value below the value
        of `fp` produced in this case.

    rank : int
        Gives the rank of the system of equations used to compute the
        final spline (as determined by a suitable machine-dependent
        threshold). When `rank` = (`nx`-4)*(`ny`-4), the solution is
        unique; otherwise the system is rank-deficient and the
        minimum-norm solution is computed. The latter case may be caused
        by too small a value of `s`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nyest` = *<value>*.

            Constraint: `nyest` >= *<value>*.

        (`errno` 1)
            On entry, `nxest` = *<value>*.

            Constraint: `nxest` >= *<value>*.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, MM < 16, where MM is the number of points with
            nonzero weight: MM = *<value>*.

        (`errno` 2)
            On entry, all the values in array `y` are equal.

        (`errno` 2)
            On entry, all the values in array `x` are equal.

        (`errno` 3)
            The number of knots needed in one direction is greater than
            `nxest` or `nyest`: `nxest`, `nyest` = *<value>*,*<value>*.
            Possibly `s` is too small: `s` = *<value>*.

        (`errno` 4)
            No more knots added; the number of B-spline coefficients
            already exceeds `m`. Either `m` or `s` is probably too
            small: `m` = *<value>* and `s` = *<value>*.

        (`errno` 5)
            No more knots added; the additional knot would coincide with
            an old one. Possibly an inaccurate data point has too large
            a weight, or `s` is too small: `s` = *<value>*.

        (`errno` 6)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_spline_sctr`` determines a smooth bicubic spline
    approximation s(x,y) to the set of data points (x_r,y_r,f_r) with
    weights w_r, for r = 1,2,...,m.

    The approximation domain is considered to be the rectangle
    [x_min,x_max]*[y_min,y_max], where x_min(y_min) and x_max(y_max)
    denote the lowest and highest data values of x(y).

    The spline is given in the B-spline representation

        s(x,y) = sum_i = 1^n_x-4sum_j = 1^n_y-4c_{ij}M_i(x)N_j(y),

    where M_i(x) and N_j(y) denote normalized cubic B-splines, the
    former defined on the knots lambda_i to lambda_{i+4} and the latter
    on the knots mu_j to mu_{j+4}.
    For further details, see Hayes and Halliday (1974) for bicubic
    splines and de Boor (1972) for normalized B-splines.

    The total numbers n_x and n_y of these knots and their values
    lambda_1,...,lambda_n_x and mu_1,...,mu_n_y are chosen automatically
    by the function.
    The knots lambda_5,...,lambda_{n_x-4} and mu_5,...,mu_{n_y-4} are
    the interior knots; they divide the approximation domain
    [x_min,x_max]*[y_min,y_max] into (n_x-7)*(n_y-7) subpanels
    [lambda_i,lambda_{i+1}]*[mu_j,mu_{j+1}], for j = 4,5,...,n_y-4, for
    i = 4,5,...,n_x-4.
    Then, much as in the curve case (see ``dim1_spline_auto``), the
    coefficients c_{ij} are determined as the solution of the following
    constrained minimization problem:

    minimize

        eta,

    subject to the constraint

        theta = sum_r = 1^m epsilon_r^2 <= S

    [table omitted]

    By means of the argument `s`, 'the smoothing factor', you will then
    control the balance between smoothness and closeness of fit, as
    measured by the sum of squares of residuals in (3).
    If `s` is too large, the spline will be too smooth and signal will
    be lost (underfit); if `s` is too small, the spline will pick up too
    much noise (overfit).
    In the extreme cases the method would return an interpolating spline
    (theta = 0) if `s` were set to zero, and returns the least squares
    bicubic polynomial (eta = 0) if `s` is set very large.
    Experimenting with `s`-values between these two extremes should
    result in a good compromise. (See [Further Comments] for advice on
    choice of `s`.) Note however, that this function, unlike
    ``dim1_spline_auto`` and ``dim2_spline_grid``, does not allow `s` to
    be set exactly to zero: to compute an interpolant to scattered data,
    ``interp.dim2_scat`` or ``interp.dim2_scat_shep`` should be used.

    The method employed is outlined in [Further Comments] and fully
    described in Dierckx (1981a) and Dierckx (1981b).
    It involves an adaptive strategy for locating the knots of the
    bicubic spline (depending on the function underlying the data and on
    the value of `s`), and an iterative method for solving the
    constrained minimization problem once the knots have been
    determined.

    Values and derivatives of the computed spline can subsequently be
    computed by calling ``dim2_spline_evalv``, ``dim2_spline_evalm`` or
    ``dim2_spline_derivm`` as described in [Further Comments].

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An algorithm for surface fitting with spline
    functions`, IMA J. Numer. Anal. (1), 267--283

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Peters, G and Wilkinson, J H, 1970, `The least squares problem and
    pseudo-inverses`, Comput. J. (13), 309--316

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_sctr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, start(len=1).
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineFloat64ArrayType.c_type, # Input, w[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_INT, # Input, nxest.
        utils._BYREF_INT, # Input, nyest.
        utils._BYREF_INT, # Input/Output, nx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, lamda[nxest].
        utils._BYREF_INT, # Input/Output, ny.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, mu[nyest].
        utils._EngineFloat64ArrayType.c_type, # Output, c[(nxest-4)*(nyest-4)].
        utils._BYREF_FLOAT64, # Output, fp.
        utils._BYREF_INT, # Output, rank.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, wrk[lwrk].
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of start.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _start = utils._EngineCharScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    )
    _y = _y_cls.to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _nxest = utils._EngineIntScalarType(
        nxest, {'fun_name': fun_name, 'entity': 'nxest'},
    ).to_ctypes()
    _nyest = utils._EngineIntScalarType(
        nyest, {'fun_name': fun_name, 'entity': 'nyest'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nxest_py = utils._EngineIntScalarType.to_py(_nxest)
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_nxest_py,),
    )
    _lamda = _lamda_cls.to_ctypes()
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _nyest_py = utils._EngineIntScalarType.to_py(_nyest)
    _mu_cls = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_nyest_py,),
    )
    _mu = _mu_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_nxest_py-4)*(_nyest_py-4),),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _fp = utils._EngineFloat64ScalarType.empty_ctypes()
    _rank = utils._EngineIntScalarType.empty_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('wrk',),
    )
    _vl_u_py = _nxest_py-4
    _vl_v_py = _nyest_py-4
    _vl_w_py = max(_vl_u_py, _vl_v_py)
    _vl_lwrk2_py = 4*_vl_u_py*_vl_v_py*_vl_w_py+2*_vl_u_py*_vl_v_py+4*_vl_w_py
    if _comm_needs_init:
        _wrk = utils._EngineFloat64ArrayType.empty_ctypes((7*_vl_u_py*_vl_v_py+25*_vl_w_py)*(_vl_w_py+1)+2*(_vl_u_py+_vl_v_py+4*_m_py)+23*_vl_w_py+56+_vl_lwrk2_py)
        comm['wrk'] = _wrk
    else:
        _wrk = comm['wrk']
    _lwrk_py = utils._EngineFloat64ArrayType.get_shape(comm['wrk'], {'fun_name': fun_name, 'entity': 'comm[\'wrk\']'}, exp_rank=1, validate=False)[0]
    _lwrk = utils._EngineIntScalarType(
        _lwrk_py, {'fun_name': fun_name, 'entity': '_lwrk'},
    ).to_ctypes()
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _liwrk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _start,
        _m,
        _x,
        _y,
        _f,
        _w,
        _s,
        _nxest,
        _nyest,
        _nx,
        _lamda,
        _ny,
        _mu,
        _c,
        _fp,
        _rank,
        _wrk,
        _lwrk,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _start._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nyest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nyest >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nxest >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start != \'W\' or \'C\': start = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, MM < 16, where MM is the number of points with',
                    'nonzero weight: MM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the values in array y are equal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the values in array x are equal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of knots needed in one direction is greater than nxest or nyest:',
                    'nxest, nyest = {:s},{:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No more knots added; the number of B-spline coefficients already exceeds m.',
                    'Either m or s is probably too small:',
                    'm = {:s} and s = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No more knots added; the additional knot would coincide with an old one.',
                    'Possibly an inaccurate data point has too large a weight, or s is',
                    'too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative process has failed to converge.',
                    'Possibly s is too small: s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        _nx_py = utils._EngineIntScalarType.to_py(_nx)
        _lamda_cls.refresh_py(_lamda)
        _ny_py = utils._EngineIntScalarType.to_py(_ny)
        _mu_cls.refresh_py(_mu)
        _c_cls.refresh_py(_c)
        _fp_py = utils._EngineFloat64ScalarType.to_py(_fp)
        _rank_py = utils._EngineIntScalarType.to_py(_rank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nx', 'ny', 'fp', 'rank'],
        )(
            _nx_py,
            _ny_py,
            _fp_py,
            _rank_py,
        )
    _err_class.manage()
    return return_data

def dim2_spline_evalv(m, px, py, x, y, lamda, mu, c, ff):
    u"""
    Evaluation of fitted bicubic spline at a vector of points.

    ``dim2_spline_evalv`` calculates values of a bicubic spline from its
    B-spline representation.

    For full information please refer to the NAG Library document for
    e02de

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02def.html

    Parameters
    ----------
    m : int
        m, the number of points at which values of the spline are
        required.

    px : int
        `px` and `py` must specify the total number of knots associated
        with the variables x and y respectively. They are such that
        `px`-8 and `py`-8 are the corresponding numbers of interior
        knots.

    py : int
        `px` and `py` must specify the total number of knots associated
        with the variables x and y respectively. They are such that
        `px`-8 and `py`-8 are the corresponding numbers of interior
        knots.

    x : float, shape (`m`)
        `x` and `y` must contain x_r and y_r, for r = 1,2,...,m,
        respectively. These are the coordinates of the points at which
        values of the spline are required. The order of the points is
        immaterial.

    y : float, shape (`m`)
        `x` and `y` must contain x_r and y_r, for r = 1,2,...,m,
        respectively. These are the coordinates of the points at which
        values of the spline are required. The order of the points is
        immaterial.

    lamda : float, shape (`px`)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    mu : float, shape (`py`)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    c : float, shape ((`px`-4)*(`py`-4))
        `c`[ (`py`-4) * (i-1) +j -1] must contain the coefficient c_{ij}
        described in [Notes], for j = 1,2,...,`py`-4, for i =
        1,2,...,`px`-4.

    ff : float, shape (`m`), modified in place
        `On exit`: `ff`[r-1] contains the value of the spline at the
        point (x_r,y_r), for r = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `py` = *<value>*.

            Constraint: `py` >= 8.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 8.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, the knots in `mu` are not in nondecreasing order.

        (`errno` 2)
            On entry, the knots in `lamda` are not in nondecreasing
            order.

        (`errno` 3)
            On entry, point (`x`[K-1],`y`[K-1]) lies outside the
            rectangle bounded by `lamda`[3], `lamda`[`px`-4], `mu`[3],
            `mu`[`py`-4]: K = *<value>*, `x`[K-1] = *<value>* and
            `y`[K-1] = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_spline_evalv`` calculates values of the bicubic spline s(x,y)
    at prescribed points (x_r,y_r), for r = 1,2,...,m, from its
    augmented knot sets {lambda} and {mu} and from the coefficients
    c_{ij}, for j = 1,2,...,`py`-4, for i = 1,2,...,`px`-4, in its
    B-spline representation

        s(x,y) = sum_{ij}c_{ij}M_i(x)N_j(y).

    Here M_i(x) and N_j(y) denote normalized cubic B-splines, the former
    defined on the knots lambda_i to lambda_{i+4} and the latter on the
    knots mu_j to mu_{j+4}.

    This function may be used to calculate values of a bicubic spline
    given in the form produced by ``interp.dim2_spline_grid``,
    ``dim2_spline_panel``, ``dim2_spline_grid`` and
    ``dim2_spline_sctr``.
    It is derived from the function B2VRE in Anthony `et al.` (1982).

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_evalv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, px.
        utils._BYREF_INT, # Input, py.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[px].
        utils._EngineFloat64ArrayType.c_type, # Input, mu[py].
        utils._EngineFloat64ArrayType.c_type, # Input, c[(px-4)*(py-4)].
        utils._EngineFloat64ArrayType.c_type, # Output, ff[m].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _px = utils._EngineIntScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineIntScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _px_py = utils._EngineIntScalarType.to_py(_px)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_px_py,),
    ).to_ctypes()
    _py_py = utils._EngineIntScalarType.to_py(_py)
    _mu = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_py_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_px_py-4)*(_py_py-4),),
    ).to_ctypes()
    _ff_cls = utils._EngineFloat64ArrayType(
        ff, {'fun_name': fun_name, 'entity': 'ff'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _ff = _ff_cls.to_ctypes(can_reference_contents=False)
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        _ff,
        _wrk,
        _iwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: py >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the knots in mu are not in nondecreasing order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the knots in lamda are not in nondecreasing order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, point (x[K-1],y[K-1]) lies outside the rectangle bounded by',
                    'lamda[3], lamda[px-4], mu[3], mu[py-4]:',
                    'K = {:s}, x[K-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'y[K-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ff_cls.refresh_py(_ff)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_evalm(mx, my, px, py, x, y, lamda, mu, c, ff):
    u"""
    Evaluation of fitted bicubic spline at a mesh of points.

    ``dim2_spline_evalm`` calculates values of a bicubic spline from its
    B-spline representation.
    The spline is evaluated at all points on a rectangular grid.

    For full information please refer to the NAG Library document for
    e02df

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02dff.html

    Parameters
    ----------
    mx : int
        `mx` and `my` must specify m_x and m_y respectively, the number
        of points along the x and y axis that define the rectangular
        grid.

    my : int
        `mx` and `my` must specify m_x and m_y respectively, the number
        of points along the x and y axis that define the rectangular
        grid.

    px : int
        `px` and `py` must specify the total number of knots associated
        with the variables x and y respectively. They are such that
        `px`-8 and `py`-8 are the corresponding numbers of interior
        knots.

    py : int
        `px` and `py` must specify the total number of knots associated
        with the variables x and y respectively. They are such that
        `px`-8 and `py`-8 are the corresponding numbers of interior
        knots.

    x : float, shape (`mx`)
        `x` and `y` must contain x_q, for q = 1,2,...,m_x, and y_r, for
        r = 1,2,...,m_y, respectively. These are the x and y coordinates
        that define the rectangular grid of points at which values of
        the spline are required.

    y : float, shape (`my`)
        `x` and `y` must contain x_q, for q = 1,2,...,m_x, and y_r, for
        r = 1,2,...,m_y, respectively. These are the x and y coordinates
        that define the rectangular grid of points at which values of
        the spline are required.

    lamda : float, shape (`px`)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    mu : float, shape (`py`)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    c : float, shape ((`px`-4)*(`py`-4))
        `c`[ (`py`-4) * (i-1) +j -1] must contain the coefficient c_{ij}
        described in [Notes], for j = 1,2,...,`py`-4, for i =
        1,2,...,`px`-4.

    ff : float, shape (`mx`*`my`), modified in place
        `On exit`: `ff`[`my`*(q-1)+r-1] contains the value of the spline
        at the point (x_q,y_r), for r = 1,2,...,m_y, for q =
        1,2,...,m_x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 1)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= 1.

        (`errno` 1)
            On entry, `py` = *<value>*.

            Constraint: `py` >= 8.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 8.

        (`errno` 3)
            On entry, the knots in `mu` are not in nondecreasing order.

        (`errno` 3)
            On entry, the knots in `lamda` are not in nondecreasing
            order.

        (`errno` 4)
            On entry, `mu`[3] <= `y`[0] < ... < `y`[`my`-1] <=
            `mu`[`py`-4] is violated.

        (`errno` 4)
            On entry, `lamda`[3] <= `x`[0] < ... < `x`[`mx`-1] <=
            `lamda`[`px`-4] is violated.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_spline_evalm`` calculates values of the bicubic spline s(x,y)
    on a rectangular grid of points in the x-y plane, from its augmented
    knot sets {lambda} and {mu} and from the coefficients c_{ij}, for j
    = 1,2,...,`py`-4, for i = 1,2,...,`px`-4, in its B-spline
    representation

        s(x,y) = sum_{ij}c_{ij}M_i(x)N_j(y).

    Here M_i(x) and N_j(y) denote normalized cubic B-splines, the former
    defined on the knots lambda_i to lambda_{i+4} and the latter on the
    knots mu_j to mu_{j+4}.

    The points in the grid are defined by coordinates x_q, for q =
    1,2,...,m_x, along the x axis, and coordinates y_r, for r =
    1,2,...,m_y, along the y axis.

    This function may be used to calculate values of a bicubic spline
    given in the form produced by ``interp.dim2_spline_grid``,
    ``dim2_spline_panel``, ``dim2_spline_grid`` and
    ``dim2_spline_sctr``.
    It is derived from the function B2VRE in Anthony `et al.` (1982).

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_evalm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, px.
        utils._BYREF_INT, # Input, py.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mx].
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[px].
        utils._EngineFloat64ArrayType.c_type, # Input, mu[py].
        utils._EngineFloat64ArrayType.c_type, # Input, c[(px-4)*(py-4)].
        utils._EngineFloat64ArrayType.c_type, # Output, ff[mx*my].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _px = utils._EngineIntScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineIntScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _px_py = utils._EngineIntScalarType.to_py(_px)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_px_py,),
    ).to_ctypes()
    _py_py = utils._EngineIntScalarType.to_py(_py)
    _mu = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_py_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_px_py-4)*(_py_py-4),),
    ).to_ctypes()
    _ff_cls = utils._EngineFloat64ArrayType(
        ff, {'fun_name': fun_name, 'entity': 'ff'},
        exp_shape=(_mx_py*_my_py,),
        check_contents=False,
    )
    _ff = _ff_cls.to_ctypes(can_reference_contents=False)
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwrk = utils._EngineIntScalarType.c_type(0)
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _liwrk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _mx,
        _my,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        _ff,
        _wrk,
        _lwrk,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: py >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the knots in mu are not in nondecreasing order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the knots in lamda are not in nondecreasing order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mu[3] <= y[0] < ... < y[my-1] <= mu[py-4]',
                    'is violated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[3] <= x[0] < ... < x[mx-1] <= lamda[px-4]',
                    'is violated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ff_cls.refresh_py(_ff)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_derivm(mx, my, px, py, x, y, lamda, mu, c, nux, nuy, z):
    u"""
    Evaluation of spline surface at mesh of points with derivatives.

    ``dim2_spline_derivm`` computes the partial derivative (of order
    nu_x, nu_y), of a bicubic spline approximation to a set of data
    values, from its B-spline representation, at points on a rectangular
    grid in the x-y plane.
    This function may be used to calculate derivatives of a bicubic
    spline given in the form produced by ``interp.dim2_spline_grid``,
    ``dim2_spline_panel``, ``dim2_spline_grid`` and
    ``dim2_spline_sctr``.

    For full information please refer to the NAG Library document for
    e02dh

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02dhf.html

    Parameters
    ----------
    mx : int
        m_x, the number of grid points along the x axis.

    my : int
        m_y, the number of grid points along the y axis.

    px : int
        The total number of knots in the x-direction of the bicubic
        spline approximation, e.g., the value nx as returned by
        ``dim2_spline_grid``.

    py : int
        The total number of knots in the y-direction of the bicubic
        spline approximation, e.g., the value ny as returned by
        ``dim2_spline_grid``.

    x : float, shape (`mx`)
        `x`[q-1] must be set to x_q, the x coordinate of the qth grid
        point along the x axis, for q = 1,2,...,m_x, on which values of
        the partial derivative are sought.

    y : float, shape (`my`)
        `y`[r-1] must be set to y_r, the y coordinate of the rth grid
        point along the y axis, for r = 1,2,...,m_y on which values of
        the partial derivative are sought.

    lamda : float, shape (`px`)
        Contains the position of the knots in the x-direction of the
        bicubic spline approximation to be differentiated, e.g., lamda
        as returned by ``dim2_spline_grid``.

    mu : float, shape (`py`)
        Contains the position of the knots in the y-direction of the
        bicubic spline approximation to be differentiated, e.g., mu as
        returned by ``dim2_spline_grid``.

    c : float, shape ((`px`-4)*(`py`-4))
        The coefficients of the bicubic spline approximation to be
        differentiated, e.g., c as returned by ``dim2_spline_grid``.

    nux : int
        Specifies the order, nu_x of the partial derivative in the
        x-direction.

    nuy : int
        Specifies the order, nu_y of the partial derivative in the
        y-direction.

    z : float, shape (`mx`*`my`), modified in place
        `On exit`: `z`[m_y*(q-1)+r-1] contains the derivative
        (d^(nu_x+nu_y))/dx^(nu_x)dy^(nu_y)s(x_q,y_r), for r =
        1,2,...,m_y, for q = 1,2,...,m_x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nux` = *<value>*.

            Constraint: 0 <= `nux` <= 2.

        (`errno` 2)
            On entry, `nuy` = *<value>*.

            Constraint: 0 <= `nuy` <= 2.

        (`errno` 3)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= 1.

        (`errno` 4)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 5)
            On entry, for i = *<value>*, `x`[i-2] = *<value>* and
            `x`[i-1] = *<value>*.

            Constraint: `x`[i-2] <= `x`[i-1], for i = 2,3,...,`mx`.

        (`errno` 6)
            On entry, for i = *<value>*, `y`[i-2] = *<value>* and
            `y`[i-1] = *<value>*.

            Constraint: `y`[i-2] <= `y`[i-1], for i = 2,3,...,`my`.

    Notes
    -----
    ``dim2_spline_derivm`` determines the partial derivative
    (d^(nu_x+nu_y))/(dx^(nu_x)dy^(nu_y)) of a smooth bicubic spline
    approximation s(x,y) at the set of data points (x_q,y_r).

    The spline is given in the B-spline representation

        s(x,y) = sum_i = 1^n_x-4sum_j = 1^n_y-4c_{ij}M_i(x)N_j(y),

    where M_i(x) and N_j(y) denote normalized cubic B-splines, the
    former defined on the knots lambda_i to lambda_{i+4} and the latter
    on the knots mu_j to mu_{j+4}, with n_x and n_y the total numbers of
    knots of the computed spline with respect to the x and y variables
    respectively.
    For further details, see Hayes and Halliday (1974) for bicubic
    splines and de Boor (1972) for normalized B-splines.
    This function is suitable for B-spline representations returned by
    ``interp.dim2_spline_grid``, ``dim2_spline_panel``,
    ``dim2_spline_grid`` and ``dim2_spline_sctr``.

    The partial derivatives can be up to order 2 in each direction; thus
    the highest mixed derivative available is (d^4)/(dx^2dy^2).

    The points in the grid are defined by coordinates x_q, for q =
    1,2,...,m_x, along the x axis, and coordinates y_r, for r =
    1,2,...,m_y, along the y axis.

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_derivm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02dhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, px.
        utils._BYREF_INT, # Input, py.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mx].
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[px].
        utils._EngineFloat64ArrayType.c_type, # Input, mu[py].
        utils._EngineFloat64ArrayType.c_type, # Input, c[(px-4)*(py-4)].
        utils._BYREF_INT, # Input, nux.
        utils._BYREF_INT, # Input, nuy.
        utils._EngineFloat64ArrayType.c_type, # Output, z[mx*my].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _px = utils._EngineIntScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineIntScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _px_py = utils._EngineIntScalarType.to_py(_px)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_px_py,),
    ).to_ctypes()
    _py_py = utils._EngineIntScalarType.to_py(_py)
    _mu = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_py_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_px_py-4)*(_py_py-4),),
    ).to_ctypes()
    _nux = utils._EngineIntScalarType(
        nux, {'fun_name': fun_name, 'entity': 'nux'},
    ).to_ctypes()
    _nuy = utils._EngineIntScalarType(
        nuy, {'fun_name': fun_name, 'entity': 'nuy'},
    ).to_ctypes()
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_mx_py*_my_py,),
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mx,
        _my,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        _nux,
        _nuy,
        _z,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nux = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= nux <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nuy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= nuy <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'x[i-2] = {:s} and x[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[i-2] <= x[i-1], for i = 2,3,...,mx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'y[i-2] = {:s} and y[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: y[i-2] <= y[i-1], for i = 2,3,...,my.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        return_data = None
    _err_class.manage()
    return return_data

def glin_l1sol(m, a, b, nplus2, toler, x):
    u"""
    L_1-approximation by general linear function.

    ``glin_l1sol`` calculates an l_1 solution to an overdetermined
    system of linear equations.

    For full information please refer to the NAG Library document for
    e02ga

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02gaf.html

    Parameters
    ----------
    m : int
        The number of equations, m (the number of rows of the matrix A).

    a : float, shape (`m`+2, `nplus2`), modified in place
        `On entry`: `a`[i-1,j-1] must contain a_{ij}, the element in the
        ith row and jth column of the matrix A, for j = 1,2,...,n, for i
        = 1,2,...,m. The remaining elements need not be set.

        `On exit`: contains the last simplex tableau generated by the
        simplex method.

    b : float, shape (`m`), modified in place
        `On entry`: `b`[i-1] must contain b_i, the ith element of the
        vector b, for i = 1,2,...,m.

        `On exit`: the ith residual r_i corresponding to the solution
        vector x, for i = 1,2,...,m.

    nplus2 : int
        n+2, where n is the number of unknowns (the number of columns of
        the matrix A).

    toler : float
        A non-negative value. In general `toler` specifies a threshold
        below which numbers are regarded as zero. The recommended
        threshold value is epsilon^(2/3) where epsilon is the machine
        precision. The recommended value can be computed within the
        function by setting `toler` to zero. If premature termination
        occurs a larger value for `toler` may result in a valid
        solution.

        Suggested value: 0.0.

    x : float, shape (`nplus2`), modified in place
        `On exit`: `x`[j-1] contains the jth element of the solution
        vector x, for j = 1,2,...,n. The elements `x`[n] and `x`[n+1]
        are unused.

    Returns
    -------
    resid : float
        The sum of the absolute values of the residuals for the solution
        vector x.

    irank : int
        The computed rank of the matrix A.

    itera : int
        The number of iterations taken by the simplex method.

    Raises
    ------
    NagValueError
        (`errno` 2)
            Premature termination due to rounding errors. Try using
            larger value of `toler`: `toler` = *<value>*.

        (`errno` 3)
            On entry, `nplus2` = *<value>* and `m` = *<value>*.

            Constraint: 3 <= `nplus2` <= `m`+2.

        (`errno` 3)
            On entry, `nplus2` = *<value>*.

            Constraint: `nplus2` >= 3.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            An optimal solution has been obtained, but may not be
            unique.

        (`errno` 4)
            More than 1000*n iterations were performed. ``glin_l1sol``
            has terminated without calculating a solution. The output
            data from the function is as computed on the last good
            iteration. Consider increasing the value of `toler`.
            Alternatively, A may be ill conditioned---try scaling its
            columns.

    Notes
    -----
    Given a matrix A with m rows and n columns (m >= n) and a vector b
    with m elements, the function calculates an l_1 solution to the
    overdetermined system of equations

        Ax = b.

    That is to say, it calculates a vector x, with n elements, which
    minimizes the l_1 norm (the sum of the absolute values) of the
    residuals

        r(x) = sum_i = 1^m|r_i|,

    where the residuals r_i are given by

        r_i = b_i-sum_j = 1^na_{ij}x_j, i = 1,2,...,m.

    Here a_{ij} is the element in row i and column j of A, b_i is the
    ith element of b and x_j the jth element of x.
    The matrix A need not be of full rank.

    Typically in applications to data fitting, data consisting of m
    points with coordinates (t_i,y_i) are to be approximated in the l_1
    norm by a linear combination of known functions phi_j(t),

        alpha_1phi_1(t)+alpha_2phi_2(t)+...+alpha_nphi_n(t).

    This is equivalent to fitting an l_1 solution to the overdetermined
    system of equations

        sum_j = 1^n phi_j(t_i)alpha_j = y_i, i = 1,2,...,m.

    Thus if, for each value of i and j, the element a_{ij} of the matrix
    A in the previous paragraph is set equal to the value of phi_j(t_i)
    and b_i is set equal to y_i, the solution vector x will contain the
    required values of the alpha_j.
    Note that the independent variable t above can, instead, be a vector
    of several independent variables (this includes the case where each
    phi_i is a function of a different variable, or set of variables).

    The algorithm is a modification of the simplex method of linear
    programming applied to the primal formulation of the l_1 problem
    (see Barrodale and Roberts (1973) and Barrodale and Roberts (1974)).
    The modification allows several neighbouring simplex vertices to be
    passed through in a single iteration, providing a substantial
    improvement in efficiency.

    References
    ----------
    Barrodale, I and Roberts, F D K, 1973, `An improved algorithm for
    discrete l_1 linear approximation`, SIAM J. Numer. Anal. (10),
    839--848

    Barrodale, I and Roberts, F D K, 1974, `Solution of an
    overdetermined system of equations in the l_1-norm`, Comm. ACM
    (17(6)), 319--320
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.glin_l1sol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, nplus2].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[m].
        utils._BYREF_INT, # Input, nplus2.
        utils._BYREF_FLOAT64, # Input, toler.
        utils._EngineFloat64ArrayType.c_type, # Output, x[nplus2].
        utils._BYREF_FLOAT64, # Output, resid.
        utils._BYREF_INT, # Output, irank.
        utils._BYREF_INT, # Output, itera.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _nplus2 = utils._EngineIntScalarType(
        nplus2, {'fun_name': fun_name, 'entity': 'nplus2'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _nplus2_py = utils._EngineIntScalarType.to_py(_nplus2)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py+2, _nplus2_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py,),
    )
    _b = _b_cls.to_ctypes()
    _toler = utils._EngineFloat64ScalarType(
        toler, {'fun_name': fun_name, 'entity': 'toler'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nplus2_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _resid = utils._EngineFloat64ScalarType.empty_ctypes()
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _a,
        _lda,
        _b,
        _nplus2,
        _toler,
        _x,
        _resid,
        _irank,
        _itera,
        _iwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'An optimal solution has been obtained, but may not be unique.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Premature termination due to rounding errors.',
                    'Try using larger value of toler: toler = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nplus2 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 3 <= nplus2 <= m+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nplus2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nplus2 >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'More than 1000*n iterations were performed.',
                    'glin_l1sol has terminated without calculating a solution.',
                    'The output data from the function is as computed on the last good iteration.',
                    'Consider increasing the value of toler.',
                    'Alternatively, A may be ill conditioned---try scaling its columns.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _x_cls.refresh_py(_x)
        _resid_py = utils._EngineFloat64ScalarType.to_py(_resid)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['resid', 'irank', 'itera'],
        )(
            _resid_py,
            _irank_py,
            _itera_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glinc_l1sol(m, n, mpl, e, f, x, mxs, monit, iprint, indx, data=None):
    u"""
    L_1-approximation by general linear function subject to linear
    inequality constraints.

    ``glinc_l1sol`` calculates an l_1 solution to an overdetermined
    system of linear equations, possibly subject to linear inequality
    constraints.

    For full information please refer to the NAG Library document for
    e02gb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02gbf.html

    Parameters
    ----------
    m : int
        The number of equations in the overdetermined system, m (i.e.,
        the number of rows of the matrix A).

    n : int
        The number of unknowns, n (the number of columns of the matrix
        A).

    mpl : int
        m+l, where l is the number of constraints (which may be zero).

    e : float, shape (`n`, `mpl`), modified in place
        `On entry`: the equation and constraint matrices stored in the
        following manner.

        The first m columns contain the m rows of the matrix A; element
        `e`[i-1,j-1] specifying the element a_{ji} in the jth row and
        ith column of A (the coefficient of the ith unknown in the jth
        equation), for j = 1,2,...,m, for i = 1,2,...,n.

        The next l columns contain the l rows of the constraint matrix
        C; element `e`[i-1,j+m-1] containing the element c_{ji} in the
        jth row and ith column of C (the coefficient of the ith unknown
        in the jth constraint), for j = 1,2,...,l, for i = 1,2,...,n.

        `On exit`: unchanged, except possibly to the extent of a small
        multiple of the machine precision. (See [Further Comments].)

    f : float, shape (`mpl`)
        `f`[i-1], for i = 1,2,...,m, must contain b_i (the ith element
        of the right-hand side vector of the overdetermined system of
        equations) and `f`[m+i-1], for i = 1,2,...,l, must contain d_i
        (the ith element of the right-hand side vector of the
        constraints), where l is the number of constraints.

    x : float, shape (`n`), modified in place
        `On entry`: `x`[i-1] must contain an estimate of the ith
        unknown, for i = 1,2,...,n. If no better initial estimate for
        `x`[i-1] is available, set `x`[i-1] = 0.0.

        `On exit`: the latest estimate of the ith unknown, for i =
        1,2,...,n. If no exception or warning is raised on exit, these
        are the solution values.

    mxs : int
        The maximum number of steps to be allowed for the solution of
        the unconstrained problem. Typically this may be a modest
        multiple of n. If, on entry, `mxs` is zero or negative, the
        value returned by ``machine.integer_max`` is used.

    monit : callable monit(n, x, niter, k, el1n, data=None)
        `monit` can be used to print out the current values of any
        selection of its arguments.

        The frequency with which `monit` is called in ``glinc_l1sol`` is
        controlled by `iprint`.

        Parameters
        ~~~~~~~~~~
        n : int
            The number n of unknowns (the number of columns of the
            matrix A).

        x : float, ndarray, shape (`n`)
            The latest estimate of the unknowns.

        niter : int
            The number of iterations so far carried out.

        k : int
            The total number of equations and constraints which are
            currently active (i.e., the number of equations with zero
            residuals plus the number of constraints which are satisfied
            as equations).

        el1n : float
            The l_1-norm of the current residuals of the overdetermined
            system of equations.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int
        The frequency of iteration print out.

        `iprint` > 0

            `monit` is called every `iprint` iterations and at the
            solution.

        `iprint` = 0

            Information is printed out at the solution only. Otherwise
            `monit` is not called.

    indx : int, shape (`mpl`), modified in place
        `On exit`: specifies which columns of `e` relate to the inactive
        equations and constraints. `indx`[0] up to `indx`[`k`-1] number
        the active columns and `indx`[`k`] up to `indx`[`mpl`-1] number
        the inactive columns.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    k : int
        The total number of equations and constraints which are then
        active (i.e., the number of equations with zero residuals plus
        the number of constraints which are satisfied as equalities).

    el1n : float
        The l_1-norm (sum of absolute values) of the equation residuals.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The constraints cannot all be satisfied simultaneously: they
            are not compatible with one another. Hence no solution is
            possible.

        (`errno` 2)
            The limit imposed by `mxs` has been reached without finding
            a solution. Consider restarting from the current point by
            simply calling ``glinc_l1sol`` again without changing the
            arguments.

        (`errno` 3)
            The function has failed because of numerical difficulties;
            the problem is too ill-conditioned. Consider rescaling the
            unknowns.

        (`errno` 4)
            Elements 1 to `m` of one of the first `mpl` columns of the
            array `e` are all zero -- this corresponds to a zero row in
            either of the matrices A or C.

        (`errno` 4)
            On entry, `mpl` = *<value>* and `m` = *<value>*.

            Constraint: `mpl` >= `m`.

        (`errno` 4)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= `n`.

        (`errno` 4)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    Given a matrix A with m rows and n columns (m >= n) and a vector b
    with m elements, the function calculates an l_1 solution to the
    overdetermined system of equations

        Ax = b.

    That is to say, it calculates a vector x, with n elements, which
    minimizes the l_1-norm (the sum of the absolute values) of the
    residuals

        r(x) = sum_i = 1^m|r_i|,

    where the residuals r_i are given by

        r_i = b_i-sum_j = 1^na_{ij}x_j, i = 1,2,...,m.

    Here a_{ij} is the element in row i and column j of A, b_i is the
    ith element of b and x_j the jth element of x.

    If, in addition, a matrix C with l rows and n columns and a vector d
    with l elements, are given, the vector x computed by the function is
    such as to minimize the l_1-norm r(x) subject to the set of
    inequality constraints Cx >= d.

    The matrices A and C need not be of full rank.

    Typically in applications to data fitting, data consisting of m
    points with coordinates (t_i,y_i) is to be approximated by a linear
    combination of known functions phi_i(t),

        alpha_1phi_1(t)+alpha_2phi_2(t)+...+alpha_nphi_n(t),

    in the l_1-norm, possibly subject to linear inequality constraints
    on the coefficients alpha_j of the form C alpha >= d where alpha is
    the vector of the alpha_j and C and d are as in the previous
    paragraph.
    This is equivalent to finding an l_1 solution to the overdetermined
    system of equations

        sum_j = 1^n phi_j(t_i)alpha_j = y_i, i = 1,2,...,m,

    subject to C alpha >= d.

    Thus if, for each value of i and j, the element a_{ij} of the matrix
    A above is set equal to the value of phi_j(t_i) and b_i is equal to
    y_i and C and d are also supplied to the function, the solution
    vector x will contain the required values of the alpha_j.
    Note that the independent variable t above can, instead, be a vector
    of several independent variables (this includes the case where each
    of phi_i is a function of a different variable, or set of
    variables).

    The algorithm follows the Conn--Pietrzykowski approach (see Bartels
    `et al.` (1978) and Conn and Pietrzykowski (1977)), which is via an
    exact penalty function

        g(x) = gamma r(x)-sum_i = 1^lmin(0, c_i^Tx-d_i),

    where gamma is a penalty argument, c_i^T is the ith row of the
    matrix C, and d_i is the ith element of the vector d.
    It proceeds in a step-by-step manner much like the simplex method
    for linear programming but does not move from vertex to vertex and
    does not require the problem to be cast in a form containing only
    non-negative unknowns.
    It uses stable procedures to update an orthogonal factorization of
    the current set of active equations and constraints.

    References
    ----------
    Bartels, R H, Conn, A R and Charalambous, C, 1976, `Minimisation
    techniques for piecewise Differentiable functions -- the l_infinity
    solution to an overdetermined linear system`, Technical Report No.
    247, CORR 76/30, Mathematical Sciences Department, The John Hopkins
    University

    Bartels, R H, Conn, A R and Sinclair, J W, 1976, `A Fortran program
    for solving overdetermined systems of linear equations in the l_1
    Sense`, Technical Report No. 236, CORR 76/7, Mathematical Sciences
    Department, The John Hopkins University

    Bartels, R H, Conn, A R and Sinclair, J W, 1978, `Minimisation
    techniques for piecewise differentiable functions -- the l_1
    solution to an overdetermined linear system`, SIAM J. Numer. Anal.
    (15), 224--241

    Conn, A R and Pietrzykowski, T, 1977, `A penalty-function method
    converging directly to a constrained optimum`, SIAM J. Numer. Anal.
    (14), 348--375
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.glinc_l1sol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    monith = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, el1n.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_monith(_monit, en_data, n, x, niter, k, el1n, _iuser, _ruser):
        """Helper for monit."""
        try:
            _n_monith = utils._EngineIntScalarType.to_py(n.contents)
            _x_monith = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_monith,),
            )
            _niter_monith = utils._EngineIntScalarType.to_py(niter.contents)
            _k_monith = utils._EngineIntScalarType.to_py(k.contents)
            _el1n_monith = utils._EngineFloat64ScalarType.to_py(el1n.contents)
            if data is not None:
                monit(
                    _n_monith,
                    _x_monith,
                    _niter_monith,
                    _k_monith,
                    _el1n_monith,
                    data,
                )
            else:
                monit(
                    _n_monith,
                    _x_monith,
                    _niter_monith,
                    _k_monith,
                    _el1n_monith,
                )
            utils._unlink_pointer(x)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('e02gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, mpl.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, e[lde, mpl].
        utils._BYREF_INT, # Input, lde.
        utils._EngineFloat64ArrayType.c_type, # Input, f[mpl].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_INT, # Input, mxs.
        utils._UNUSED_FUNCTYPE,
        monith,
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, el1n.
        utils._EngineIntArrayType.c_type, # Output, indx[mpl].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._BYREF_INT, # Input, iw.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((e, 'e'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mpl = utils._EngineIntScalarType(
        mpl, {'fun_name': fun_name, 'entity': 'mpl'},
    ).to_ctypes()
    _lde_py = utils._EngineFloat64ArrayType.get_nd(e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lde = utils._EngineIntScalarType(
        _lde_py, {'fun_name': fun_name, 'entity': '_lde'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _mpl_py = utils._EngineIntScalarType.to_py(_mpl)
    _e_cls = utils._EngineFloat64ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_n_py, _mpl_py),
        mutual_sorder=_m_sorder,
    )
    _e = _e_cls.to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_mpl_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _mxs = utils._EngineIntScalarType(
        mxs, {'fun_name': fun_name, 'entity': 'mxs'},
    ).to_ctypes()
    utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'})
    _monit = utils._cast_to_unused_callback(monit)
    _monith = utils._cast_callback(
        monith, cb_monith,
        {'fun_name': fun_name, 'entity': 'monit'},
    )
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _el1n = utils._EngineFloat64ScalarType.empty_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_mpl_py,),
        check_contents=False,
    )
    _indx = _indx_cls.to_ctypes(can_reference_contents=False)
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iw = utils._EngineIntScalarType.c_type(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _m,
        _n,
        _mpl,
        _e,
        _lde,
        _f,
        _x,
        _mxs,
        _monit,
        _monith,
        _iprint,
        _k,
        _el1n,
        _indx,
        _w,
        _iw,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The constraints cannot all be satisfied simultaneously: they are not',
                    'compatible with one another. Hence no solution is possible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The limit imposed by mxs has been reached without finding a',
                    'solution. Consider restarting from the current point by simply calling',
                    'glinc_l1sol again without changing the arguments.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The function has failed because of numerical difficulties; the problem is',
                    'too ill-conditioned. Consider rescaling the unknowns.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Elements 1 to m of one of the first mpl columns of the array',
                    'e are all zero -- this corresponds to a zero row in either of the matrices A',
                    'or C.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mpl = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mpl >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _e_cls.refresh_py(_e)
        _x_cls.refresh_py(_x)
        _indx_cls.refresh_py(_indx)
        _k_py = utils._EngineIntScalarType.to_py(_k)
        _el1n_py = utils._EngineFloat64ScalarType.to_py(_el1n)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['k', 'el1n'],
        )(
            _k_py,
            _el1n_py,
        )
    _err_class.manage()
    return return_data

def glin_linf(m, n, a, b, tol, relerr, x):
    u"""
    L_infinity-approximation by general linear function.

    ``glin_linf`` calculates an l_infinity solution to an overdetermined
    system of linear equations.

    For full information please refer to the NAG Library document for
    e02gc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02gcf.html

    Parameters
    ----------
    m : int
        The number of equations, m (the number of rows of the matrix A).

    n : int
        The number of unknowns, n (the number of columns of the matrix
        A).

    a : float, shape (`n`+3, `m`+1), modified in place
        `On entry`: `a`[j-1,i-1] must contain a_{ij}, the element in the
        ith row and jth column of the matrix A, for j = 1,2,...,n, for i
        = 1,2,...,m, (that is, the **transpose** of the matrix). The
        remaining elements need not be set. Preferably, the columns of
        the matrix A (rows of the argument `a`) should be scaled before
        entry: see [Accuracy].

        `On exit`: contains the last simplex tableau.

    b : float, shape (`m`), modified in place
        `On entry`: `b`[i-1] must contain b_i, the ith element of the
        vector b, for i = 1,2,...,m.

        `On exit`: the ith residual r_i corresponding to the solution
        vector x, for i = 1,2,...,m. Note however that these residuals
        may contain few significant figures, especially when `resmax` is
        within one or two orders of magnitude of `tol`. Indeed if
        `resmax` <= `tol`, the elements `b`[i-1] may all be set to zero.
        It is, therefore, often advisable to compute the residuals
        directly.

    tol : float
        A threshold below which numbers are regarded as zero. The
        recommended threshold value is 10.0*epsilon, where epsilon is
        the machine precision. If `tol` <= 0.0 on entry, the recommended
        value is used within the function. If premature termination
        occurs, a larger value for `tol` may result in a valid solution.

        Suggested value: 0.0.

    relerr : float
        Must be set to a bound on the relative error acceptable in the
        maximum residual at the solution.

        If `relerr` <= 0.0, the l_infinity solution is computed, and
        `relerr` is set to 0.0 on exit.

        If `relerr` > 0.0, the function obtains instead an approximate
        solution for which the largest residual is less than
        1.0+`relerr` times that of the l_infinity solution; on exit,
        `relerr` contains a smaller value such that the above bound
        still applies. (The usual result of this option, say with
        `relerr` = 0.1, is a saving in the number of simplex
        iterations).

    x : float, shape (`n`), modified in place
        `On exit`: if the function exits successfully or `errno` = 1,
        `x`[j-1] contains the jth element of the solution vector x, for
        j = 1,2,...,n. Whether this is an l_infinity solution or an
        approximation to one, depends on the value of `relerr` on entry.

    Returns
    -------
    relerr : float
        Is altered as described above.

    resmax : float
        If the function exits successfully or `errno` = 1, `resmax`
        contains the absolute value of the largest residual(s) for the
        solution vector x. (See `b`.)

    irank : int
        If the function exits successfully or `errno` = 1, `irank`
        contains the computed rank of the matrix A.

    itera : int
        If the function exits successfully or `errno` = 1, `itera`
        contains the number of iterations taken by the simplex method.

    Raises
    ------
    NagValueError
        (`errno` 2)
            Premature termination due to rounding errors. Try using
            larger value of `tol`: `tol` = *<value>*.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= `n`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            An optimal solution has been obtained, but may not be
            unique.

    Notes
    -----
    Given a matrix A with m rows and n columns (m >= n) and a vector b
    with m elements, the function calculates an l_infinity solution to
    the overdetermined system of equations

        Ax = b.

    That is to say, it calculates a vector x, with n elements, which
    minimizes the l_infinity norm of the residuals (the absolutely
    largest residual)

        r(x) = max_{1 <= i <= m}|r_i|

    where the residuals r_i are given by

        r_i = b_i-sum_j = 1^na_{ij}x_j, i = 1,2,...,m.

    Here a_{ij} is the element in row i and column j of A, b_i is the
    ith element of b and x_j the jth element of x.
    The matrix A need not be of full rank.
    The solution is not unique in this case, and may not be unique even
    if A is of full rank.

    Alternatively, in applications where a complete minimization of the
    l_infinity norm is not necessary, you may obtain an approximate
    solution, usually in shorter time, by giving an appropriate value to
    the argument `relerr`.

    Typically in applications to data fitting, data consisting of m
    points with coordinates (t_i,y_i) is to be approximated in the
    l_infinity norm by a linear combination of known functions phi_j(t),

        alpha_1phi_1(t)+alpha_2phi_2(t)+...+alpha_nphi_n(t).

    This is equivalent to finding an l_infinity solution to the
    overdetermined system of equations

        sum_j = 1^n phi_j(t_i)alpha_j = y_i, i = 1,2,...,m.

    Thus if, for each value of i and j the element a_{ij} of the matrix
    A above is set equal to the value of phi_j(t_i) and b_i is set equal
    to y_i, the solution vector x will contain the required values of
    the alpha_j.
    Note that the independent variable t above can, instead, be a vector
    of several independent variables (this includes the case where each
    phi_i is a function of a different variable, or set of variables).

    The algorithm is a modification of the simplex method of linear
    programming applied to the dual formation of the l_infinity problem
    (see Barrodale and Phillips (1974) and Barrodale and Phillips
    (1975)).
    The modifications are designed to improve the efficiency and
    stability of the simplex method for this particular application.

    References
    ----------
    Barrodale, I and Phillips, C, 1974, `An improved algorithm for
    discrete Chebyshev linear approximation`, Proc. 4th Manitoba Conf.
    Numerical Mathematics, 177--190, University of Manitoba, Canada

    Barrodale, I and Phillips, C, 1975, `Solution of an overdetermined
    system of linear equations in the Chebyshev norm [F4] (Algorithm
    495)`, ACM Trans. Math. Software (1(3)), 264--270
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.glin_linf'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02gcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, sda.
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, sda].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[m].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Input/Output, relerr.
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_FLOAT64, # Output, resmax.
        utils._BYREF_INT, # Output, irank.
        utils._BYREF_INT, # Output, itera.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _sda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sda = utils._EngineIntScalarType(
        _sda_py, {'fun_name': fun_name, 'entity': '_sda'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py+3, _m_py+1),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py,),
    )
    _b = _b_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _relerr = utils._EngineFloat64ScalarType(
        relerr, {'fun_name': fun_name, 'entity': 'relerr'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _resmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _sda,
        _lda,
        _a,
        _b,
        _tol,
        _relerr,
        _x,
        _resmax,
        _irank,
        _itera,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'An optimal solution has been obtained, but may not be unique.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Premature termination due to rounding errors.',
                    'Try using larger value of tol: tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _relerr_py = utils._EngineFloat64ScalarType.to_py(_relerr)
        _x_cls.refresh_py(_x)
        _resmax_py = utils._EngineFloat64ScalarType.to_py(_resmax)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['relerr', 'resmax', 'irank', 'itera'],
        )(
            _relerr_py,
            _resmax_py,
            _irank_py,
            _itera_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim2_spline_ts_sctr(n, x, y, f, lsminp, lsmaxp, nxcels, nycels, comm):
    u"""
    Spline approximation to a set of scattered data using a two-stage
    approximation method.

    Note: this function uses optional algorithmic parameters, see also:
    ``opt_set``, ``opt_get``.

    ``dim2_spline_ts_sctr`` computes a spline approximation to a set of
    scattered data using a two-stage approximation method.

    The computational complexity of the method grows linearly with the
    number of data points; hence large datasets are easily accommodated.

    For full information please refer to the NAG Library document for
    e02jd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02jdf.html

    Parameters
    ----------
    n : int
        n, the number of data values to be fitted.

    x : float, shape (`n`)
        The (x_i,y_i,f_i) data values to be fitted.

    y : float, shape (`n`)
        The (x_i,y_i,f_i) data values to be fitted.

    f : float, shape (`n`)
        The (x_i,y_i,f_i) data values to be fitted.

    lsminp : int
        `lsminp` and `lsmaxp` are control parameters for the local
        approximations.

        Each local approximation is computed on a local domain
        containing one of the triangles in the discretization of the
        bounding box.

        The size of each local domain will be adaptively chosen such
        that if it contains fewer than `lsminp` sample points it is
        expanded, else if it contains greater than `lsmaxp` sample
        points a thinning method is applied. `lsmaxp` mainly controls
        computational cost (in that working with a thinned set of points
        is cheaper and may be appropriate if the input data is densely
        distributed), while `lsminp` allows handling of different types
        of scattered data.

        Setting `lsmaxp` < `lsminp`, and, therefore, forcing either
        expansion or thinning, may be useful for computing initial
        coarse approximations.

        In general smaller values for these arguments reduces cost.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate values for `lsminp` and `lsmaxp`.

    lsmaxp : int
        `lsminp` and `lsmaxp` are control parameters for the local
        approximations.

        Each local approximation is computed on a local domain
        containing one of the triangles in the discretization of the
        bounding box.

        The size of each local domain will be adaptively chosen such
        that if it contains fewer than `lsminp` sample points it is
        expanded, else if it contains greater than `lsmaxp` sample
        points a thinning method is applied. `lsmaxp` mainly controls
        computational cost (in that working with a thinned set of points
        is cheaper and may be appropriate if the input data is densely
        distributed), while `lsminp` allows handling of different types
        of scattered data.

        Setting `lsmaxp` < `lsminp`, and, therefore, forcing either
        expansion or thinning, may be useful for computing initial
        coarse approximations.

        In general smaller values for these arguments reduces cost.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate values for `lsminp` and `lsmaxp`.

    nxcels : int
        `nxcels` (respectively `nycels`) is the number of cells in the x
        (respectively y) direction that will be used to create the
        triangulation of the bounding box of the domain of the function
        to be fitted.

        Greater efficiency generally comes when `nxcels` and `nycels`
        are chosen to be of the same order of magnitude and are such
        that `n` is O(`nxcels`*`nycels`).

        Thus for a 'square' triangulation --- when `nxcels` = `nycels`
        --- the quantities sqrt(`n`) and `nxcels` should be of the same
        order of magnitude.

        See also [Further Comments].

    nycels : int
        `nxcels` (respectively `nycels`) is the number of cells in the x
        (respectively y) direction that will be used to create the
        triangulation of the bounding box of the domain of the function
        to be fitted.

        Greater efficiency generally comes when `nxcels` and `nycels`
        are chosen to be of the same order of magnitude and are such
        that `n` is O(`nxcels`*`nycels`).

        Thus for a 'square' triangulation --- when `nxcels` = `nycels`
        --- the quantities sqrt(`n`) and `nxcels` should be of the same
        order of magnitude.

        See also [Further Comments].

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt_set``.

    Other Parameters
    ----------------
    'Averaged Spline' : str
        Default = 'NO'

        When the bounding box is triangulated there are 8 equivalent
        configurations of the mesh.
        Setting 'Averaged Spline' = 'YES' will use the averaged value of
        the 8 possible local polynomial approximations over each
        triangle in the mesh.
        This usually gives better results but at (about 8 times) higher
        computational cost.

        Constraint: 'Averaged Spline' = 'YES' or 'NO'.

    'Global Smoothing Level' : int
        Default = 1

        The smoothness level for the global spline approximation.

        'Global Smoothing Level' = 1
            Will use C^1 piecewise cubics.

        'Global Smoothing Level' = 2
            Will use C^2 piecewise sextics.

    'Interpolation Only RBF' : str
        Default = 'YES'

        If 'Interpolation Only RBF' = 'YES', each local RBF
        approximation is computed by interpolation.

        If 'Interpolation Only RBF' = 'NO', each local RBF approximation
        is computed by a discrete least squares approach.
        This is likely to be more accurate and more expensive than
        interpolation.

        If 'Local Method' = 'HYBRID' or 'POLYNOMIAL', this option
        setting is ignored.

    'Local Method' : str
        Default = 'POLYNOMIAL'

        The local approximation scheme to use.

        'Local Method' = 'POLYNOMIAL'
            Uses least squares polynomial approximations.

        'Local Method' = 'HYBRID'
            Uses hybrid polynomial and RBF approximations.

        'Local Method' = 'RBF'
            Uses pure RBF approximations.

        In general 'POLYNOMIAL' is less computationally expensive than
        'HYBRID' is less computationally expensive than 'RBF' with the
        reverse ordering holding for accuracy of results.

    'Minimum Singular Value LHA' : float
        Default = 1.0

        A tolerance measure for accepting or rejecting a local hybrid
        approximation (LHA) as reliable.

        The solution of a local least squares problem solved on each
        triangle subdomain is accepted as reliable if the minimum
        singular value sigma of the collocation matrix (of polynomial
        and radial basis function terms) associated with the least
        squares problem satisfies 'Minimum Singular Value LHA' <= sigma.

        In general the approximation power will be reduced as 'Minimum
        Singular Value LHA' is reduced. (A small sigma indicates that
        the local data has hidden redundancies which prevent it from
        carrying enough information for a good approximation to be
        made.) Setting 'Minimum Singular Value LHA' very large may have
        the detrimental effect that only approximations of low degree
        are deemed reliable.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate value for this parameter.

        If 'Local Method' = 'POLYNOMIAL' or 'RBF', this option setting
        is ignored.

    'Minimum Singular Value LPA' : float
        Default = 1.0

        A tolerance measure for accepting or rejecting a local
        polynomial approximation (LPA) as reliable.
        Clearly this setting is relevant when 'Local Method' =
        'POLYNOMIAL', but it also may be used when 'Local Method' =
        'HYBRID' (see [Further Comments].)

        The solution of a local least squares problem solved on each
        triangle subdomain is accepted as reliable if the minimum
        singular value sigma of the matrix (of Bernstein polynomial
        values) associated with the least squares problem satisfies
        'Minimum Singular Value LPA' <= sigma.

        In general the approximation power will be reduced as 'Minimum
        Singular Value LPA' is reduced. (A small sigma indicates that
        the local data has hidden redundancies which prevent it from
        carrying enough information for a good approximation to be
        made.) Setting 'Minimum Singular Value LPA' very large may have
        the detrimental effect that only approximations of low degree
        are deemed reliable.

        'Minimum Singular Value LPA' will have no effect if 'Polynomial
        Starting Degree' = 0, and it will have little effect if the
        input data is 'smooth' (e.g., from a known function).

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate value for this parameter.

        If 'Local Method' = 'RBF', this option setting is ignored.

    'Polynomial Starting Degree' : int
        Default = 5 if 'Local Method' = 'HYBRID',
        Default = 1 otherwise

        The degree to be used for the polynomial part in the initial
        step of each local approximation.

        At this initial step the method will attempt to fit with a local
        approximation having polynomial part of degree 'Polynomial
        Starting Degree'.
        If 'Local Method' = 'POLYNOMIAL' and the approximation is deemed
        unreliable (according to 'Minimum Singular Value LPA'), the
        degree will be decremented by one and a new local approximation
        computed, ending with a constant approximation if no other is
        reliable.
        If 'Local Method' = 'HYBRID' and the approximation is deemed
        unreliable (according to 'Minimum Singular Value LHA'), a pure
        polynomial approximation of this degree will be tried instead.
        The method then proceeds as in the 'POLYNOMIAL' case.

        'Polynomial Starting Degree' is bounded from above by the
        maximum possible spline degree, which is 6 (when performing C^2
        global super-smoothing).
        Note that the best-case approximation error (see [Accuracy]) for
        C^2 smoothing with 'Supersmooth C2' = 'NO' is achieved for local
        polynomials of degree 5; that is, for this level of global
        smoothing no further benefit is gained by setting 'Polynomial
        Starting Degree' = 6.

        The default value gives a good compromise between efficiency and
        accuracy.
        In general the best approximation can be obtained by setting:

            If 'Local Method' = 'POLYNOMIAL'

                if 'Global Smoothing Level' = 1, 'Polynomial Starting
                Degree' = 3;

                if 'Global Smoothing Level' = 2;

                    if 'Supersmooth C2' = 'NO', 'Polynomial Starting
                    Degree' = 5;

                    otherwise 'Polynomial Starting Degree' = 6.

            If 'Local Method' = 'HYBRID', 'Polynomial Starting Degree'
            as small as possible.

        If 'Local Method' = 'RBF', this option setting is ignored.

    'Radial Basis Function' : str
        Default = 'MQ'

        'Radial Basis Function' selects the RBF to use in each local RBF
        approximation, while 'Scaling Coefficient RBF' selects the scale
        factor to use in its evaluation, as described below.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate scale factor and RBF.

        If 'Local Method' = 'POLYNOMIAL', these option settings are
        ignored.

        If 'Local Method' = 'HYBRID' or 'RBF', the following
        (conditionally) positive definite functions may be chosen.

        Define R = sqrt(x^2+y^2) and rho = R/r.

        [table omitted]

        If 'Local Method' = 'HYBRID' the following conditionally
        positive definite functions may also be chosen.

        [table omitted]

    'Scaling Coefficient RBF' : float
        Default = 1.0

        'Radial Basis Function' selects the RBF to use in each local RBF
        approximation, while 'Scaling Coefficient RBF' selects the scale
        factor to use in its evaluation, as described below.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate scale factor and RBF.

        If 'Local Method' = 'POLYNOMIAL', these option settings are
        ignored.

        If 'Local Method' = 'HYBRID' or 'RBF', the following
        (conditionally) positive definite functions may be chosen.

        Define R = sqrt(x^2+y^2) and rho = R/r.

        [table omitted]

        If 'Local Method' = 'HYBRID' the following conditionally
        positive definite functions may also be chosen.

        [table omitted]

    'Separation LRBFA' : float
        Default = 16.0/'Scaling Coefficient RBF'

        A knot-separation parameter used to control the condition number
        of the matrix used in each local RBF approximation (LRBFA).
        A smaller value may mean greater numerical stability but fewer
        knots.

        If 'Local Method' = 'HYBRID' or 'POLYNOMIAL', this option
        setting is ignored.

    'Supersmooth C2' : str
        Default = 'NO'

        If 'Supersmooth C2' = 'YES', the C^2 spline is generated using
        additional smoothness constraints.
        This usually gives better results but at higher computational
        cost.

        If 'Global Smoothing Level' = 1 this option setting is ignored.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 4)
            On entry, `lsminp` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `lsminp` <= `n`.

        (`errno` 5)
            On entry, `lsmaxp` = *<value>*.

            Constraint: `lsmaxp` >= 1.

        (`errno` 6)
            On entry, `nxcels` = *<value>*.

            Constraint: `nxcels` >= 1.

        (`errno` 7)
            On entry, `nycels` = *<value>*.

            Constraint: `nycels` >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 11)
            An unexpected algorithmic failure was encountered. Please
            contact NAG.

        (`errno` 12)
            On entry, all elements of `x` or of `y` are equal.

        (`errno` 20)
            The selected radial basis function cannot be used with the
            RBF local method.

        (`errno` 21)
            The value of option 'Polynomial Starting Degree' was
            invalid.

    Notes
    -----
    ``dim2_spline_ts_sctr`` determines a smooth bivariate spline
    approximation to a set of data points (x_i,y_i,f_i), for i =
    1,2,...,n.
    Here, 'smooth' means C^1 or C^2. (You may select the degree of
    smoothing using the option 'Global Smoothing Level'.)

    The approximation domain is the bounding box
    [x_min,x_max]*[y_min,y_max], where x_min (respectively y_min) and
    x_max (respectively y_max) denote the lowest and highest data values
    of the (x_i) (respectively (y_i)).

    The spline is computed by local approximations on a uniform
    triangulation of the bounding box.
    These approximations are extended to a smooth spline representation
    of the surface over the domain.
    The local approximation scheme is controlled by the option 'Local
    Method'.
    The schemes provided are: by least squares polynomial approximation
    (Davydov and Zeilfelder (2004)); by hybrid polynomial and radial
    basis function (RBF) approximation (Davydov `et al.` (2006)); or by
    pure RBF approximation (Davydov `et al.` (2005)).

    The two-stage approximation method employed by
    ``dim2_spline_ts_sctr`` is derived from the TSFIT package of O.
    Davydov and F. Zeilfelder.

    Values of the computed spline can subsequently be computed by
    calling ``dim2_spline_ts_evalv`` or ``dim2_spline_ts_evalm``.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_ts_sctr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02jdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._BYREF_INT, # Input, lsminp.
        utils._BYREF_INT, # Input, lsmaxp.
        utils._BYREF_INT, # Input, nxcels.
        utils._BYREF_INT, # Input, nycels.
        utils._BYREF_INT, # Input, lcoefs.
        utils._EngineFloat64ArrayType.c_type, # Output, coefs[lcoefs].
        utils._EngineIntArrayType.c_type, # Input/Output, iopts[liopts].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, opts[lopts].
        utils._BYREF_INT, # Input, liopts.
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _lsminp = utils._EngineIntScalarType(
        lsminp, {'fun_name': fun_name, 'entity': 'lsminp'},
    ).to_ctypes()
    _lsmaxp = utils._EngineIntScalarType(
        lsmaxp, {'fun_name': fun_name, 'entity': 'lsmaxp'},
    ).to_ctypes()
    _nxcels = utils._EngineIntScalarType(
        nxcels, {'fun_name': fun_name, 'entity': 'nxcels'},
    ).to_ctypes()
    _nycels = utils._EngineIntScalarType(
        nycels, {'fun_name': fun_name, 'entity': 'nycels'},
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('coefs',),
        req_keys=('iopts', 'opts'),
    )
    _nxcels_py = utils._EngineIntScalarType.to_py(_nxcels)
    _nycels_py = utils._EngineIntScalarType.to_py(_nycels)
    _vl_liopts_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (0,), validate=False)
    _vl_gsmoothness_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (11,), validate=False)
    if _vl_liopts_py >= 12:
        if _vl_gsmoothness_py == 1:
            _coefs_dim1 = (((_nxcels_py+2)*(_nycels_py+2)+1)//2)*10+1
        elif _vl_gsmoothness_py == 2:
            _coefs_dim1 = 28*(_nxcels_py+2)*(_nycels_py+2)*4+1
        else:
            _coefs_dim1 = 0
    else:
        _coefs_dim1 = 0
    _coefs = utils._EngineFloat64ArrayType.empty_ctypes(_coefs_dim1)
    comm['coefs'] = _coefs
    _lcoefs_py = utils._EngineFloat64ArrayType.get_shape(comm['coefs'], {'fun_name': fun_name, 'entity': 'comm[\'coefs\']'}, exp_rank=1, validate=False)[0]
    _lcoefs = utils._EngineIntScalarType(
        _lcoefs_py, {'fun_name': fun_name, 'entity': '_lcoefs'},
    ).to_ctypes()
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _f,
        _lsminp,
        _lsmaxp,
        _nxcels,
        _nycels,
        _lcoefs,
        _coefs,
        _iopts,
        _opts,
        _liopts,
        _lopts,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _lsminp_py = utils._EngineIntScalarType.to_py(_lsminp)
        _lsmaxp_py = utils._EngineIntScalarType.to_py(_lsmaxp)
        if _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_DATA_SIZE:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_MINP:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lsminp = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_lsminp_py),
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: 1 <= lsminp <= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_MAXP:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lsmaxp = {:s}.'.format(
                        utils._STR_FMT_INT.format(_lsmaxp_py),
                    ),
                    'Constraint: lsmaxp >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_GRIDX:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxcels = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nxcels_py),
                    ),
                    'Constraint: nxcels >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_GRIDY:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nycels = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nycels_py),
                    ),
                    'Constraint: nycels >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_INIT:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Option arrays are not initialized or are corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_SVD and _ifmt == _nag_e_consts.IERR_TSFIT_SVD:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected algorithmic failure was encountered. Please contact NAG.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_LSE:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected algorithmic failure was encountered. Please contact NAG.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_DEGENERATE:
            _ierr_fl = 12
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all elements of x or of y are equal.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_OPTION and _ifmt == _nag_e_consts.IFMT_TSFIT_RBF:
            _ierr_fl = 20
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The selected radial basis function cannot be used with the RBF local method.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_OPTION and _ifmt == _nag_e_consts.IFMT_TSFIT_DEGREE:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of option \'Polynomial Starting Degree\' was invalid.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_ts_evalv(nevalv, xevalv, yevalv, comm, fevalv):
    u"""
    Evaluation at a vector of points of a spline computed by
    ``dim2_spline_ts_sctr``.

    ``dim2_spline_ts_evalv`` calculates a vector of values of a spline
    computed by ``dim2_spline_ts_sctr``.

    For full information please refer to the NAG Library document for
    e02je

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02jef.html

    Parameters
    ----------
    nevalv : int
        n, the number of values at which the spline is to be evaluated.

    xevalv : float, shape (`nevalv`)
        The (x_i) values at which the spline is to be evaluated.

    yevalv : float, shape (`nevalv`)
        The (y_i) values at which the spline is to be evaluated.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``dim2_spline_ts_sctr`` and ``opt_set``.

    fevalv : float, shape (`nevalv`), modified in place
        `On exit`: if no exception or warning is raised on exit
        `fevalv`[i-1] contains the computed spline value at (x_i,y_i).

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `nevalv` = *<value>*.

            Constraint: `nevalv` >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 10)
            The fitting routine has not been called, or the array of
            coefficients has been corrupted.

        (`errno` 13)
            On entry, `xevalv`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `xevalv`[i-1] <= *<value>* for all
            i.

        (`errno` 14)
            On entry, `yevalv`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `yevalv`[i-1] <= *<value>* for all
            i.

    Notes
    -----
    ``dim2_spline_ts_evalv`` calculates values at prescribed points
    (x_i,y_i), for i = 1,2,...,n, of a bivariate spline computed by
    ``dim2_spline_ts_sctr``.
    It is derived from the TSFIT package of O. Davydov and F.
    Zeilfelder.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    Farin, G and Hansford, D, 2000, `The Essentials of CAGD`, Natic, MA:
    A K Peters, Ltd.

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_ts_evalv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02jeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nevalv.
        utils._EngineFloat64ArrayType.c_type, # Input, xevalv[nevalv].
        utils._EngineFloat64ArrayType.c_type, # Input, yevalv[nevalv].
        utils._BYREF_INT, # Input, lcoefs.
        utils._EngineFloat64ArrayType.c_type, # Input, coefs[lcoefs].
        utils._EngineFloat64ArrayType.c_type, # Output, fevalv[nevalv].
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, liopts.
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nevalv = utils._EngineIntScalarType(
        nevalv, {'fun_name': fun_name, 'entity': 'nevalv'},
    ).to_ctypes()
    _nevalv_py = utils._EngineIntScalarType.to_py(_nevalv)
    _xevalv = utils._EngineFloat64ArrayType(
        xevalv, {'fun_name': fun_name, 'entity': 'xevalv'},
        exp_shape=(_nevalv_py,),
    ).to_ctypes()
    _yevalv = utils._EngineFloat64ArrayType(
        yevalv, {'fun_name': fun_name, 'entity': 'yevalv'},
        exp_shape=(_nevalv_py,),
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('coefs', 'iopts', 'opts'),
    )
    _lcoefs_py = utils._EngineFloat64ArrayType.get_shape(comm['coefs'], {'fun_name': fun_name, 'entity': 'comm[\'coefs\']'}, exp_rank=1, validate=False)[0]
    _lcoefs = utils._EngineIntScalarType(
        _lcoefs_py, {'fun_name': fun_name, 'entity': '_lcoefs'},
    ).to_ctypes()
    _coefs = comm['coefs']
    _fevalv_cls = utils._EngineFloat64ArrayType(
        fevalv, {'fun_name': fun_name, 'entity': 'fevalv'},
        exp_shape=(_nevalv_py,),
        check_contents=False,
    )
    _fevalv = _fevalv_cls.to_ctypes(can_reference_contents=False)
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nevalv,
        _xevalv,
        _yevalv,
        _lcoefs,
        _coefs,
        _fevalv,
        _iopts,
        _opts,
        _liopts,
        _lopts,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_DATA_SIZE:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nevalv = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nevalv_py),
                    ),
                    'Constraint: nevalv >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_INIT:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Option arrays are not initialized or are corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_COEF:
            _ierr_fl = 10
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The fitting routine has not been called, or the array of coefficients has',
                    'been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_BOUNDS and _ifmt == _nag_e_consts.IFMT_TSFIT_X_BOUND:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xevalv[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'was outside the bounding box.',
                    'Constraint: {:s} <= xevalv[i-1] <= {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_BOUNDS and _ifmt == _nag_e_consts.IFMT_TSFIT_Y_BOUND:
            _ierr_fl = 14
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, yevalv[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'was outside the bounding box.',
                    'Constraint: {:s} <= yevalv[i-1] <= {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fevalv_cls.refresh_py(_fevalv)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_ts_evalm(nxeval, nyeval, xevalm, yevalm, comm, fevalm):
    u"""
    Evaluation at a mesh of points of a spline computed by
    ``dim2_spline_ts_sctr``.

    ``dim2_spline_ts_evalm`` calculates a mesh of values of a spline
    computed by ``dim2_spline_ts_sctr``.

    For full information please refer to the NAG Library document for
    e02jf

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02jff.html

    Parameters
    ----------
    nxeval : int
        n_x, the number of values in the x direction forming the mesh on
        which the spline is to be evaluated.

    nyeval : int
        n_y, the number of values in the y direction forming the mesh on
        which the spline is to be evaluated.

    xevalm : float, shape (`nxeval`)
        The (x_i) values forming the mesh on which the spline is to be
        evaluated.

    yevalm : float, shape (`nyeval`)
        The (y_j) values forming the mesh on which the spline is to be
        evaluated.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``dim2_spline_ts_sctr`` and ``opt_set``.

    fevalm : float, shape (`nxeval`, `nyeval`), modified in place
        `On exit`: if no exception or warning is raised on exit
        `fevalm`[i-1,j-1] contains the computed spline value at
        (x_i,y_j).

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, `nxeval` = *<value>*.

            Constraint: `nxeval` >= 1.

        (`errno` 7)
            On entry, `nyeval` = *<value>*.

            Constraint: `nyeval` >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 10)
            The fitting routine has not been called, or the array of
            coefficients has been corrupted.

        (`errno` 13)
            On entry, `xevalm`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `xevalm`[i-1] <= *<value>* for all
            i.

        (`errno` 14)
            On entry, `yevalm`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `yevalm`[j-1] <= *<value>* for all
            j.

    Notes
    -----
    ``dim2_spline_ts_evalm`` calculates values on a rectangular mesh of
    a bivariate spline computed by ``dim2_spline_ts_sctr``.
    The points in the mesh are defined by x coordinates (x_i), for i =
    1,2,...,n_x, and y coordinates (y_j), for j = 1,2,...,n_y.
    This function is derived from the TSFIT package of O. Davydov and F.
    Zeilfelder.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    Farin, G and Hansford, D, 2000, `The Essentials of CAGD`, Natic, MA:
    A K Peters, Ltd.
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_ts_evalm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02jfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nxeval.
        utils._BYREF_INT, # Input, nyeval.
        utils._EngineFloat64ArrayType.c_type, # Input, xevalm[nxeval].
        utils._EngineFloat64ArrayType.c_type, # Input, yevalm[nyeval].
        utils._BYREF_INT, # Input, lcoefs.
        utils._EngineFloat64ArrayType.c_type, # Input, coefs[lcoefs].
        utils._EngineFloat64ArrayType.c_type, # Output, fevalm[nxeval, nyeval].
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, liopts.
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((fevalm, 'fevalm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nxeval = utils._EngineIntScalarType(
        nxeval, {'fun_name': fun_name, 'entity': 'nxeval'},
    ).to_ctypes()
    _nyeval = utils._EngineIntScalarType(
        nyeval, {'fun_name': fun_name, 'entity': 'nyeval'},
    ).to_ctypes()
    _nxeval_py = utils._EngineIntScalarType.to_py(_nxeval)
    _xevalm = utils._EngineFloat64ArrayType(
        xevalm, {'fun_name': fun_name, 'entity': 'xevalm'},
        exp_shape=(_nxeval_py,),
    ).to_ctypes()
    _nyeval_py = utils._EngineIntScalarType.to_py(_nyeval)
    _yevalm = utils._EngineFloat64ArrayType(
        yevalm, {'fun_name': fun_name, 'entity': 'yevalm'},
        exp_shape=(_nyeval_py,),
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('coefs', 'iopts', 'opts'),
    )
    _lcoefs_py = utils._EngineFloat64ArrayType.get_shape(comm['coefs'], {'fun_name': fun_name, 'entity': 'comm[\'coefs\']'}, exp_rank=1, validate=False)[0]
    _lcoefs = utils._EngineIntScalarType(
        _lcoefs_py, {'fun_name': fun_name, 'entity': '_lcoefs'},
    ).to_ctypes()
    _coefs = comm['coefs']
    _fevalm_cls = utils._EngineFloat64ArrayType(
        fevalm, {'fun_name': fun_name, 'entity': 'fevalm'},
        exp_shape=(_nxeval_py, _nyeval_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _fevalm = _fevalm_cls.to_ctypes(can_reference_contents=False)
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nxeval,
        _nyeval,
        _xevalm,
        _yevalm,
        _lcoefs,
        _coefs,
        _fevalm,
        _iopts,
        _opts,
        _liopts,
        _lopts,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_GRIDX:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxeval = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nxeval_py),
                    ),
                    'Constraint: nxeval >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_CONSTRAINT and _ifmt == _nag_e_consts.IFMT_TSFIT_GRIDY:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nyeval = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nyeval_py),
                    ),
                    'Constraint: nyeval >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_INIT:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Option arrays are not initialized or are corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_COEF:
            _ierr_fl = 10
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The fitting routine has not been called, or the array of coefficients has',
                    'been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_BOUNDS and _ifmt == _nag_e_consts.IFMT_TSFIT_X_BOUND:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xevalm[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'was outside the bounding box.',
                    'Constraint: {:s} <= xevalm[i-1] <= {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_TSFIT_BOUNDS and _ifmt == _nag_e_consts.IFMT_TSFIT_Y_BOUND:
            _ierr_fl = 14
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, yevalm[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'was outside the bounding box.',
                    'Constraint: {:s} <= yevalm[j-1] <= {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'for all j.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fevalm_cls.refresh_py(_fevalm)
        return_data = None
    _err_class.manage()
    return return_data

def pade_app(ia, ib, c, ic, a, b):
    u"""
    Padé approximants.

    ``pade_app`` calculates the coefficients in a Padé approximant to a
    function from its user-supplied Maclaurin expansion.

    For full information please refer to the NAG Library document for
    e02ra

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02raf.html

    Parameters
    ----------
    ia : int
        `ia` must specify l+1 and `ib` must specify m+1, where l and m
        are the degrees of the numerator and denominator of the
        approximant, respectively.

    ib : int
        `ia` must specify l+1 and `ib` must specify m+1, where l and m
        are the degrees of the numerator and denominator of the
        approximant, respectively.

    c : float, shape (`ic`)
        `c`[i-1] must specify, for i = 1,2,...,l+m+1, the coefficient of
        x^(i-1) in the given power series.

    ic : int
        The dimension of the array `c`.

    a : float, shape (`ia`), modified in place
        `On exit`: `a`[j], for j = 1,2,...,l+1, contains the coefficient
        a_j in the numerator of the approximant.

    b : float, shape (`ib`), modified in place
        `On exit`: `b`[k], for k = 1,2,...,m+1, contains the coefficient
        b_k in the denominator of the approximant.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ic` = *<value>*, `ia` = *<value>* and `ib` =
            *<value>*.

            Constraint: `ic` >= `ia`+`ib`-1.

        (`errno` 1)
            On entry, `ib` = *<value>*.

            Constraint: `ib` >= 1.

        (`errno` 1)
            On entry, `ia` = *<value>*.

            Constraint: `ia` >= 1.

        (`errno` 2)
            The Pade approximant is degenerate.

    Notes
    -----
    Given a power series

        c_0+c_1 x+c_2 x^2+...+c_{l+m}x^(l+m)+...

    ``pade_app`` uses the coefficients c_i, for i = 0,1,...,l+m, to form
    the [l/m] Padé approximant of the form

        (a_0+a_1 x+a_2 x^2+...+a_l x^l)/(b_0+b_1 x+b_2 x^2+...+b_m x^m)

    with b_0 defined to be unity.
    The two sets of coefficients a_j, for j = 0,1,...,l, and b_k, for k
    = 0,1,...,m, in the numerator and denominator are calculated by
    direct solution of the Padé equations (see Graves--Morris (1979));
    these values are returned through the argument list unless the
    approximant is degenerate.

    Padé approximation is a useful technique when values of a function
    are to be obtained from its Maclaurin expansion but convergence of
    the series is unacceptably slow or even nonexistent.
    It is based on the hypothesis of the existence of a sequence of
    convergent rational approximations, as described in Baker and
    Graves--Morris (1981) and Graves--Morris (1979).

    Unless there are reasons to the contrary (as discussed in Module 4,
    Section 2, Modules 5 and 6 of Baker and Graves--Morris (1981)), one
    normally uses the diagonal sequence of Padé approximants, namely

        {[m/m],m = 0,1,2,...}.

    Subsequent evaluation of the approximant at a given value of x may
    be carried out using ``pade_eval``.

    References
    ----------
    Baker, G A Jr and Graves--Morris, P R, 1981, `Padé approximants,
    Part 1: Basic theory`, encyclopaedia of Mathematics and its
    Applications, Addison--Wesley

    Graves--Morris, P R, 1979, `The numerical calculation of Padé
    approximants`, Padé Approximation and its Applications. Lecture
    Notes in Mathematics, (ed L Wuytack) (765), 231--245, Adison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.pade_app'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02raft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, ia.
        utils._BYREF_INT, # Input, ib.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ic].
        utils._BYREF_INT, # Input, ic.
        utils._EngineFloat64ArrayType.c_type, # Output, a[ia].
        utils._EngineFloat64ArrayType.c_type, # Output, b[ib].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._BYREF_INT, # Input, jw.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _ia = utils._EngineIntScalarType(
        ia, {'fun_name': fun_name, 'entity': 'ia'},
    ).to_ctypes()
    _ib = utils._EngineIntScalarType(
        ib, {'fun_name': fun_name, 'entity': 'ib'},
    ).to_ctypes()
    _ic = utils._EngineIntScalarType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
    ).to_ctypes()
    _ia_py = utils._EngineIntScalarType.to_py(_ia)
    _ib_py = utils._EngineIntScalarType.to_py(_ib)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ia_py+_ib_py-1,),
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_ia_py,),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ib_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _jw = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _ia,
        _ib,
        _c,
        _ic,
        _a,
        _b,
        _w,
        _jw,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ic = {:s}, ia = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ib = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ic >= ia+ib-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ib = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ib >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ia = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ia >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The Pade approximant is degenerate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def pade_eval(a, ia, b, ib, x):
    u"""
    Evaluation of fitted rational function as computed by ``pade_app``.

    ``pade_eval`` evaluates a rational function at a user-supplied
    point, given the numerator and denominator coefficients.

    For full information please refer to the NAG Library document for
    e02rb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02rbf.html

    Parameters
    ----------
    a : float, shape (`ia`)
        `a`[j], for j = 1,2,...,l+1, must contain the value of the
        coefficient a_j in the numerator of the rational function.

    ia : int
        The value of l+1, where l is the degree of the numerator.

    b : float, shape (`ib`)
        `b`[k], for k = 1,2,...,m+1, must contain the value of the
        coefficient b_k in the denominator of the rational function.

    ib : int
        The value of m+1, where m is the degree of the denominator.

    x : float
        The point x at which the rational function is to be evaluated.

    Returns
    -------
    ans : float
        The result of evaluating the rational function at the given
        point x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Evaluation at or near a pole.

        (`errno` 2)
            The first `ib` entries in `b` are zero: `ib` = *<value>*.

        (`errno` 2)
            On entry, `ib` = *<value>*.

            Constraint: `ib` >= 1.

        (`errno` 2)
            On entry, `ia` = *<value>*.

            Constraint: `ia` >= 1.

    Notes
    -----
    Given a real value x and the coefficients a_j, for j = 0,1,...,l and
    b_k, for k = 0,1,...,m, ``pade_eval`` evaluates the rational
    function

        sum_j = 0^la_j x^j/(sum_k = 0^mb_k x^k).

    using nested multiplication (see Conte and de Boor (1965)).

    A particular use of ``pade_eval`` is to compute values of the Padé
    approximants determined by ``pade_app``.

    References
    ----------
    Conte, S D and de Boor, C, 1965, `Elementary Numerical Analysis`,
    McGraw--Hill

    Peters, G and Wilkinson, J H, 1971, `Practical problems arising in
    the solution of polynomial equations`, J. Inst. Maths. Applics. (8),
    16--35
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.pade_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02rbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, a[ia].
        utils._BYREF_INT, # Input, ia.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ib].
        utils._BYREF_INT, # Input, ib.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ans.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ia = utils._EngineIntScalarType(
        ia, {'fun_name': fun_name, 'entity': 'ia'},
    ).to_ctypes()
    _ia_py = utils._EngineIntScalarType.to_py(_ia)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_ia_py,),
    ).to_ctypes()
    _ib = utils._EngineIntScalarType(
        ib, {'fun_name': fun_name, 'entity': 'ib'},
    ).to_ctypes()
    _ib_py = utils._EngineIntScalarType.to_py(_ib)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ib_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ans = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _ia,
        _b,
        _ib,
        _x,
        _ans,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Evaluation at or near a pole.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The first ib entries in b are zero: ib = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ib = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ib >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ia = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ia >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ans_py = utils._EngineFloat64ScalarType.to_py(_ans)
        return_data = _ans_py
    _err_class.manage()
    return return_data

def dim2_spline_sort(px, py, lamda, mu, m, x, y, point, npoint):
    u"""
    Sort two-dimensional data into panels for fitting bicubic splines.

    ``dim2_spline_sort`` sorts two-dimensional data into rectangular
    panels.

    For full information please refer to the NAG Library document for
    e02za

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02zaf.html

    Parameters
    ----------
    px : int
        `px` and `py` must specify eight more than the number of
        intercepts on the x axis and y axis, respectively.

    py : int
        `px` and `py` must specify eight more than the number of
        intercepts on the x axis and y axis, respectively.

    lamda : float, shape (`px`)
        `lamda`[4] to `lamda`[`px`-5] must contain, in nondecreasing
        order, the intercepts on the x axis of the sides of the panels
        parallel to the y axis.

    mu : float, shape (`py`)
        `mu`[4] to `mu`[`py`-5] must contain, in nondecreasing order,
        the intercepts on the y axis of the sides of the panels parallel
        to the x axis.

    m : int
        The number m of data points.

    x : float, shape (`m`)
        The coordinates of the rth data point (x_r,y_r), for r =
        1,2,...,m.

    y : float, shape (`m`)
        The coordinates of the rth data point (x_r,y_r), for r =
        1,2,...,m.

    point : int, shape (`npoint`), modified in place
        `On exit`: for i = 1,2,...,`npoint`, `point`[m+i-1] = I1 is the
        index of the first point in panel i, `point`[I1-1] = I2 is the
        index of the second point in panel i and so on.

        `point`[In-1] = 0 indicates that `x`[In-1],`y`[In-1] was the
        last point in the panel.

        The coordinates of points in panel i can be accessed in turn by
        means of the following instructions:

        ::

          i_n = point[m+i-1]
          while i_n != 0:
              xi = x[i_n-1]
              yi = y[i_n-1]
              .
              .
              .
              i_n = point[i_n]
          ...

    npoint : int
        The dimension of the array `point`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, I = *<value>*, `mu`[I-1] = *<value>* and `mu`[I-2]
            = *<value>*.

            Constraint: `mu`[I-1] >= `mu`[I-2].

        (`errno` 1)
            On entry, I = *<value>*, `lamda`[I-1] = *<value>* and
            `lamda`[I-2] = *<value>*.

            Constraint: `lamda`[I-1] >= `lamda`[I-2].

        (`errno` 2)
            On entry, `npoint` = *<value>*, `m` = *<value>*, `px` =
            *<value>* and `py` = *<value>*.

            Constraint: `npoint` >= `m`+(`px`-7)*(`py`-7).

        (`errno` 2)
            On entry, `py` = *<value>*.

            Constraint: `py` >= 8.

        (`errno` 2)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 8.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

    Notes
    -----
    A set of m data points with rectangular Cartesian coordinates
    x_r,y_r are sorted into panels defined by lines parallel to the y
    and x axes.
    The intercepts of these lines on the x and y axes are given in
    `lamda`[i-1], for i = 5,6,...,`px`-4 and `mu`[j-1], for j =
    5,6,...,`py`-4, respectively.
    The function orders the data so that all points in a panel occur
    before data in succeeding panels, where the panels are numbered from
    bottom to top and then left to right, with the usual arrangement of
    axes, as shown in the diagram.
    Within a panel the points maintain their original order.

    [figure omitted]

    A data point lying exactly on one or more panel sides is taken to be
    in the highest-numbered panel adjacent to the point.
    The function does not physically rearrange the data, but provides
    the array `point` which contains a linked list for each panel,
    pointing to the data in that panel.
    The total number of panels is (`px`-7)*(`py`-7).
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.dim2_spline_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, px.
        utils._BYREF_INT, # Input, py.
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[px].
        utils._EngineFloat64ArrayType.c_type, # Input, mu[py].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineIntArrayType.c_type, # Output, point[npoint].
        utils._BYREF_INT, # Input, npoint.
        utils._EngineIntArrayType.c_type, # Workspace, adres.
        utils._BYREF_INT, # Input, nadres.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _px = utils._EngineIntScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineIntScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _px_py = utils._EngineIntScalarType.to_py(_px)
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_px_py,),
    ).to_ctypes()
    _py_py = utils._EngineIntScalarType.to_py(_py)
    _mu = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_py_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _npoint = utils._EngineIntScalarType(
        npoint, {'fun_name': fun_name, 'entity': 'npoint'},
    ).to_ctypes()
    _point_cls = utils._EngineIntArrayType(
        point, {'fun_name': fun_name, 'entity': 'point'},
        exp_shape=(_m_py+(_px_py-7)*(_py_py-7),),
        check_contents=False,
    )
    _point = _point_cls.to_ctypes(can_reference_contents=False)
    _adres = utils._EngineIntArrayType.empty_ctypes(0)
    _nadres = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _px,
        _py,
        _lamda,
        _mu,
        _m,
        _x,
        _y,
        _point,
        _npoint,
        _adres,
        _nadres,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, mu[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'mu[I-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: mu[I-1] >= mu[I-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, lamda[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and lamda[I-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lamda[I-1] >= lamda[I-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npoint = {:s}, m = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'px = {:s} and py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: npoint >= m+(px-7)*(py-7).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, py = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: py >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _point_cls.refresh_py(_point)
        return_data = None
    _err_class.manage()
    return return_data

def opt_set(optstr, comm):
    u"""
    Option setting routine.

    ``opt_set`` either initializes or resets the option arrays or sets a
    single option for supported problem solving functions in submodule
    ``fit``.
    Currently, only ``dim2_spline_ts_sctr`` is supported.

    For full information please refer to the NAG Library document for
    e02zk

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02zkf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option to be set.

        Initialize = function name

            Initialize the option arrays `comm`['iopts'] and
            `comm`['opts'] for use with function function name, where
            function name is the name of the problem solving function
            you wish to use.

        Defaults

            Resets all options to their default values.

        option = optval

            See [Other Parameters for dim2_spline_ts_sctr] for details
            of valid values for option and optval. The equals sign (=)
            delimiter must be used to separate the option from its
            optval.

        The processing of `optstr` does not depend on its case.

        Each token in the option and optval component must be separated
        by at least one space.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            `optstr` = *<value>*.

        (`errno` 14)
            On entry, attempting to initialize the option arrays but
            specified function name was not valid: name = *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            `optstr` = *<value>*.

        (`errno` 21)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    Notes
    -----
    ``opt_set`` has three purposes: to initialize option arrays, to
    reset all options to their default values or to set a single option
    to a user-supplied value.

    Options and their values are, in general, presented as a character
    string, `optstr`, of the form 'option = optval'; alphabetic
    characters can be supplied in either upper or lower case.
    Both option and optval may consist of one or more tokens separated
    by white space.
    The tokens that comprise optval will normally be either an integer,
    real or character value as defined in the description of the
    specific optional argument.
    In addition all options can take an optval 'DEFAULT' which resets
    the option to its default value.

    It is imperative that option arrays are initialized before any
    options are set, before the relevant problem solving function is
    called and before any options are queried using ``opt_get``.

    Information relating to available option names and their
    corresponding valid values is given in [Other Parameters for
    dim2_spline_ts_sctr].

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.opt_set'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02zkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._EngineIntArrayType.c_type, # Input/Output, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('iopts', 'opts'),
    )
    if _comm_needs_init:
        _iopts = utils._EngineIntArrayType.empty_ctypes(100)
        comm['iopts'] = _iopts
    else:
        _iopts = comm['iopts']
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    if _comm_needs_init:
        _opts = utils._EngineFloat64ArrayType.empty_ctypes(100)
        comm['opts'] = _opts
    else:
        _opts = comm['opts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == _nag_e_consts.IERR_OPT_UNKNOWN and _ifmt == _nag_e_consts.IFMT_OPT_UNKNOWN:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option in optstr was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NO_EQUAL and _ifmt == _nag_e_consts.IFMT_OPT_NO_EQUAL:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the expected delimiter \'=\' was not found in optstr:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NON_NUMERIC and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to an integer:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NON_NUMERIC and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to a real:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_RNAME and _ifmt == _nag_e_consts.IFMT_OPT_RNAME:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, attempting to initialize the option arrays but',
                    'specified function name was not valid:',
                    'name = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_IVALUE and _ifmt == _nag_e_consts.IERR_OPT_IVALUE:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the integer option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_RVALUE and _ifmt == _nag_e_consts.IERR_OPT_RVALUE:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the real option is not valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_CVALUE and _ifmt == _nag_e_consts.IFMT_OPT_CVALUE_NOREC:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the character option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_NOT_INIT and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_INIT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_LINKED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT and _ifmt == _nag_e_consts.IFMT_OPTS_CORRUPT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def opt_get(optstr, comm):
    u"""
    Option getting routine.

    ``opt_get`` is used to query the value of options available to
    supported problem solving functions in submodule ``fit``.
    Currently, only ``dim2_spline_ts_sctr`` is supported.

    For full information please refer to the NAG Library document for
    e02zl

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/e02/e02zlf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option whose current value is required.
        See [Other Parameters for dim2_spline_ts_sctr] for information
        on valid options. In addition, the following is a valid option:

        Identify

            ``opt_get`` returns in `cvalue` the function name supplied
            to ``opt_set`` when the option arrays `comm`['iopts'] and
            `comm`['opts'] were initialized.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt_set``.

    Returns
    -------
    ivalue : int
        If the option supplied in `optstr` is an integer valued
        parameter, `ivalue` will hold its current value.

    rvalue : float
        If the option supplied in `optstr` is a real valued parameter,
        `rvalue` will hold its current value.

    cvalue : str
        If the option supplied in `optstr` is a character valued
        parameter, `cvalue` will hold its current value, unless Identify
        is specified (see `optstr`).

    optype : int
        Indicates whether the option supplied in `optstr` is an integer,
        real or character valued parameter and hence which of `ivalue`,
        `rvalue` or `cvalue` holds the current value.

        `optype` = 1

            `optstr` is an integer valued option, its current value has
            been returned in `ivalue`.

        `optype` = 2

            `optstr` is a real valued option, its current value has been
            returned in `rvalue`.

        `optype` = 3

            `optstr` is a character valued option, its current value has
            been returned in `cvalue`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 61)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    Notes
    -----
    ``opt_get`` is used to query the current values of options.
    It is necessary to initalize option arrays using ``opt_set`` before
    any options are queried.

    ``opt_get`` will normally return either an integer, real or
    character value dependent upon the type associated with the option
    being queried.
    Whether the option queried is of integer, real or character type is
    indicated by the returned value of `optype`.

    Information on option names and whether these options are real,
    integer or character can be found in [Other Parameters for
    dim2_spline_ts_sctr].
    """

    excs = []
    fun_name = 'naginterfaces.base.fit.opt_get'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e02zlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_INT, # Output, ivalue.
        utils._BYREF_FLOAT64, # Output, rvalue.
        utils._BYREF_STRING, # Output, cvalue(len=*).
        utils._BYREF_INT, # Output, optype.
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cvalue.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _ivalue = utils._EngineIntScalarType.empty_ctypes()
    _rvalue = utils._EngineFloat64ScalarType.empty_ctypes()
    _cvalue = utils._EngineCharScalarType.empty_ctypes(chrlen=39)
    _optype = utils._EngineIntScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iopts', 'opts'),
    )
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _ivalue,
        _rvalue,
        _cvalue,
        _optype,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _cvalue._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == _nag_e_consts.IERR_OPT_UNKNOWN and _ifmt == _nag_e_consts.IFMT_OPT_UNKNOWN:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option in optstr was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPTS_CORRUPT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ivalue_py = utils._EngineIntScalarType.to_py(_ivalue)
        _rvalue_py = utils._EngineFloat64ScalarType.to_py(_rvalue)
        _optype_py = utils._EngineIntScalarType.to_py(_optype)
        if _optype_py == _nag_e_consts.OPT_TYPE_IDENTIFY:
            _optype_py = _nag_e_consts.OPT_TYPE_CHAR
            if _ivalue_py == _nag_e_consts.TSFIT_SOLVER_ID:
                _cvalue_py = 'e02jd'
            else:
                utils._EngineErrorExit(
                    fun_name=fun_name,
                    msg_lines=[
                        'The communication structure comm has either not been initialized,',
                        'has become corrupted, or is not compatible with this option',
                        'setting function.',
                    ],
                    errcodes=(_nag_e_consts.IERR_OPTS_CORRUPT_GETTER,),
                ).manage()
            _ivalue_py = 0
        else:
            _cvalue_py = utils._EngineCharScalarType.to_py(_cvalue)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ivalue', 'rvalue', 'cvalue', 'optype'],
        )(
            _ivalue_py,
            _rvalue_py,
            _cvalue_py,
            _optype_py,
        )
    _err_class.manage()
    return return_data
