# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.6 `linsys` Chapter.

``linsys`` - Simultaneous Linear Equations

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the solution of the matrix equation AX =
B, where B may be a single vector or a matrix of multiple right-hand
sides.
The matrix A may be real, complex, symmetric, Hermitian, positive
definite, positive definite Toeplitz or banded.
It may also be rectangular, in which case a least squares solution is
obtained.

Much of the functionality of this module has been superseded by
functions from submodule ``lapacklin`` and submodule ``lapackeig``
(LAPACK routines) as those modules have grown and have included driver
and expert driver functions.

For a general introduction to sparse systems of equations, see [the F11
Introduction], which provides functions for large sparse systems. Some
functions for sparse problems are also included in this module; they are
described in [Sparse Matrix Functions].

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04intro.html
"""

# NAG Copyright 2017-2022.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def real_gen_lsqsol(a, x, b, m, n, ir, eps, qr, alpha, ipiv):
    u"""
    Least squares solution of m real equations in n unknowns, rank = n,
    m >= n using iterative refinement (Black Box).

    ``real_gen_lsqsol`` calculates the accurate least squares solution
    of a set of m linear equations in n unknowns, m >= n and rank = n,
    with multiple right-hand sides, AX = B, using a QR factorization and
    iterative refinement.

    For full information please refer to the NAG Library document for
    f04am

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04amf.html

    Parameters
    ----------
    a : float, shape (`m`, `n`)
        The m*n matrix A.

    x : float, shape (`n`, `ir`), modified in place
        `On exit`: the n*r solution matrix X.

    b : float, shape (`m`, `ir`)
        The m*r right-hand side matrix B.

    m : int
        m, the number of rows of the matrix A, i.e., the number of
        equations.

    n : int
        n, the number of columns of the matrix A, i.e., the number of
        unknowns.

    ir : int
        r, the number of right-hand sides.

    eps : float
        Must be set to the value of the machine precision.

    qr : float, shape (`m`, `n`), modified in place
        `On exit`: details of the QR factorization.

    alpha : float, shape (`n`), modified in place
        `On exit`: the diagonal elements of the upper triangular matrix
        R.

    ipiv : int, shape (`n`), modified in place
        `On exit`: details of the column interchanges.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The rank of `a` is less than `n`. The problem does not have
            a unique solution.

        (`errno` 2)
            The iterative refinement fails to converge. The matrix A is
            too ill-conditioned.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    To compute the least squares solution to a set of m linear equations
    in n unknowns (m >= n) AX = B, ``real_gen_lsqsol`` first computes a
    QR factorization of A with column pivoting, AP = QR, where R is
    upper triangular, Q is an m*m orthogonal matrix, and P is a
    permutation matrix. Q^T is applied to the m*r right-hand side matrix
    B to give C = Q^TB, and the n*r solution matrix X is calculated, to
    a first approximation, by back-substitution in RX = C.
    The residual matrix S = B-AX is calculated using additional
    precision, and a correction D to X is computed as the least squares
    solution to AD = S. X is replaced by X+D and this iterative
    refinement of the solution is repeated until full machine accuracy
    has been obtained.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_gen_lsqsol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04amft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, ir].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, ir].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ir.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Output, qr[ldqr, n].
        utils._BYREF_INT, # Input, ldqr.
        utils._EngineFloat64ArrayType.c_type, # Output, alpha[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, e.
        utils._EngineFloat64ArrayType.c_type, # Workspace, y.
        utils._EngineFloat64ArrayType.c_type, # Workspace, z.
        utils._EngineFloat64ArrayType.c_type, # Workspace, r.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (x, 'x'), (b, 'b'), (qr, 'qr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _ir = utils._EngineIntScalarType(
        ir, {'fun_name': fun_name, 'entity': 'ir'},
    ).to_ctypes()
    _ir_py = utils._EngineIntScalarType.to_py(_ir)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _ir_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _ir_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _ldqr_py = utils._EngineFloat64ArrayType.get_nd(qr, {'fun_name': fun_name, 'entity': 'qr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldqr = utils._EngineIntScalarType(
        _ldqr_py, {'fun_name': fun_name, 'entity': '_ldqr'},
    ).to_ctypes()
    _qr_cls = utils._EngineFloat64ArrayType(
        qr, {'fun_name': fun_name, 'entity': 'qr'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _qr = _qr_cls.to_ctypes(can_reference_contents=False)
    _alpha_cls = utils._EngineFloat64ArrayType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _alpha = _alpha_cls.to_ctypes(can_reference_contents=False)
    _e = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _y = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _z = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _lda,
        _x,
        _ldx,
        _b,
        _ldb,
        _m,
        _n,
        _ir,
        _eps,
        _qr,
        _ldqr,
        _alpha,
        _e,
        _y,
        _z,
        _r,
        _ipiv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The rank of a is less than n.',
                    'The problem does not have a unique solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative refinement fails to converge.',
                    'The matrix A is too ill-conditioned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _qr_cls.refresh_py(_qr)
        _alpha_cls.refresh_py(_alpha)
        _ipiv_cls.refresh_py(_ipiv)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_sparse_fac_solve(n, a, licn, comm, rhs, mtype):
    u"""
    Solution of real sparse simultaneous linear equations (coefficient
    matrix already factorized).

    ``real_sparse_fac_solve`` calculates the approximate solution of a
    set of real sparse linear equations with a single right-hand side,
    Ax = b or A^Tx = b, where A has been factorized by
    ``matop.real_gen_sparse_lu`` or ``matop.real_gen_sparse_lu_reuse``.

    For full information please refer to the NAG Library document for
    f04ax

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04axf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, shape (`licn`)
        The nonzero elements in the factorization of the matrix A, as
        returned by ``matop.real_gen_sparse_lu`` or
        ``matop.real_gen_sparse_lu_reuse``.

    licn : int
        The dimension of the arrays `a` and `comm`['icn'].

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``matop.real_gen_sparse_lu``.

    rhs : float, shape (`n`), modified in place
        `On entry`: the right-hand side vector b.

        `On exit`: `rhs` is overwritten by the solution vector x.

    mtype : int
        `mtype` specifies the task to be performed.

        `mtype` = 1

            Solve Ax = b.

        `mtype` != 1

            Solve A^Tx = b.

    Returns
    -------
    resid : float
        The value of the maximum residual, max(|b_i-sum_j^a_{ij}x_j|),
        over all the unsatisfied equations, in case
        ``matop.real_gen_sparse_lu`` or
        ``matop.real_gen_sparse_lu_reuse`` has been used to factorize a
        singular or rectangular matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    To solve a system of real linear equations Ax = b or A^Tx = b, where
    A is a general sparse matrix, A must first be factorized by
    ``matop.real_gen_sparse_lu`` or ``matop.real_gen_sparse_lu_reuse``.
    ``real_sparse_fac_solve`` then computes x by block forward or
    backward substitution using simple forward and backward substitution
    within each diagonal block.

    The method is fully described in Duff (1977).

    A more recent method is available through solver function
    ``sparse.direct_real_gen_solve`` and factorization function
    ``sparse.direct_real_gen_lu``.

    References
    ----------
    Duff, I S, 1977, `MA28 -- a set of Fortran subroutines for sparse
    unsymmetric linear equations`, AERE Report R8730, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_sparse_fac_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04axft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[licn].
        utils._BYREF_INT, # Input, licn.
        utils._EngineIntArrayType.c_type, # Input, icn[licn].
        utils._EngineIntArrayType.c_type, # Input/Output, ikeep[5*n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rhs[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._BYREF_INT, # Input, mtype.
        utils._EngineIntArrayType.c_type, # Input, idisp[2].
        utils._BYREF_FLOAT64, # Output, resid.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _licn = utils._EngineIntScalarType(
        licn, {'fun_name': fun_name, 'entity': 'licn'},
    ).to_ctypes()
    _licn_py = utils._EngineIntScalarType.to_py(_licn)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_licn_py,),
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('icn', 'ikeep', 'idisp'),
    )
    _icn = comm['icn']
    _ikeep = comm['ikeep']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rhs_cls = utils._EngineFloat64ArrayType(
        rhs, {'fun_name': fun_name, 'entity': 'rhs'},
        exp_shape=(_n_py,),
    )
    _rhs = _rhs_cls.to_ctypes()
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mtype = utils._EngineIntScalarType(
        mtype, {'fun_name': fun_name, 'entity': 'mtype'},
    ).to_ctypes()
    _idisp = comm['idisp']
    _resid = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _licn,
        _icn,
        _ikeep,
        _rhs,
        _w,
        _mtype,
        _idisp,
        _resid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rhs_cls.refresh_py(_rhs)
        _resid_py = utils._EngineFloat64ScalarType.to_py(_resid)
        return_data = _resid_py
    _err_class.manage()
    return return_data

def real_square_solve(n, nrhs, a, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real system of linear equations.

    ``real_square_solve`` computes the solution to a real system of
    linear equations AX = B, where A is an n*n matrix and X and B are
    n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04ba

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04baf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : float, shape (`n`, `n`), modified in place
        `On entry`: the n*n coefficient matrix A.

        `On exit`: if no exception is raised, the factors L and U from
        the factorization A = PLU. The unit diagonal elements of L are
        not stored.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i
        indicates a row interchange was not required.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    unit lower triangular, and U is upper triangular.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_square_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _a,
        _lda,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_band_solve(n, kl, ku, nrhs, ab, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real banded system of linear equations.

    ``real_band_solve`` computes the solution to a real system of linear
    equations AX = B, where A is an n*n band matrix, with k_l
    subdiagonals and k_u superdiagonals, and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bbf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    kl : int
        The number of subdiagonals k_l, within the band of A.

    ku : int
        The number of superdiagonals k_u, within the band of A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ab : float, shape (2*`kl`+`ku`+1, `n`), modified in place
        `On entry`: the n*n matrix A.

        See [Further Comments] for further details.

        `On exit`: if no exception is raised, `ab` is overwritten by
        details of the factorization.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i
        indicates a row interchange was not required.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -6)
            On entry, ldab = *<value>*, `kl` = *<value>* and `ku` =
            *<value>*.

            Constraint: ldab >= 2*`kl`+`ku`+1.

        (`errno` -4)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -3)
            On entry, `ku` = *<value>*.

            Constraint: `ku` >= 0.

        (`errno` -2)
            On entry, `kl` = *<value>*.

            Constraint: `kl` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    the product of permutation matrices and unit lower triangular
    matrices with k_l subdiagonals, and U is upper triangular with
    (k_l+k_u) superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_band_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab_cls = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(2*_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    )
    _ab = _ab_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _kl,
        _ku,
        _nrhs,
        _ab,
        _ldab,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -6 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldab = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'kl = {:s} and ku = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ldab >= 2*kl+ku+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ku = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ku >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kl >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 8:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ab_cls.refresh_py(_ab)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_tridiag_solve(n, nrhs, dl, d, du, du2, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real tridiagonal system of linear equations.

    ``real_tridiag_solve`` computes the solution to a real system of
    linear equations AX = B, where A is an n*n tridiagonal matrix and X
    and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bcf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    dl : float, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) subdiagonal elements of the
        matrix A.

        `On exit`: if no exception is raised, `dl` is overwritten by the
        (n-1) multipliers that define the matrix L from the LU
        factorization of A.

    d : float, shape (`n`), modified in place
        `On entry`: must contain the n diagonal elements of the matrix
        A.

        `On exit`: if no exception is raised, `d` is overwritten by the
        n diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : float, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) superdiagonal elements of the
        matrix A

        `On exit`: if no exception is raised, `du` is overwritten by the
        (n-1) elements of the first superdiagonal of U.

    du2 : float, shape (`n`-2), modified in place
        `On exit`: if no exception is raised, `du2` returns the (n-2)
        elements of the second superdiagonal of U.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] will
        always be either i or (i+1); `ipiv`[i-1] = i indicates a row
        interchange was not required.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    unit lower triangular with at most one nonzero subdiagonal element,
    and U is an upper triangular band matrix with two superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    Note that the equations A^TX = B may be solved by interchanging the
    order of the arguments `du` and `dl`.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_tridiag_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, dl[n-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, d[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, du[n-1].
        utils._EngineFloat64ArrayType.c_type, # Output, du2[n-2].
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _dl_cls = utils._EngineFloat64ArrayType(
        dl, {'fun_name': fun_name, 'entity': 'dl'},
        exp_shape=(_n_py-1,),
    )
    _dl = _dl_cls.to_ctypes()
    _d_cls = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    )
    _d = _d_cls.to_ctypes()
    _du_cls = utils._EngineFloat64ArrayType(
        du, {'fun_name': fun_name, 'entity': 'du'},
        exp_shape=(_n_py-1,),
    )
    _du = _du_cls.to_ctypes()
    _du2_cls = utils._EngineFloat64ArrayType(
        du2, {'fun_name': fun_name, 'entity': 'du2'},
        exp_shape=(_n_py-2,),
        check_contents=False,
    )
    _du2 = _du2_cls.to_ctypes(can_reference_contents=False)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py == _n_py and _ifmt == 4:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 5:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _dl_cls.refresh_py(_dl)
        _d_cls.refresh_py(_d)
        _du_cls.refresh_py(_du)
        _du2_cls.refresh_py(_du2)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_posdef_solve(uplo, n, nrhs, a, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric positive definite system of linear equations.

    ``real_posdef_solve`` computes the solution to a real system of
    linear equations AX = B, where A is an n*n symmetric positive
    definite matrix and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bdf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : float, shape (`n`, `n`), modified in place
        `On entry`: the n*n symmetric matrix A.

        If `uplo` = 'U', the leading `n` by `n` upper triangular part of
        `a` contains the upper triangular part of the matrix A, and the
        strictly lower triangular part of `a` is not referenced.

        If `uplo` = 'L', the leading `n` by `n` lower triangular part of
        `a` contains the lower triangular part of the matrix A, and the
        strictly upper triangular part of `a` is not referenced.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^TU or
        A = LL^T.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^TU, if
    `uplo` = 'U', or A = LL^T, if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_posdef_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _a,
        _lda,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo != \'U\' or \'L\': uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_posdef_packed_solve(uplo, n, nrhs, ap, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric positive definite system of linear equations,
    packed storage.

    ``real_posdef_packed_solve`` computes the solution to a real system
    of linear equations AX = B, where A is an n*n symmetric positive
    definite matrix, stored in packed format, and X and B are n*r
    matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04be

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bef.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ap : float, shape (`n`*(`n`+1)/2), modified in place
        `On entry`: the n*n symmetric matrix A. The upper or lower
        triangular part of the symmetric matrix is packed column-wise in
        a linear array. The jth column of A is stored in the array `ap`
        as follows:

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^TU or
        A = LL^T, in the same storage format as A.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^TU, if
    `uplo` = 'U', or A = LL^T, if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_posdef_packed_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ap_cls = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _ap,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ap_cls.refresh_py(_ap)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_posdef_band_solve(uplo, n, kd, nrhs, ab, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric positive definite banded system of linear
    equations.

    ``real_posdef_band_solve`` computes the solution to a real system of
    linear equations AX = B, where A is an n*n symmetric positive
    definite band matrix of band width 2k+1, and X and B are n*r
    matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bf

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bff.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    kd : int
        The number of superdiagonals k (and the number of subdiagonals)
        of the band matrix A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ab : float, shape (`kd`+1, `n`), modified in place
        `On entry`: the n*n symmetric band matrix A. The upper or lower
        triangular part of the symmetric matrix is stored in the first
        `kd`+1 rows of the array. The jth column of A is stored in the
        jth column of the array `ab` as follows:

        See [Further Comments] below for further details.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^TU or
        A = LL^T, in the same storage format as A.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -3)
            On entry, `kd` = *<value>*.

            Constraint: `kd` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^TU, if
    `uplo` = 'U', or A = LL^T, if `uplo` = 'L', where U is an upper
    triangular band matrix with k superdiagonals, and L is a lower
    triangular band matrix with k subdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_posdef_band_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kd.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kd = utils._EngineIntScalarType(
        kd, {'fun_name': fun_name, 'entity': 'kd'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kd_py = utils._EngineIntScalarType.to_py(_kd)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab_cls = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kd_py+1, _n_py),
        mutual_sorder=_m_sorder,
    )
    _ab = _ab_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _kd,
        _nrhs,
        _ab,
        _ldab,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kd = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kd >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 8:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ab_cls.refresh_py(_ab)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_posdef_tridiag_solve(n, nrhs, d, e, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric positive definite tridiagonal system of linear
    equations.

    ``real_posdef_tridiag_solve`` computes the solution to a real system
    of linear equations AX = B, where A is an n*n symmetric positive
    definite tridiagonal matrix and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bg

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bgf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    d : float, shape (`n`), modified in place
        `On entry`: must contain the n diagonal elements of the
        tridiagonal matrix A.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, `d` is overwritten by the n diagonal elements of the diagonal
        matrix D from the LDL^T factorization of A.

    e : float, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) subdiagonal elements of the
        tridiagonal matrix A.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, `e` is overwritten by the (n-1) subdiagonal elements of the
        unit lower bidiagonal matrix L from the LDL^T factorization of
        A. (`e` can also be regarded as the superdiagonal of the unit
        upper bidiagonal factor U from the U^TDU factorization of A.)

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    A is factorized as A = LDL^T, where L is a unit lower bidiagonal
    matrix and D is diagonal, and the factored form of A is then used to
    solve the system of equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_posdef_tridiag_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, d[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, e[n-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _d_cls = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    )
    _d = _d_cls.to_ctypes()
    _e_cls = utils._EngineFloat64ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_n_py-1,),
    )
    _e = _e_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _d,
        _e,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 5:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _d_cls.refresh_py(_d)
        _e_cls.refresh_py(_e)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_symm_solve(uplo, n, nrhs, a, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric system of linear equations.

    ``real_symm_solve`` computes the solution to a real system of linear
    equations AX = B, where A is an n*n symmetric matrix and X and B are
    n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bh

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bhf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : float, shape (`n`, `n`), modified in place
        `On entry`: the n*n symmetric matrix A.

        If `uplo` = 'U', the leading `n` by `n` upper triangular part of
        the array `a` contains the upper triangular part of the matrix
        A, and the strictly lower triangular part of `a` is not
        referenced.

        If `uplo` = 'L', the leading `n` by `n` lower triangular part of
        the array `a` contains the lower triangular part of the matrix
        A, and the strictly upper triangular part of `a` is not
        referenced.

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by
        ``lapacklin.dsytrf``.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.dsytrf``.

        `ipiv`[k-1] > 0

            Rows and columns k and `ipiv`[k-1] were interchanged, and
            d_{kk} is a 1*1 diagonal block.

        `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0

            Rows and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block.

        `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0

            Rows and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^T, if
    `uplo` = 'U', or A = LDL^T, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_symm_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _a,
        _lda,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_symm_packed_solve(uplo, n, nrhs, ap, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a real symmetric system of linear equations, packed storage.

    ``real_symm_packed_solve`` computes the solution to a real system of
    linear equations AX = B, where A is an n*n symmetric matrix, stored
    in packed format and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04bj

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04bjf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ap : float, shape (`n`*(`n`+1)/2), modified in place
        `On entry`: the n*n symmetric matrix A, packed column-wise in a
        linear array. The jth column of the matrix A is stored in the
        array `ap` as follows:

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by
        ``lapacklin.dsptrf``, stored as a packed triangular matrix in
        the same storage format as A.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.dsptrf``.

            If `ipiv`[k-1] > 0, then rows and columns k and `ipiv`[k-1]
            were interchanged, and d_{kk} is a 1*1 diagonal block;

            if `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0, then rows
            and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block;

            if `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0, then rows
            and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : float, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^T, if
    `uplo` = 'U', or A = LDL^T, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_symm_packed_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04bjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ap_cls = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ap_cls.refresh_py(_ap)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_square_solve(n, nrhs, a, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex system of linear equations.

    ``complex_square_solve`` computes the solution to a complex system
    of linear equations AX = B, where A is an n*n matrix and X and B are
    n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04ca

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04caf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : complex, shape (`n`, `n`), modified in place
        `On entry`: the n*n coefficient matrix A.

        `On exit`: if no exception is raised, the factors L and U from
        the factorization A = PLU. The unit diagonal elements of L are
        not stored.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i
        indicates a row interchange was not required.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    unit lower triangular, and U is upper triangular.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_square_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _a,
        _lda,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_band_solve(n, kl, ku, nrhs, ab, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex banded system of linear equations.

    ``complex_band_solve`` computes the solution to a complex system of
    linear equations AX = B, where A is an n*n band matrix, with k_l
    subdiagonals and k_u superdiagonals, and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cb

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cbf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    kl : int
        The number of subdiagonals k_l, within the band of A.

    ku : int
        The number of superdiagonals k_u, within the band of A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ab : complex, shape (2*`kl`+`ku`+1, `n`), modified in place
        `On entry`: the n*n matrix A.

        See [Further Comments] for further details.

        `On exit`: if no exception is raised, `ab` is overwritten by
        details of the factorization.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i
        indicates a row interchange was not required.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no exception is raised, an estimate of the reciprocal of the
        condition number of the matrix A, computed as `rcond` =
        (||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -6)
            On entry, ldab = *<value>*, `kl` = *<value>* and `ku` =
            *<value>*.

            Constraint: ldab >= 2*`kl`+`ku`+1.

        (`errno` -4)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -3)
            On entry, `ku` = *<value>*.

            Constraint: `ku` >= 0.

        (`errno` -2)
            On entry, `kl` = *<value>*.

            Constraint: `kl` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    the product of permutation matrices and unit lower triangular
    matrices with k_l subdiagonals, and U is upper triangular with
    (k_l+k_u) superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_band_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab_cls = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(2*_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    )
    _ab = _ab_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _kl,
        _ku,
        _nrhs,
        _ab,
        _ldab,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -6 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldab = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'kl = {:s} and ku = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ldab >= 2*kl+ku+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ku = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ku >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kl >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 8:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ab_cls.refresh_py(_ab)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_tridiag_solve(n, nrhs, dl, d, du, du2, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex tridiagonal system of linear equations.

    ``complex_tridiag_solve`` computes the solution to a complex system
    of linear equations AX = B, where A is an n*n tridiagonal matrix and
    X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04ccf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    dl : complex, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) subdiagonal elements of the
        matrix A.

        `On exit`: if no exception is raised, `dl` is overwritten by the
        (n-1) multipliers that define the matrix L from the LU
        factorization of A.

    d : complex, shape (`n`), modified in place
        `On entry`: must contain the n diagonal elements of the matrix
        A.

        `On exit`: if no exception is raised, `d` is overwritten by the
        n diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : complex, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) superdiagonal elements of the
        matrix A

        `On exit`: if no exception is raised, `du` is overwritten by the
        (n-1) elements of the first superdiagonal of U.

    du2 : complex, shape (`n`-2), modified in place
        `On exit`: if no exception is raised, `du2` returns the (n-2)
        elements of the second superdiagonal of U.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, the pivot indices that
        define the permutation matrix P; at the ith step row i of the
        matrix was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] will
        always be either i or (i+1); `ipiv`[i-1] = i indicates a row
        interchange was not required.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal element *<value>* of the upper triangular factor is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The LU decomposition with partial pivoting and row interchanges is
    used to factor A as A = PLU, where P is a permutation matrix, L is
    unit lower triangular with at most one nonzero subdiagonal element,
    and U is an upper triangular band matrix with two superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    Note that the equations A^TX = B may be solved by interchanging the
    order of the arguments `du` and `dl`.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_tridiag_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, dl[n-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, d[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, du[n-1].
        utils._EngineComplex128ArrayType.c_type, # Output, du2[n-2].
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _dl_cls = utils._EngineComplex128ArrayType(
        dl, {'fun_name': fun_name, 'entity': 'dl'},
        exp_shape=(_n_py-1,),
    )
    _dl = _dl_cls.to_ctypes()
    _d_cls = utils._EngineComplex128ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    )
    _d = _d_cls.to_ctypes()
    _du_cls = utils._EngineComplex128ArrayType(
        du, {'fun_name': fun_name, 'entity': 'du'},
        exp_shape=(_n_py-1,),
    )
    _du = _du_cls.to_ctypes()
    _du2_cls = utils._EngineComplex128ArrayType(
        du2, {'fun_name': fun_name, 'entity': 'du2'},
        exp_shape=(_n_py-2,),
        check_contents=False,
    )
    _du2 = _du2_cls.to_ctypes(can_reference_contents=False)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal element {:s} of the upper triangular factor is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero. The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 5:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _dl_cls.refresh_py(_dl)
        _d_cls.refresh_py(_d)
        _du_cls.refresh_py(_du)
        _du2_cls.refresh_py(_du2)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_posdef_solve(uplo, n, nrhs, a, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex Hermitian positive definite system of linear equations.

    ``complex_posdef_solve`` computes the solution to a complex system
    of linear equations AX = B, where A is an n*n Hermitian positive
    definite matrix and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cdf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : complex, shape (`n`, `n`), modified in place
        `On entry`: the n*n Hermitian matrix A.

        If `uplo` = 'U', the leading `n` by `n` upper triangular part of
        `a` contains the upper triangular part of the matrix A, and the
        strictly lower triangular part of `a` is not referenced.

        If `uplo` = 'L', the leading `n` by `n` lower triangular part of
        `a` contains the lower triangular part of the matrix A, and the
        strictly upper triangular part of `a` is not referenced.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^HU or
        A = LL^H.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^HU, if
    `uplo` = 'U', or A = LL^H, if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_posdef_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _a,
        _lda,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_posdef_packed_solve(uplo, n, nrhs, ap, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex Hermitian positive definite system of linear equations,
    packed storage.

    ``complex_posdef_packed_solve`` computes the solution to a complex
    system of linear equations AX = B, where A is an n*n Hermitian
    positive definite matrix, stored in packed format, and X and B are
    n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04ce

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cef.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        `On entry`: the n*n Hermitian matrix A. The upper or lower
        triangular part of the Hermitian matrix is packed column-wise in
        a linear array. The jth column of A is stored in the array `ap`
        as follows:

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^HU or
        A = LL^H, in the same storage format as A.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^HU, if
    `uplo` = 'U', or A = LL^H, if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_posdef_packed_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _ap,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo != \'U\' or \'L\': uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ap_cls.refresh_py(_ap)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_posdef_band_solve(uplo, n, kd, nrhs, ab, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex Hermitian positive definite banded system of linear
    equations.

    ``complex_posdef_band_solve`` computes the solution to a complex
    system of linear equations AX = B, where A is an n*n Hermitian
    positive definite band matrix of band width 2k+1, and X and B are
    n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cf

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cff.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    kd : int
        The number of superdiagonals k (and the number of subdiagonals)
        of the band matrix A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ab : complex, shape (`kd`+1, `n`), modified in place
        `On entry`: the n*n Hermitian band matrix A. The upper or lower
        triangular part of the Hermitian matrix is stored in the first
        `kd`+1 rows of the array. The jth column of A is stored in the
        jth column of the array `ab` as follows:

        See [Further Comments] below for further details.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the factor U or L from the Cholesky factorization A = U^HU or
        A = LL^H, in the same storage format as A.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -3)
            On entry, `kd` = *<value>*.

            Constraint: `kd` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The Cholesky factorization is used to factor A as A = U^HU, if
    `uplo` = 'U', or A = LL^H, if `uplo` = 'L', where U is an upper
    triangular band matrix with k superdiagonals, and L is a lower
    triangular band matrix with k subdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_posdef_band_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04cfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kd.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kd = utils._EngineIntScalarType(
        kd, {'fun_name': fun_name, 'entity': 'kd'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kd_py = utils._EngineIntScalarType.to_py(_kd)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab_cls = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kd_py+1, _n_py),
        mutual_sorder=_m_sorder,
    )
    _ab = _ab_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _kd,
        _nrhs,
        _ab,
        _ldab,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kd = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kd >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 8:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ab_cls.refresh_py(_ab)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_posdef_tridiag_solve(n, nrhs, d, e, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex Hermitian positive definite tridiagonal system of linear
    equations.

    ``complex_posdef_tridiag_solve`` computes the solution to a complex
    system of linear equations AX = B, where A is an n*n Hermitian
    positive definite tridiagonal matrix and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cg

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cgf.html

    Parameters
    ----------
    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    d : float, shape (`n`), modified in place
        `On entry`: must contain the n diagonal elements of the
        tridiagonal matrix A.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, `d` is overwritten by the n diagonal elements of the diagonal
        matrix D from the LDL^H factorization of A.

    e : complex, shape (`n`-1), modified in place
        `On entry`: must contain the (n-1) subdiagonal elements of the
        tridiagonal matrix A.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, `e` is overwritten by the (n-1) subdiagonal elements of the
        unit lower bidiagonal matrix L from the LDL^H factorization of
        A. (`e` can also be regarded as the conjugate of the
        superdiagonal of the unit upper bidiagonal factor U from the
        U^HDU factorization of A.)

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the reciprocal of the condition number of the matrix
        A, computed as `rcond` = 1/(||A||_1,||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The principal minor of order *<value>* of the matrix A is
            not positive definite. The factorization has not been
            completed and the solution could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    A is factorized as A = LDL^H, where L is a unit lower bidiagonal
    matrix and D is a real diagonal matrix, and the factored form of A
    is then used to solve the system of equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_posdef_tridiag_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04cgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, d[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, e[n-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _d_cls = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    )
    _d = _d_cls.to_ctypes()
    _e_cls = utils._EngineComplex128ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_n_py-1,),
    )
    _e = _e_cls.to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nrhs,
        _d,
        _e,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The principal minor of order {:s} of the matrix A is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'positive definite. The factorization has not been completed and the solution',
                    'could not be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 5:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _d_cls.refresh_py(_d)
        _e_cls.refresh_py(_e)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_herm_solve(uplo, n, nrhs, a, ipiv, b):
    u"""
    Computes the solution and error-bound to a complex Hermitian system
    of linear equations.

    ``complex_herm_solve`` computes the solution to a complex system of
    linear equations AX = B, where A is an n*n Hermitian matrix and X
    and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04ch

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04chf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : complex, shape (`n`, `n`), modified in place
        `On entry`: the n*n Hermitian matrix A.

        If `uplo` = 'U', the leading `n` by `n` upper triangular part of
        the array `a` contains the upper triangular part of the matrix
        A, and the strictly lower triangular part of `a` is not
        referenced.

        If `uplo` = 'L', the leading `n` by `n` lower triangular part of
        the array `a` contains the lower triangular part of the matrix
        A, and the strictly upper triangular part of `a` is not
        referenced.

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by
        ``lapacklin.zhetrf``.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.zhetrf``.

            If `ipiv`[k-1] > 0, then rows and columns k and `ipiv`[k-1]
            were interchanged, and d_{kk} is a 1*1 diagonal block;

            if `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0, then rows
            and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block;

            if `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0, then rows
            and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^H, if
    `uplo` = 'U', or A = LDL^H, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is Hermitian and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_herm_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04chft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _a,
        _lda,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_herm_packed_solve(uplo, n, nrhs, ap, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex Hermitian system of linear equations, packed storage.

    ``complex_herm_packed_solve`` computes the solution to a complex
    system of linear equations AX = B, where A is an n*n complex
    Hermitian matrix, stored in packed format and X and B are n*r
    matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04cj

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04cjf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        `On entry`: the n*n Hermitian matrix A, packed column-wise in a
        linear array. The jth column of the matrix A is stored in the
        array `ap` as follows:

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by
        ``lapacklin.zhptrf``, stored as a packed triangular matrix in
        the same storage format as A.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.zhptrf``.

            If `ipiv`[k-1] > 0, then rows and columns k and `ipiv`[k-1]
            were interchanged, and d_{kk} is a 1*1 diagonal block;

            if `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0, then rows
            and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block;

            if `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0, then rows
            and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^H, if
    `uplo` = 'U', or A = LDL^H, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is Hermitian and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_herm_packed_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04cjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo != \'U\' or \'L\': uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ap_cls.refresh_py(_ap)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_symm_solve(uplo, n, nrhs, a, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex symmetric system of linear equations.

    ``complex_symm_solve`` computes the solution to a complex system of
    linear equations AX = B, where A is an n*n complex symmetric matrix
    and X and B are n*r matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04dh

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04dhf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    a : complex, shape (`n`, `n`), modified in place
        `On entry`: the n*n complex symmetric matrix A.

        If `uplo` = 'U', the leading `n` by `n` upper triangular part of
        the array `a` contains the upper triangular part of the matrix
        A, and the strictly lower triangular part of `a` is not
        referenced.

        If `uplo` = 'L', the leading `n` by `n` lower triangular part of
        the array `a` contains the lower triangular part of the matrix
        A, and the strictly upper triangular part of `a` is not
        referenced.

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by
        ``lapacklin.zsytrf``.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no exception is raised, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.zsytrf``.

            If `ipiv`[k-1] > 0, then rows and columns k and `ipiv`[k-1]
            were interchanged, and d_{kk} is a 1*1 diagonal block;

            if `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0, then rows
            and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block;

            if `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0, then rows
            and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` not one of 'U' or 'u' or 'L' or 'l': `uplo`
            = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^T, if
    `uplo` = 'U', or A = LDL^T, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_symm_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04dhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _a,
        _lda,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo not one of \'U\' or \'u\' or \'L\' or \'l\':',
                    'uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_symm_packed_solve(uplo, n, nrhs, ap, ipiv, b):
    u"""
    Computes the solution, estimated condition number and error-bound to
    a complex symmetric system of linear equations, packed storage.

    ``complex_symm_packed_solve`` computes the solution to a complex
    system of linear equations AX = B, where A is an n*n complex
    symmetric matrix, stored in packed format and X and B are n*r
    matrices.
    An estimate of the condition number of A and an error bound for the
    computed solution are also returned.

    For full information please refer to the NAG Library document for
    f04dj

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04djf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    n : int
        The number of linear equations n, i.e., the order of the matrix
        A.

    nrhs : int
        The number of right-hand sides r, i.e., the number of columns of
        the matrix B.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        `On entry`: the n*n symmetric matrix A, packed column-wise in a
        linear array. The jth column of the matrix A is stored in the
        array `ap` as follows:

        `On exit`: if no exception is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by
        ``lapacklin.zsptrf``, stored as a packed triangular matrix in
        the same storage format as A.

    ipiv : int, shape (`n`), modified in place
        `On exit`: if no constraints are violated, details of the
        interchanges and the block structure of D, as determined by
        ``lapacklin.zsptrf``.

            If `ipiv`[k-1] > 0, then rows and columns k and `ipiv`[k-1]
            were interchanged, and d_{kk} is a 1*1 diagonal block;

            if `uplo` = 'U' and `ipiv`[k-1] = `ipiv`[k-2] < 0, then rows
            and columns k-1 and -`ipiv`[k-1] were interchanged and
            d_{k-1:k,k-1:k} is a 2*2 diagonal block;

            if `uplo` = 'L' and `ipiv`[k-1] = `ipiv`[k] < 0, then rows
            and columns k+1 and -`ipiv`[k-1] were interchanged and
            d_{k:k+1,k:k+1} is a 2*2 diagonal block.

    b : complex, shape (`n`, `nrhs`), modified in place
        `On entry`: the n*r matrix of right-hand sides B.

        `On exit`: if the function exits successfully or `errno` = `n` +
        1, the n*r solution matrix X.

    Returns
    -------
    rcond : float
        If no constraints are violated, an estimate of the reciprocal of
        the condition number of the matrix A, computed as `rcond` =
        1/(||A||_1||A^-1||_1).

    errbnd : float
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for a computed solution x^,
        such that ||x^-x||_1/||x||_1 <= `errbnd`, where x^ is a column
        of the computed solution returned in the array `b` and x is the
        corresponding column of the exact solution X. If `rcond` is less
        than machine precision, `errbnd` is returned as unity.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` -2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -1)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` (i > 0) and (i<=`n`))
            Diagonal block *<value>* of the block diagonal matrix is
            zero. The factorization has been completed, but the solution
            could not be computed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            A solution has been computed, but `rcond` is less than
            machine precision so that the matrix A is numerically
            singular.

    Notes
    -----
    The diagonal pivoting method is used to factor A as A = UDU^T, if
    `uplo` = 'U', or A = LDL^T, if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_symm_packed_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04djft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._EngineIntArrayType.c_type, # Output, ipiv[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._BYREF_FLOAT64, # Output, errbnd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ipiv = _ipiv_cls.to_ctypes(can_reference_contents=False)
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbnd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
        _ldb,
        _rcond,
        _errbnd,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo != \'U\' or \'L\': uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ierr_py <= _n_py and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Diagonal block {:s} of the block diagonal matrix is zero.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'The factorization has been completed, but the solution could not',
                    'be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _n_py+1 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A solution has been computed, but rcond is less than machine precision',
                    'so that the matrix A is numerically singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ap_cls.refresh_py(_ap)
        _ipiv_cls.refresh_py(_ipiv)
        _b_cls.refresh_py(_b)
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        _errbnd_py = utils._EngineFloat64ScalarType.to_py(_errbnd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rcond', 'errbnd'],
        )(
            _rcond_py,
            _errbnd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_toeplitz_yule(n, t, x, wantp, p, wantv, v):
    u"""
    Solution of the Yule--Walker equations for real symmetric positive
    definite Toeplitz matrix, one right-hand side.

    ``real_toeplitz_yule`` solves the Yule--Walker equations for a real
    symmetric positive definite Toeplitz system.

    For full information please refer to the NAG Library document for
    f04fe

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04fef.html

    Parameters
    ----------
    n : int
        The order of the Toeplitz matrix T.

    t : float, shape (`n`+1)
        `t`[0] must contain the value tau_0 of the diagonal elements of
        T, and the remaining `n` elements of `t` must contain the
        elements of the vector t.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    wantp : bool
        Must be set to True if the partial (auto)correlation
        coefficients are required, and must be set to False otherwise.

    p : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `wantp`= True: `n`; otherwise: 1.

        `On exit`: with `wantp` as True, the ith element of `p` contains
        the partial (auto)correlation coefficient, or reflection
        coefficient, p_i for the ith step. (See [Further Comments] and
        submodule ``tsa``.) If `wantp` is False, `p` is not referenced.
        Note that in any case, x_n = p_n.

    wantv : bool
        Must be set to True if the mean square prediction errors are
        required, and must be set to False otherwise.

    v : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `wantv`= True: `n`; otherwise: 1.

        `On exit`: with `wantv` as True, the ith element of `v` contains
        the mean square prediction error, or predictor error variance
        ratio, v_i, for the ith step. (See [Further Comments] and
        submodule ``tsa``.) If `wantv` is False, `v` is not referenced.

    Returns
    -------
    vlast : float
        The value of v_n, the mean square prediction error for the final
        step.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, `t`[0] = *<value>*.

            Constraint: `t`[0] > 0.0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Principal minor *<value>* is not positive definite. Value of
            the reflection coefficient is *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_toeplitz_yule`` solves the equations

        Tx = -t,

    where T is the n*n symmetric positive definite Toeplitz matrix

        T = [table omitted]

    and t is the vector

        t^T = (tau_1,tau_2...tau_n).

    The function uses the method of Durbin (see Durbin (1960) and Golub
    and Van Loan (1996)).
    Optionally the mean square prediction errors and/or the partial
    correlation coefficients for each step can be returned.

    References
    ----------
    Bunch, J R, 1985, `Stability of methods for solving Toeplitz systems
    of equations`, SIAM J. Sci. Statist. Comput. (6), 349--364

    Bunch, J R, 1987, `The weak and strong stability of algorithms in
    numerical linear algebra`, Linear Algebra Appl. (88/89), 49--66

    Cybenko, G, 1980, `The numerical stability of the Levinson--Durbin
    algorithm for Toeplitz systems of equations`, SIAM J. Sci. Statist.
    Comput. (1), 303--319

    Durbin, J, 1960, `The fitting of time series models`, Rev. Inst.
    Internat. Stat. (28), 233

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_toeplitz_yule'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04feft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n+1].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_LOGICAL, # Input, wantp.
        utils._EngineFloat64ArrayType.c_type, # Output, p[*].
        utils._BYREF_LOGICAL, # Input, wantv.
        utils._EngineFloat64ArrayType.c_type, # Output, v[*].
        utils._BYREF_FLOAT64, # Output, vlast.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _wantp = utils._EngineLogicalScalarType(
        wantp, {'fun_name': fun_name, 'entity': 'wantp'},
    ).to_ctypes()
    _wantp_py = utils._EngineLogicalScalarType.to_py(_wantp)
    if _wantp_py:
        _p_dim1 = _n_py
    else:
        _p_dim1 = 1
    exp_shape_p = (_p_dim1,)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=exp_shape_p,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _wantv = utils._EngineLogicalScalarType(
        wantv, {'fun_name': fun_name, 'entity': 'wantv'},
    ).to_ctypes()
    _wantv_py = utils._EngineLogicalScalarType.to_py(_wantv)
    if _wantv_py:
        _v_dim1 = _n_py
    else:
        _v_dim1 = 1
    exp_shape_v = (_v_dim1,)
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=exp_shape_v,
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _vlast = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _t,
        _x,
        _wantp,
        _p,
        _wantv,
        _v,
        _vlast,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: t[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Principal minor {:s} is not positive definite.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Value of the reflection coefficient is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _p_cls.refresh_py(_p)
        _v_cls.refresh_py(_v)
        _vlast_py = utils._EngineFloat64ScalarType.to_py(_vlast)
        return_data = _vlast_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['vlast'],
            )(
                _vlast_py,
            )
    _err_class.manage()
    return return_data

def real_toeplitz_solve(n, t, b, x, wantp, p):
    u"""
    Solution of real symmetric positive definite Toeplitz system, one
    right-hand side.

    ``real_toeplitz_solve`` solves the equations Tx = b, where T is a
    real symmetric positive definite Toeplitz matrix.

    For full information please refer to the NAG Library document for
    f04ff

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04fff.html

    Parameters
    ----------
    n : int
        The order of the Toeplitz matrix T.

    t : float, shape (`n`)
        `t`[i] must contain the value tau_i, for i = 0,1,...,`n`-1.

    b : float, shape (`n`)
        The right-hand side vector b.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    wantp : bool
        Must be set to True if the reflection coefficients are required,
        and must be set to False otherwise.

    p : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `wantp`= True: `n`-1; otherwise: 1.

        `On exit`: with `wantp` as True, the ith element of `p` contains
        the reflection coefficient, p_i, for the ith step, for i =
        1,2,...,`n`-1. (See [Further Comments].) If `wantp` is False,
        `p` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, `t`[0] = *<value>*.

            Constraint: `t`[0] > 0.0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Principal minor *<value>* is not positive definite. Value of
            the reflection coefficient is *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_toeplitz_solve`` solves the equations

        Tx = b,

    where T is the n*n symmetric positive definite Toeplitz matrix

        T = [table omitted]

    and b is an n-element vector.

    The function uses the method of Levinson (see Levinson (1947) and
    Golub and Van Loan (1996)).
    Optionally, the reflection coefficients for each step may also be
    returned.

    References
    ----------
    Bunch, J R, 1985, `Stability of methods for solving Toeplitz systems
    of equations`, SIAM J. Sci. Statist. Comput. (6), 349--364

    Bunch, J R, 1987, `The weak and strong stability of algorithms in
    numerical linear algebra`, Linear Algebra Appl. (88/89), 49--66

    Cybenko, G, 1980, `The numerical stability of the Levinson--Durbin
    algorithm for Toeplitz systems of equations`, SIAM J. Sci. Statist.
    Comput. (1), 303--319

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Levinson, N, 1947, `The Weiner RMS error criterion in filter design
    and prediction`, J. Math. Phys. (25), 261--278
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_toeplitz_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04ffft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_LOGICAL, # Input, wantp.
        utils._EngineFloat64ArrayType.c_type, # Output, p[*].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _wantp = utils._EngineLogicalScalarType(
        wantp, {'fun_name': fun_name, 'entity': 'wantp'},
    ).to_ctypes()
    _wantp_py = utils._EngineLogicalScalarType.to_py(_wantp)
    if _wantp_py:
        _p_dim1 = _n_py-1
    else:
        _p_dim1 = 1
    exp_shape_p = (_p_dim1,)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=exp_shape_p,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _t,
        _b,
        _x,
        _wantp,
        _p,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: t[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py > 0 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Principal minor {:s} is not positive definite.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Value of the reflection coefficient is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _p_cls.refresh_py(_p)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_solve(m, n, a, b, tol, work, lwork):
    u"""
    Least squares (if rank = n) or minimal least squares (if rank < n)
    solution of m real equations in n unknowns, m >= n.

    ``real_gen_solve`` finds the solution of a linear least squares
    problem, Ax = b, where A is a real m*n(m >= n) matrix and b is an m
    element vector.
    If the matrix of observations is not of full rank, then the minimal
    least squares solution is returned.

    For full information please refer to the NAG Library document for
    f04jg

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04jgf.html

    Parameters
    ----------
    m : int
        m, the number of rows of `a`.

    n : int
        n, the number of columns of `a`.

    a : float, shape (`m`, `n`), modified in place
        `On entry`: the m*n matrix A.

        `On exit`: if `svd` is returned as False, `a` is overwritten by
        details of the QR factorization of A.

        If `svd` is returned as True, the first n rows of `a` are
        overwritten by the right-hand singular vectors, stored by rows;
        and the remaining rows of the array are used as workspace.

    b : float, shape (`m`), modified in place
        `On entry`: the right-hand side vector b.

        `On exit`: the first n elements of `b` contain the minimal least
        squares solution vector x. The remaining m-n elements are used
        for workspace.

    tol : float
        A relative tolerance to be used to determine the rank of A.
        `tol` should be chosen as approximately the largest relative
        error in the elements of A. For example, if the elements of A
        are correct to about 4 significant figures then `tol` should be
        set to about 5*10^-4. See [Further Comments] for a description
        of how `tol` is used to determine rank. If `tol` is outside the
        range (epsilon,1.0), where epsilon is the machine precision, the
        value epsilon is used in place of `tol`. For most problems this
        is unreasonably small.

    work : float, shape (`lwork`), modified in place
        `On exit`: if `svd` is returned as False, then the first n
        elements of `work` contain information on the QR factorization
        of A (see argument `a` above), and `work`[n] contains the
        condition number ||R||_E||R^-1||_E of the upper triangular
        matrix R.

        If `svd` is returned as True, then the first n elements of
        `work` contain the singular values of A arranged in descending
        order and `work`[n] contains the total number of iterations
        taken by the QR algorithm.

        The rest of `work` is used as workspace.

    lwork : int
        The dimension of the array `work`.

    Returns
    -------
    svd : bool
        Is returned as False if the least squares solution has been
        obtained from the QR factorization of A. In this case A is of
        full rank. `svd` is returned as True if the least squares
        solution has been obtained from the singular value decomposition
        of A.

    sigma : float
        The standard error, i.e., the value sqrt(r^Tr/(m-k)) when m > k,
        and the value zero when m = k. Here r is the residual vector
        b-Ax and k is the rank of A.

    irank : int
        k, the rank of the matrix A. It should be noted that it is
        possible for `irank` to be returned as n and `svd` to be
        returned as True. This means that the matrix R only just failed
        the test for nonsingularity.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lwork` is too small. Minimum size required:
            *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= `n`.

        (`errno` 2)
            The QR algorithm has failed to converge to the singular
            values in 50*`n` iterations. This failure can only happen
            when the singular value decomposition is employed, but even
            then it is not likely to occur.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The minimal least squares solution of the problem Ax = b is the
    vector x of minimum (Euclidean) length which minimizes the length of
    the residual vector r = b-Ax.

    The real m*n(m >= n) matrix A is factorized as

        A = Q(R;0)

    where Q is an m*m orthogonal matrix and R is an n*n upper triangular
    matrix.
    If R is of full rank, then the least squares solution is given by

        x = (R^-1 0)Q^Tb.

    If R is not of full rank, then the singular value decomposition of R
    is obtained so that R is factorized as

        R = UDV^T,

    where U and V are n*n orthogonal matrices and D is the n*n diagonal
    matrix

        D = diag(sigma_1,sigma_2,...,sigma_n),

    with sigma_1 >= sigma_2 >= ... >= sigma_n >= 0, these being the
    singular values of A.
    If the singular values sigma_{k+1},...,sigma_n are negligible, but
    sigma_k is not negligible, relative to the data errors in A, then
    the rank of A is taken to be k and the minimal least squares
    solution is given by

        x = V[table omitted][table omitted]Q^Tb,

    where S = diag(sigma_1,sigma_2,...,sigma_k).

    The function also returns the value of the standard error

        [table omitted]

    r^Tr being the residual sum of squares and k the rank of A.

    References
    ----------
    Lawson, C L and Hanson, R J, 1974, `Solving Least Squares Problems`,
    Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_gen_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04jgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[m].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_LOGICAL, # Output, svd.
        utils._BYREF_FLOAT64, # Output, sigma.
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py,),
    )
    _b = _b_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _svd = utils._EngineLogicalScalarType.empty_ctypes()
    _sigma = utils._EngineFloat64ScalarType.empty_ctypes()
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType(
        lwork, {'fun_name': fun_name, 'entity': 'lwork'},
    ).to_ctypes()
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work_cls = utils._EngineFloat64ArrayType(
        work, {'fun_name': fun_name, 'entity': 'work'},
        exp_shape=(_lwork_py,),
        check_contents=False,
    )
    _work = _work_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _a,
        _lda,
        _b,
        _tol,
        _svd,
        _sigma,
        _irank,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwork is too small. Minimum size required: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The QR algorithm has failed to converge to the singular values in 50*n',
                    'iterations. This failure can only happen when the singular value',
                    'decomposition is employed, but even then it is not likely to occur.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _work_cls.refresh_py(_work)
        _svd_py = utils._EngineLogicalScalarType.to_py(_svd)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['svd', 'sigma', 'irank'],
        )(
            _svd_py,
            _sigma_py,
            _irank_py,
        )
    _err_class.manage()
    return return_data

def real_tridiag_fac_solve(job, n, a, b, c, d, ipiv, y, tol):
    u"""
    Solution of real tridiagonal simultaneous linear equations
    (coefficient matrix already factorized by
    ``matop.real_gen_tridiag_lu``).

    ``real_tridiag_fac_solve`` solves a system of tridiagonal equations
    following the factorization by ``matop.real_gen_tridiag_lu``.
    This function is intended for applications such as inverse iteration
    as well as straightforward linear equation applications.

    For full information please refer to the NAG Library document for
    f04le

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04lef.html

    Parameters
    ----------
    job : int
        Must specify the equations to be solved.

        `job` = 1

            The equations (T-lambda I)x = y are to be solved, but
            diagonal elements of U are not to be perturbed.

        `job` = -1

            The equations (T-lambda I)x = y are to be solved and, if
            overflow would otherwise occur, diagonal elements of U are
            to be perturbed. See argument `tol`.

        `job` = 2

            The equations (T-lambda I)^Tx = y are to be solved, but
            diagonal elements of U are not to be perturbed.

        `job` = -2

            The equations (T-lambda I)^Tx = y are to be solved and, if
            overflow would otherwise occur, diagonal elements of U are
            to be perturbed. See argument `tol`.

        `job` = 3

            The equations Ux = y are to be solved, but diagonal elements
            of U are not to be perturbed.

        `job` = -3

            The equations Ux = y are to be solved and, if overflow would
            otherwise occur, diagonal elements of U are to be perturbed.
            See argument `tol`.

    n : int
        n, the order of the matrix T.

    a : float, shape (`n`)
        The diagonal elements of U as returned by
        ``real_tridiag_fac_solve``.

    b : float, shape (`n`)
        The elements of the first superdiagonal of U as returned by
        ``real_tridiag_fac_solve``.

    c : float, shape (`n`)
        The subdiagonal elements of L as returned by
        ``real_tridiag_fac_solve``.

    d : float, shape (`n`)
        The elements of the second superdiagonal of U as returned by
        ``real_tridiag_fac_solve``.

    ipiv : int, shape (`n`)
        Details of the matrix P as returned by
        ``matop.real_gen_tridiag_lu``.

    y : float, shape (`n`), modified in place
        `On entry`: the right-hand side vector y.

        `On exit`: `y` is overwritten by the solution vector x.

    tol : float
        The minimum perturbation to be made to very small diagonal
        elements of U. `tol` is only referenced when `job` is negative.
        `tol` should normally be chosen as about epsilon||U||, where
        epsilon is the machine precision, but if `tol` is supplied as
        non-positive, it is reset to epsilon max(|u_{ij}|).

    Returns
    -------
    tol : float
        If on entry `tol` is non-positive, it is reset as just
        described. Otherwise `tol` is unchanged.

    Raises
    ------
    NagValueError
        (`errno` i > 1)
            Overflow would occur computing the element I in the
            solution. I = *<value>*.

        (`errno` 1)
            On entry, `job` = *<value>*.

            Constraint: `job` != 0 and `job` >= -3 and `job` <= 3.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    Following the factorization of the n*n tridiagonal matrix (T-lambda
    I) as

        T-lambda I = PLU

    by ``matop.real_gen_tridiag_lu``, ``real_tridiag_fac_solve`` may be
    used to solve any of the equations

        (T-lambda I)x = y, (T-lambda I)^Tx = y, Ux = y

    for x, the choice of equation being controlled by the argument
    `job`.
    In each case there is an option to perturb zero or very small
    diagonal elements of U, this option being intended for use in
    applications such as inverse iteration.

    References
    ----------
    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford

    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_tridiag_fac_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04left')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, job.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Input, c[n].
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._EngineIntArrayType.c_type, # Input, ipiv[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._BYREF_FLOAT64, # Input/Output, tol.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _job = utils._EngineIntScalarType(
        job, {'fun_name': fun_name, 'entity': 'job'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ipiv = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _job,
        _n,
        _a,
        _b,
        _c,
        _d,
        _ipiv,
        _y,
        _tol,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py > 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow would occur computing the element I in the solution.',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, job = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: job != 0 and job >= -3 and job <= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        return_data = _tol_py
    _err_class.manage()
    return return_data

def real_blkdiag_fac_solve(trans, n, nbloks, blkstr, a, lena, pivot, b, ir):
    u"""
    Solution of real almost block diagonal simultaneous linear equations
    (coefficient matrix already factorized by
    ``matop.real_gen_blkdiag_lu``).

    ``real_blkdiag_fac_solve`` calculates the approximate solution of a
    set of real linear equations with multiple right-hand sides, AX = B
    or A^TX = B, where A is an almost block-diagonal matrix which has
    been factorized by ``matop.real_gen_blkdiag_lu``.

    For full information please refer to the NAG Library document for
    f04lh

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04lhf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved.

        `trans` = 'N'

            Solve AX = B.

        `trans` = 'T'

            Solve A^TX = B.

    n : int
        n, the order of the matrix A.

    nbloks : int
        The total number of blocks of the matrix A, as supplied to
        ``real_blkdiag_fac_solve``.

    blkstr : int, shape (3, `nbloks`)
        Information which describes the block structure of A, as
        supplied to ``real_blkdiag_fac_solve``.

    a : float, shape (`lena`)
        The elements in the factorization of A, as returned by
        ``real_blkdiag_fac_solve``.

    lena : int
        The dimension of the array `a`.

    pivot : int, shape (`n`)
        Details of the interchanges in the factorization, as returned by
        ``real_blkdiag_fac_solve``.

    b : float, shape (`n`, `ir`), modified in place
        `On entry`: the n*r right-hand side matrix B.

        `On exit`: `b` is overwritten by the n*r solution matrix X.

    ir : int
        r, the number of right-hand sides.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lena` is too small. `lena` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, the following equality does not hold:
            `blkstr`[1,0]+sum(`blkstr`[1,k-1]-`blkstr`[2,k-2]:k =
            2,`nbloks`) = `n`.

        (`errno` 1)
            On entry, the following equality does not hold:
            sum(`blkstr`[0,k-1]:k = 1,`nbloks`) = `n`.

        (`errno` 1)
            On entry, K = *<value>*, `blkstr`[2,K-1] = *<value>*,
            `blkstr`[2,K-2] = *<value>* and `blkstr`[1,K-1] = *<value>*.

            Constraint: `blkstr`[2,K-1]+`blkstr`[2,K-2] >=
            `blkstr`[1,K-1].

        (`errno` 1)
            On entry, the following inequality was not satisfied for: J
            = *<value>*. sum(`blkstr`[1,k-1]-`blkstr`[2,k-1]:k = 1,J) <=
            sum(`blkstr`[0,k-1]:k = 1,J) <=
            `blkstr`[1,0]+sum(`blkstr`[1,k-1]-`blkstr`[2,k-2]:k = 2,J).

        (`errno` 1)
            On entry, K = *<value>*, `blkstr`[1,K-1] = *<value>*,
            `blkstr`[2,K-1] = *<value>* and `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1]-`blkstr`[2,K-1] <=
            `blkstr`[0,K-1].

        (`errno` 1)
            On entry, K = *<value>*, `blkstr`[1,K-1] = *<value>* and
            `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1] >= `blkstr`[0,K-1].

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[2,K-1] = *<value>*.

            Constraint: `blkstr`[2,K-1] >= 0.

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[1,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1] >= 1.

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[0,K-1] >= 1.

        (`errno` 1)
            On entry, `n` = *<value>* and `nbloks` = *<value>*.

            Constraint: `n` >= `nbloks`.

        (`errno` 1)
            On entry, `nbloks` = *<value>*.

            Constraint: `nbloks` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 1.

        (`errno` 1)
            On entry, `trans` != 'N' or 'T': `trans` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_blkdiag_fac_solve`` solves a set of real linear equations AX
    = B or A^TX = B, where A is almost block-diagonal. A must first be
    factorized by ``matop.real_gen_blkdiag_lu``.
    ``real_blkdiag_fac_solve`` then computes X by forward and backward
    substitution over the blocks.

    References
    ----------
    Diaz, J C, Fairweather, G and Keast, P, 1983, `Fortran packages for
    solving certain almost block diagonal linear systems by modified
    alternate row and column elimination`, ACM Trans. Math. Software
    (9), 358--375
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_blkdiag_fac_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04lhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nbloks.
        utils._EngineIntArrayType.c_type, # Input, blkstr[3, nbloks].
        utils._EngineFloat64ArrayType.c_type, # Input, a[lena].
        utils._BYREF_INT, # Input, lena.
        utils._EngineIntArrayType.c_type, # Input, pivot[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, ir].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, ir.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((blkstr, 'blkstr'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nbloks = utils._EngineIntScalarType(
        nbloks, {'fun_name': fun_name, 'entity': 'nbloks'},
    ).to_ctypes()
    _nbloks_py = utils._EngineIntScalarType.to_py(_nbloks)
    _blkstr = utils._EngineIntArrayType(
        blkstr, {'fun_name': fun_name, 'entity': 'blkstr'},
        exp_shape=(3, _nbloks_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _lena = utils._EngineIntScalarType(
        lena, {'fun_name': fun_name, 'entity': 'lena'},
    ).to_ctypes()
    _lena_py = utils._EngineIntScalarType.to_py(_lena)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_lena_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _pivot = utils._EngineIntArrayType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _ir = utils._EngineIntScalarType(
        ir, {'fun_name': fun_name, 'entity': 'ir'},
    ).to_ctypes()
    _ir_py = utils._EngineIntScalarType.to_py(_ir)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _ir_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nbloks,
        _blkstr,
        _a,
        _lena,
        _pivot,
        _b,
        _ldb,
        _ir,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lena is too small.',
                    'lena = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Minimum possible dimension: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the following equality does not hold:',
                    'blkstr[1,0]+sum(blkstr[1,k-1]-blkstr[2,k-2]:k = 2,nbloks) = n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the following equality does not hold:',
                    'sum(blkstr[0,k-1]:k = 1,nbloks) = n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s}, blkstr[2,K-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'blkstr[2,K-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'blkstr[1,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: blkstr[2,K-1]+blkstr[2,K-2] >= blkstr[1,K-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the following inequality was not satisfied for:',
                    'J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'sum(blkstr[1,k-1]-blkstr[2,k-1]:k = 1,J) <=',
                    'sum(blkstr[0,k-1]:k = 1,J) <=',
                    'blkstr[1,0]+sum(blkstr[1,k-1]-blkstr[2,k-2]:k = 2,J).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s}, blkstr[1,K-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'blkstr[2,K-1] = {:s} and blkstr[0,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: blkstr[1,K-1]-blkstr[2,K-1] <= blkstr[0,K-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s}, blkstr[1,K-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and blkstr[0,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: blkstr[1,K-1] >= blkstr[0,K-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'blkstr[2,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: blkstr[2,K-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'blkstr[1,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: blkstr[1,K-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, K = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'blkstr[0,K-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: blkstr[0,K-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and nbloks = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= nbloks.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nbloks = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nbloks >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ir = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ir >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans != \'N\' or \'T\': trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def real_posdef_vband_solve(n, al, lal, d, nrow, ir, b, iselct, x):
    u"""
    Solution of real symmetric positive definite variable-bandwidth
    simultaneous linear equations (coefficient matrix already factorized
    by ``matop.real_vband_posdef_fac``).

    ``real_posdef_vband_solve`` computes the approximate solution of a
    system of real linear equations with multiple right-hand sides, AX =
    B, where A is a symmetric positive definite variable-bandwidth
    matrix, which has previously been factorized by
    ``matop.real_vband_posdef_fac``.
    Related systems may also be solved.

    For full information please refer to the NAG Library document for
    f04mc

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04mcf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix L.

    al : float, shape (`lal`)
        The elements within the envelope of the lower triangular matrix
        L, taken in row by row order, as returned by
        ``matop.real_vband_posdef_fac``. The unit diagonal elements of L
        must be stored explicitly.

    lal : int
        The dimension of the array `al`.

    d : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `iselct`>=4: 1; otherwise: `n`.

        The diagonal elements of the diagonal matrix D. `d` is not
        referenced if `iselct` >= 4.

    nrow : int, shape (`n`)
        `nrow`[i-1] must contain the width of row i of L, i.e., the
        number of elements between the first (leftmost) nonzero element
        and the element on the diagonal, inclusive.

    ir : int
        r, the number of right-hand sides.

    b : float, shape (`n`, `ir`)
        The n*r right-hand side matrix B. See also [Further Comments].

    iselct : int
        Must specify the type of system to be solved, as follows:

        `iselct` = 1

            Solve LDL^TX = B.

        `iselct` = 2

            Solve LDX = B.

        `iselct` = 3

            Solve DL^TX = B.

        `iselct` = 4

            Solve LL^TX = B.

        `iselct` = 5

            Solve LX = B.

        `iselct` = 6

            Solve L^TX = B.

    x : float, shape (`n`, `ir`), modified in place
        `On exit`: the n*r solution matrix X. See also [Further
        Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lal` = *<value>* and `nrow`[0]+...+`nrow`[`n`-1]
            = *<value>*.

            Constraint: `lal` >= `nrow`[0]+...+`nrow`[`n`-1].

        (`errno` 1)
            On entry, I = *<value>* and `nrow`[I-1] = *<value>*.

            Constraint: `nrow`[I-1] >= 1 and `nrow`[I-1] <= I.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 1.

        (`errno` 3)
            On entry, `iselct` = *<value>*.

            Constraint: `iselct` >= 1 and `iselct` <= 6.

        (`errno` 4)
            On entry, I = *<value>*.

            Constraint: `d`[I-1] != 0.0.

        (`errno` 5)
            At least one diagonal entry of `al` is not unit.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The normal use of this function is the solution of the systems AX =
    B, following a call of ``matop.real_vband_posdef_fac`` to determine
    the Cholesky factorization A = LDL^T of the symmetric positive
    definite variable-bandwidth matrix A.

    However, the function may be used to solve any one of the following
    systems of linear algebraic equations:

    (1) LDL^TX = B (usual system),

    (#) LDX = B (lower triangular system),

    (#) DL^TX = B (upper triangular system),

    (#) LL^TX = B

    (#) LX = B (unit lower triangular system),

    (#) L^TX = B (unit upper triangular system).

    L denotes a unit lower triangular variable-bandwidth matrix of order
    n, D a diagonal matrix of order n, and B a set of right-hand sides.

    The matrix L is represented by the elements lying within its
    **envelope**, i.e., between the first nonzero of each row and the
    diagonal.
    The width `nrow`[i-1] of the ith row is the number of elements
    between the first nonzero element and the element on the diagonal
    inclusive.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_posdef_vband_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04mcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, al[lal].
        utils._BYREF_INT, # Input, lal.
        utils._EngineFloat64ArrayType.c_type, # Input, d[*].
        utils._EngineIntArrayType.c_type, # Input, nrow[n].
        utils._BYREF_INT, # Input, ir.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, ir].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, iselct.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, ir].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lal = utils._EngineIntScalarType(
        lal, {'fun_name': fun_name, 'entity': 'lal'},
    ).to_ctypes()
    _lal_py = utils._EngineIntScalarType.to_py(_lal)
    _al = utils._EngineFloat64ArrayType(
        al, {'fun_name': fun_name, 'entity': 'al'},
        exp_shape=(_lal_py,),
    ).to_ctypes()
    _iselct = utils._EngineIntScalarType(
        iselct, {'fun_name': fun_name, 'entity': 'iselct'},
    ).to_ctypes()
    _iselct_py = utils._EngineIntScalarType.to_py(_iselct)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _iselct_py >= 4:
        _d_dim1 = 1
    else:
        _d_dim1 = _n_py
    exp_shape_d = (_d_dim1,)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=exp_shape_d,
    ).to_ctypes()
    _nrow = utils._EngineIntArrayType(
        nrow, {'fun_name': fun_name, 'entity': 'nrow'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ir = utils._EngineIntScalarType(
        ir, {'fun_name': fun_name, 'entity': 'ir'},
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _ir_py = utils._EngineIntScalarType.to_py(_ir)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _ir_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _ir_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _al,
        _lal,
        _d,
        _nrow,
        _ir,
        _b,
        _ldb,
        _iselct,
        _x,
        _ldx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lal = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nrow[0]+...+nrow[n-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lal >= nrow[0]+...+nrow[n-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s} and nrow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nrow[I-1] >= 1 and nrow[I-1] <= I.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ir = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ir >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iselct = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iselct >= 1 and iselct <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d[I-1] != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one diagonal entry of al is not unit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_toeplitz_yule_update(n, t, x, v):
    u"""
    Update solution of the Yule--Walker equations for real symmetric
    positive definite Toeplitz matrix.

    ``real_toeplitz_yule_update`` updates the solution to the
    Yule--Walker equations for a real symmetric positive definite
    Toeplitz system.

    For full information please refer to the NAG Library document for
    f04me

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04mef.html

    Parameters
    ----------
    n : int
        The order of the Toeplitz matrix T.

    t : float, shape (`n`+1)
        `t`[0] must contain the value tau_0 of the diagonal elements of
        T, and the remaining `n` elements of `t` must contain the
        elements of the vector t_n.

    x : float, shape (`n`), modified in place
        `On entry`: with `n` > 1 the (n-1) elements of the solution
        vector x_{n-1} as returned by a previous call to
        ``real_toeplitz_yule_update``. The element `x`[`n`-1] need not
        be specified.

        `On exit`: the solution vector x_n. The element `x`[`n`-1]
        returns the partial (auto)correlation coefficient, or reflection
        coefficient, for the nth step. If |`x`[`n`-1]| >= 1.0, the
        matrix T_{n+1} will not be positive definite to working
        accuracy.

    v : float
        With `n` > 1 the mean square prediction error for the (n-1)th
        step, as returned by a previous call to
        ``real_toeplitz_yule_update``.

    Returns
    -------
    v : float
        The mean square prediction error, or predictor error variance
        ratio, nu_n, for the nth step. (See [Further Comments] and [the
        G13 Introduction].)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, `n` = *<value>* and `x`[`n`-2] = *<value>*.

            Constraint: if `n` > 1, abs(`x`[`n`-2]) < 1.0.

        (`errno` -1)
            On entry, `t`[0] = *<value>*.

            Constraint: `t`[0] > 0.0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Matrix of order *<value>* would not be positive definite.
            Value of the reflection coefficient is *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_toeplitz_yule_update`` solves the equations

        T_n x_n = -t_n,

    where T_n is the n*n symmetric positive definite Toeplitz matrix

        T_n = [table omitted]

    and t_n is the vector

        t_n^T = (tau_1tau_2...tau_n),

    given the solution of the equations

        T_{n-1}x_{n-1} = -t_{n-1}.

    The function will normally be used to successively solve the
    equations

        T_k x_k = -t_k, k = 1,2,...,n.

    If it is desired to solve the equations for a single value of n,
    then function ``real_toeplitz_yule`` may be called.
    This function uses the method of Durbin (see Durbin (1960) and Golub
    and Van Loan (1996)).

    References
    ----------
    Bunch, J R, 1985, `Stability of methods for solving Toeplitz systems
    of equations`, SIAM J. Sci. Statist. Comput. (6), 349--364

    Bunch, J R, 1987, `The weak and strong stability of algorithms in
    numerical linear algebra`, Linear Algebra Appl. (88/89), 49--66

    Cybenko, G, 1980, `The numerical stability of the Levinson--Durbin
    algorithm for Toeplitz systems of equations`, SIAM J. Sci. Statist.
    Comput. (1), 303--319

    Durbin, J, 1960, `The fitting of time series models`, Rev. Inst.
    Internat. Stat. (28), 233

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_toeplitz_yule_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04meft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Input/Output, v.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _v = utils._EngineFloat64ScalarType(
        v, {'fun_name': fun_name, 'entity': 'v'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _t,
        _x,
        _v,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and x[n-2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if n > 1, abs(x[n-2]) < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: t[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Matrix of order {:s} would not be positive definite.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Value of the reflection coefficient is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _v_py = utils._EngineFloat64ScalarType.to_py(_v)
        return_data = _v_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['v'],
            )(
                _v_py,
            )
    _err_class.manage()
    return return_data

def real_toeplitz_update(n, t, b, x, work):
    u"""
    Update solution of real symmetric positive definite Toeplitz system.

    ``real_toeplitz_update`` updates the solution of the equations Tx =
    b, where T is a real symmetric positive definite Toeplitz matrix.

    For full information please refer to the NAG Library document for
    f04mf

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04mff.html

    Parameters
    ----------
    n : int
        The order of the Toeplitz matrix T.

    t : float, shape (`n`)
        `t`[i] must contain the value tau_i, for i = 0,1,...,`n`-1.

    b : float, shape (`n`)
        The right-hand side vector b_n.

    x : float, shape (`n`), modified in place
        `On entry`: with `n` > 1 the (n-1) elements of the solution
        vector x_{n-1} as returned by a previous call to
        ``real_toeplitz_update``. The element `x`[`n`-1] need not be
        specified.

        `On exit`: the solution vector x_n.

    work : float, shape (2*`n`-1), modified in place
        `On entry`: with `n` > 2 the elements of `work` should be as
        returned from a previous call to ``real_toeplitz_update`` with
        (`n`-1) as the argument `n`.

        `On exit`: the first (`n`-1) elements of `work` contain the
        solution to the Yule--Walker equations

            T_{n-1}y_{n-1} = -t_{n-1},

        where t_{n-1} = (tau_1tau_2...tau_{n-1})^`t`.

    Returns
    -------
    p : float
        The reflection coefficient p_{n-1}. (See [Further Comments].)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, `t`[0] = *<value>*.

            Constraint: `t`[0] > 0.0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The Toeplitz Matrix is not positive definite Value of the
            reflection coefficient is *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_toeplitz_update`` solves the equations

        T_n x_n = b_n,

    where T_n is the n*n symmetric positive definite Toeplitz matrix

        T_n = [table omitted]

    and b_n is the n-element vector b_n = (beta_1beta_2...beta_n)^T,
    given the solution of the equations

        T_{n-1}x_{n-1} = b_{n-1}.

    This function will normally be used to successively solve the
    equations

        T_k x_k = b_k, k = 1,2,...,n.

    If it is desired to solve the equations for a single value of n,
    then function ``real_toeplitz_solve`` may be called.
    This function uses the method of Levinson (see Levinson (1947) and
    Golub and Van Loan (1996)).

    References
    ----------
    Bunch, J R, 1985, `Stability of methods for solving Toeplitz systems
    of equations`, SIAM J. Sci. Statist. Comput. (6), 349--364

    Bunch, J R, 1987, `The weak and strong stability of algorithms in
    numerical linear algebra`, Linear Algebra Appl. (88/89), 49--66

    Cybenko, G, 1980, `The numerical stability of the Levinson--Durbin
    algorithm for Toeplitz systems of equations`, SIAM J. Sci. Statist.
    Comput. (1), 303--319

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Levinson, N, 1947, `The Weiner RMS error criterion in filter design
    and prediction`, J. Math. Phys. (25), 261--278
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_toeplitz_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04mfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, p.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[2*n-1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    _work_cls = utils._EngineFloat64ArrayType(
        work, {'fun_name': fun_name, 'entity': 'work'},
        exp_shape=(2*_n_py-1,),
    )
    _work = _work_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _t,
        _b,
        _x,
        _p,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: t[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The Toeplitz Matrix is not positive definite',
                    'Value of the reflection coefficient is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _work_cls.refresh_py(_work)
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        return_data = _p_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['p'],
            )(
                _p_py,
            )
    _err_class.manage()
    return return_data

def real_gen_sparse_lsqsol(m, n, b, x, se, aprod, damp, atol, btol, conlim, itnlim, msglvl, data=None, io_manager=None):
    u"""
    Sparse linear least squares problem, m real equations in n unknowns.

    ``real_gen_sparse_lsqsol`` solves sparse nonsymmetric equations,
    sparse linear least squares problems and sparse damped linear least
    squares problems, using a Lanczos algorithm.

    For full information please refer to the NAG Library document for
    f04qa

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04qaf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    b : float, shape (`m`), modified in place
        `On entry`: the right-hand side vector b.

        `On exit`: `b` is overwritten.

    x : float, shape (`n`), modified in place
        `On exit`: the solution vector x.

    se : float, shape (`n`), modified in place
        `On exit`: the estimates of the standard errors of the
        components of x. Thus `se`[i-1] contains an estimate of
        sqrt(nu_{ii}), where nu_{ii} is the ith diagonal element of the
        estimated variance-covariance matrix V. The estimates returned
        in `se` will be the lower bounds on the actual estimated
        standard errors, but will usually be correct to at least one
        significant figure.

    aprod : callable mode = aprod(mode, m, n, x, y, data=None)
        `aprod` must perform the operations y := y+Ax and x := x+A^Ty
        for given vectors x and y.

        Parameters
        ~~~~~~~~~~
        mode : int
            Specifies which operation is to be performed.

            `mode` = 1

                `aprod` must compute y+Ax.

            `mode` = 2

                `aprod` must compute x+A^Ty.

        m : int
            m, the number of rows of A.

        n : int
            n, the number of columns of A.

        x : float, ndarray, shape (`n`), to be modified in place
            `On entry`: the vector x.

            `On exit`: if `mode` = 1, `x` must be unchanged.

            If `mode` = 2, `x` must contain x+A^Ty.

        y : float, ndarray, shape (`m`), to be modified in place
            `On entry`: the vector y.

            `On exit`: if `mode` = 1, `y` must contain y+Ax.

            If `mode` = 2, `y` must be unchanged.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        mode : int
            May be used as a flag to indicate a failure in the
            computation of y+Ax or x+A^Ty. If `mode` is negative on exit
            from `aprod`, ``real_gen_sparse_lsqsol`` will exit
            immediately with errno set to `mode`.

    damp : float
        The value lambda. If either problem (1) or problem (2) is to be
        solved, `damp` must be supplied as zero.

    atol : float
        The tolerance, tol_1, of the convergence criteria (6) and (7);
        it should be an estimate of the largest relative error in the
        elements of A. For example, if the elements of A are correct to
        about 4 significant figures, then `atol` should be set to about
        5*10^-4. If `atol` is supplied as less than epsilon, where
        epsilon is the machine precision, the value epsilon is used
        instead of `atol`.

    btol : float
        The tolerance, tol_2, of the convergence criterion (6); it
        should be an estimate of the largest relative error in the
        elements of B. For example, if the elements of B are correct to
        about 4 significant figures, then `btol` should be set to about
        5*10^-4. If `btol` is supplied as less than epsilon, the value
        epsilon is used instead of `btol`.

    conlim : float
        The value c_lim of equation (8); it should be an upper limit on
        the condition number of A_. `conlim` should not normally be
        chosen much larger than 1.0/`atol`. If `conlim` is supplied as
        zero, the value 1.0/epsilon is used instead of `conlim`.

    itnlim : int
        An upper limit on the number of iterations. If `itnlim` <= 0,
        the value `n` is used in place of `itnlim`, but for
        ill-conditioned problems a higher value of `itnlim` is likely to
        be necessary.

    msglvl : int
        The level of printing from ``real_gen_sparse_lsqsol``. If
        `msglvl` <= 0, then no printing occurs, but otherwise messages
        will be output on the file object associated with the advisory
        I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). A
        description of the printed output is given in [Further
        Comments]. The level of printing is determined as follows:

        `msglvl` <= 0

            No printing.

        `msglvl` = 1

            A brief summary is printed just prior to return from
            ``real_gen_sparse_lsqsol``.

        `msglvl` >= 2

            A summary line is printed periodically to monitor the
            progress of ``real_gen_sparse_lsqsol``, together with a
            brief summary just prior to return from
            ``real_gen_sparse_lsqsol``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    itnlim : int
        Unchanged unless `itnlim` <= 0 on entry, in which case it is set
        to `n`.

    itn : int
        The number of iterations performed.

    anorm : float
        An estimate of ||A_|| for the matrix A_ of (4).

    acond : float
        An estimate of cond(A_) which is a lower bound.

    rnorm : float
        An estimate of ||r_|| for the residual, r_, of (5) corresponding
        to the solution x returned in `x`. Note that ||r_|| is the
        function being minimized.

    arnorm : float
        An estimate of the ||A_^Tr_|| corresponding to the solution x
        returned in `x`.

    xnorm : float
        An estimate of ||x|| for the solution x returned in `x`.

    inform : int
        The reason for termination of ``real_gen_sparse_lsqsol``.

        `inform` = 0

            The exact solution is x = 0. No iterations are performed in
            this case.

        `inform` = 1

            The termination criterion of (6) has been satisfied with
            tol_1 and tol_2 as the values supplied in `atol` and `btol`
            respectively.

        `inform` = 2

            The termination criterion of (7) has been satisfied with
            tol_1 as the value supplied in `atol`.

        `inform` = 3

            The termination criterion of (6) has been satisfied with
            tol_1 and/or tol_2 as the value epsilon, where epsilon is
            the machine precision. One or both of the values supplied in
            `atol` and `btol` must have been less than epsilon and was
            too small for this machine.

        `inform` = 4

            The termination criterion of (7) has been satisfied with
            tol_1 as the value epsilon. The value supplied in `atol`
            must have been less than epsilon and was too small for this
            machine.

        The values `inform` = 5, 6 or 7 correspond to failure with
        `errno` = 2, 3 and 4 respectively (see [Raises/Warns]) and when
        errno is negative `inform` will be set to the same negative
        value.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            Termination criteria not satisfied, but the condition number
            bound supplied in `conlim` has been met.

        (`errno` 3)
            Termination criteria not satisfied, but the condition number
            is bounded by 1/``machine.precision``.

        (`errno` 4)
            Maximum number of iterations reached.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            `mode` in `aprod` has been set to: *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``real_gen_sparse_lsqsol`` can be used to solve a system of linear
    equations

        Ax = b

    where A is an n*n sparse nonsymmetric matrix, or can be used to
    solve linear least squares problems, so that
    ``real_gen_sparse_lsqsol`` minimizes the value rho given by

        rho = ||r||, r = b-Ax

    where A is an m*n sparse matrix and ||r|| denotes the Euclidean
    length of r so that ||r||^2 = r^Tr.
    A damping argument, lambda, may be included in the least squares
    problem in which case ``real_gen_sparse_lsqsol`` minimizes the value
    rho given by

        rho^2 = ||r||^2+lambda^2||x||^2.

    lambda is supplied as the argument `damp` and should of course be
    zero if the solution to problems (1) and (2) is required.
    Minimizing rho in (3) is often called ridge regression.

    ``real_gen_sparse_lsqsol`` is based upon algorithm LSQR (see Paige
    and Saunders (1982a) and Paige and Saunders (1982b)) and solves the
    problems by an algorithm based upon the Lanczos process.
    The function does not require A explicitly, but A is specified via
    `aprod` which must perform the operations (y+Ax) and (x+A^Ty) for a
    given n-element vector x and m element vector y. An argument to
    `aprod` specifies which of the two operations is required on a given
    entry.

    The function also returns estimates of the standard errors of the
    sample regression coefficients ( x_i, for i = 1,2,...,n) given by
    the diagonal elements of the estimated variance-covariance matrix V.
    When problem (2) is being solved and A is of full rank, then V is
    given by

        V = s^2(A^TA)^-1, s^2 = rho^2/(m-n), m > n

    and when problem (3) is being solved then V is given by

        V = s^2(A^TA+lambda^2I)^-1, s^2 = rho^2/m, lambda != 0.

    Let A_ denote the matrix

        A_= A, lambda = 0; A_= (A;lambda I), lambda != 0,

    let r_ denote the residual vector

        r_= r, lambda = 0; r_= (b;0)-A_x, lambda != 0

    corresponding to an iterate x, so that rho = ||r_|| is the function
    being minimized, and let ||A|| denote the Frobenius (Euclidean) norm
    of A.
    Then the function accepts x as a solution if it is estimated that
    one of the following two conditions is satisfied:

        rho <= tol_1||A_||.||x||+tol_2||b||

        ||A_^Tr_|| <= tol_1||A_||rho

    where tol_1 and tol_2 are user-supplied tolerances which estimate
    the relative errors in A and b respectively.
    Condition (6) is appropriate for compatible problems where, in
    theory, we expect the residual to be zero and will be satisfied by
    an acceptable solution x to a compatible problem.
    Condition (7) is appropriate for incompatible systems where we do
    not expect the residual to be zero and is based on the observation
    that, in theory,

        A_^Tr_= 0

    when x is a solution to the least squares problem, and so (7) will
    be satisfied by an acceptable solution x to a linear least squares
    problem.

    The function also includes a test to prevent convergence to
    solutions, x, with unacceptably large elements.
    This can happen if A is nearly singular or is nearly rank deficient.
    If we let the singular values of A_ be

        sigma_1 >= sigma_2 >= ... >= sigma_n >= 0

    then the condition number of A_ is defined as

        cond(A_) = sigma_1/sigma_k

    where sigma_k is the smallest nonzero singular value of A_ and hence
    k is the rank of A_.
    When k < n, then A_ is rank deficient, the least squares solution is
    not unique and ``real_gen_sparse_lsqsol`` will normally converge to
    the minimal length solution.
    In practice A_ will not have exactly zero singular values, but may
    instead have small singular values that we wish to regard as zero.

    The function provides for this possibility by terminating if

        cond(A_) >= c_lim

    where c_lim is a user-supplied limit on the condition number of A_.
    For problem (1) termination with this condition indicates that A is
    nearly singular and for problem (2) indicates that A is nearly rank
    deficient and so has near linear dependencies in its columns.
    In this case inspection of ||r||, ||A^Tr|| and ||x||, which are all
    returned by the function, will indicate whether or not an acceptable
    solution has been found.
    Condition (8), perhaps in conjunction with lambda != 0, can be used
    to try and 'regularize' least squares solutions.
    A full discussion of the stopping criteria is given in Section 6 of
    Paige and Saunders (1982a).

    Introduction of a nonzero damping argument lambda tends to reduce
    the size of the computed solution and to make its components less
    sensitive to changes in the data, and ``real_gen_sparse_lsqsol`` is
    applicable when a value of lambda is known `a priori`.
    To have an effect, lambda should normally be at least
    sqrt(epsilon)||A|| where epsilon is the machine precision.
    For further discussion see Paige and Saunders (1982b) and the
    references given there.

    Whenever possible the matrix A should be scaled so that the relative
    errors in the elements of A are all of comparable size.
    Such a scaling helps to prevent the least squares problem from being
    unnecessarily sensitive to data errors and will normally reduce the
    number of iterations required.
    At the very least, in the absence of better information, the columns
    of A should be scaled to have roughly equal column length.

    References
    ----------
    Paige, C C and Saunders, M A, 1982, `LSQR: An algorithm for sparse
    linear equations and sparse least squares`, ACM Trans. Math.
    Software (8), 43--71

    Paige, C C and Saunders, M A, 1982, `Algorithm 583 LSQR: Sparse
    linear equations and least squares problems`, ACM Trans. Math.
    Software (8), 195--209
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_gen_sparse_lsqsol'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_aprod_name = fun_name + ':aprod'

    aprodh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, mode.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[m].
        utils._BYREF_INT, # Input, lruser.
        utils._BYREF_INT, # Input, liuser.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_aprodh(_aprod, en_data, mode, m, n, x, y, lruser, liuser, _iuser, _ruser):
        """Helper for aprod."""
        try:
            _mode_aprodh = utils._EngineIntScalarType.to_py(mode.contents)
            _m_aprodh = utils._EngineIntScalarType.to_py(m.contents)
            _n_aprodh = utils._EngineIntScalarType.to_py(n.contents)
            _x_aprodh = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_aprodh,),
            )
            _y_aprodh = utils._EngineFloat64ArrayType.from_pointer(
                y,
                (_m_aprodh,),
            )
            if data is not None:
                _mode_aprodh = aprod(
                    _mode_aprodh,
                    _m_aprodh,
                    _n_aprodh,
                    _x_aprodh,
                    _y_aprodh,
                    data,
                )
            else:
                _mode_aprodh = aprod(
                    _mode_aprodh,
                    _m_aprodh,
                    _n_aprodh,
                    _x_aprodh,
                    _y_aprodh,
                )
            utils._EngineIntScalarType(
                _mode_aprodh, {'fun_name': cb_aprod_name, 'returned': 'mode'},
            ).refresh_ctypes_buffer(mode)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _x_aprodh, {'fun_name': cb_aprod_name, 'returned': 'x'}, x,
            )
            utils._unlink_pointer(x)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _y_aprodh, {'fun_name': cb_aprod_name, 'returned': 'y'}, y,
            )
            utils._unlink_pointer(y)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('f04qaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[m].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, se[n].
        utils._UNUSED_FUNCTYPE,
        aprodh,
        utils._BYREF_FLOAT64, # Input, damp.
        utils._BYREF_FLOAT64, # Input, atol.
        utils._BYREF_FLOAT64, # Input, btol.
        utils._BYREF_FLOAT64, # Input, conlim.
        utils._BYREF_INT, # Input/Output, itnlim.
        utils._BYREF_INT, # Input, msglvl.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, acond.
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_FLOAT64, # Output, arnorm.
        utils._BYREF_FLOAT64, # Output, xnorm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lruser.
        utils._BYREF_INT, # Input, liuser.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Output, inform.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py,),
    )
    _b = _b_cls.to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    utils._check_callable(aprod, {'fun_name': fun_name, 'entity': 'aprod'})
    _aprod = utils._cast_to_unused_callback(aprod)
    _aprodh = utils._cast_callback(
        aprodh, cb_aprodh,
        {'fun_name': fun_name, 'entity': 'aprod'},
    )
    _damp = utils._EngineFloat64ScalarType(
        damp, {'fun_name': fun_name, 'entity': 'damp'},
    ).to_ctypes()
    _atol = utils._EngineFloat64ScalarType(
        atol, {'fun_name': fun_name, 'entity': 'atol'},
    ).to_ctypes()
    _btol = utils._EngineFloat64ScalarType(
        btol, {'fun_name': fun_name, 'entity': 'btol'},
    ).to_ctypes()
    _conlim = utils._EngineFloat64ScalarType(
        conlim, {'fun_name': fun_name, 'entity': 'conlim'},
    ).to_ctypes()
    _itnlim = utils._EngineIntScalarType(
        itnlim, {'fun_name': fun_name, 'entity': 'itnlim'},
    ).to_ctypes()
    _msglvl = utils._EngineIntScalarType(
        msglvl, {'fun_name': fun_name, 'entity': 'msglvl'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _acond = utils._EngineFloat64ScalarType.empty_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _arnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _xnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lruser = utils._EngineIntScalarType.c_type(0)
    _liuser = utils._EngineIntScalarType.c_type(0)
    _iuser = None
    _ruser = None
    _inform = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _m,
        _n,
        _b,
        _x,
        _se,
        _aprod,
        _aprodh,
        _damp,
        _atol,
        _btol,
        _conlim,
        _itnlim,
        _msglvl,
        _nadv,
        _itn,
        _anorm,
        _acond,
        _rnorm,
        _arnorm,
        _xnorm,
        _work,
        _lruser,
        _liuser,
        _iuser,
        _ruser,
        _inform,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py < 0 and _ifmt == 99976:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'mode in aprod has been set to: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Termination criteria not satisfied, but the condition number',
                    'bound supplied in conlim has been met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Termination criteria not satisfied, but the condition number',
                    'is bounded by 1/machine.precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Maximum number of iterations reached.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _x_cls.refresh_py(_x)
        _se_cls.refresh_py(_se)
        _itnlim_py = utils._EngineIntScalarType.to_py(_itnlim)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _acond_py = utils._EngineFloat64ScalarType.to_py(_acond)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _arnorm_py = utils._EngineFloat64ScalarType.to_py(_arnorm)
        _xnorm_py = utils._EngineFloat64ScalarType.to_py(_xnorm)
        _inform_py = utils._EngineIntScalarType.to_py(_inform)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itnlim', 'itn', 'anorm', 'acond', 'rnorm', 'arnorm', 'xnorm', 'inform'],
        )(
            _itnlim_py,
            _itn_py,
            _anorm_py,
            _acond_py,
            _rnorm_py,
            _arnorm_py,
            _xnorm_py,
            _inform_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_lsq_covmat(job, p, sigma, a, svd, irank, sv, cj):
    u"""
    Covariance matrix for linear least squares problems, m real
    equations in n unknowns.

    ``real_gen_lsq_covmat`` returns elements of the estimated
    variance-covariance matrix of the sample regression coefficients for
    the solution of a linear least squares problem.

    The function can be used to find the estimated variances of the
    sample regression coefficients.

    For full information please refer to the NAG Library document for
    f04ya

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04yaf.html

    Parameters
    ----------
    job : int
        Specifies which elements of C are required.

        `job` = -1

            The upper triangular part of C is required.

        `job` = 0

            The diagonal elements of C are required.

        `job` > 0

            The elements of column `job` of C are required.

    p : int
        p, the order of the variance-covariance matrix C.

    sigma : float
        The standard error of the residual vector.

    a : float, shape (:, `p`), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`svd`) or `job`=-1: `p`; if
        `svd`= True and `job`>=0: `irank`; otherwise: 0.

        `On entry`: if `svd` = False, `a` must contain the upper
        triangular matrix U of the QU factorization of X, or of the
        Cholesky factorization of X^TX; elements of the array below the
        diagonal need not be set.

        If `svd` = True, A must contain the first k rows of the matrix
        V^T, where k is the rank of X and V is the right-hand orthogonal
        matrix of the singular value decomposition of X.

        Thus the ith row must contain the ith right-hand singular vector
        of X.

        `On exit`: if `job` >= 0, A is unchanged.

        If `job` = -1, `a` contains the upper triangle of the symmetric
        matrix C.

        If `svd` = True, elements of the array below the diagonal are
        used as workspace.

        If `svd` = False, they are unchanged.

    svd : bool
        Must be True if the least squares solution was obtained from a
        singular value decomposition of X. `svd` must be False if the
        least squares solution was obtained from either a QU
        factorization of X or a Cholesky factorization of X^TX. In the
        latter case the rank of X is assumed to be p and so is
        applicable only to full rank problems with n >= p.

    irank : int
        If `svd` = True, `irank` must specify the rank k of the matrix
        X.

        If `svd` = False, `irank` is not referenced and the rank of X is
        assumed to be p.

    sv : float, shape (`p`)
        If `svd` = True, `sv` must contain the first `irank` singular
        values of X.

        If `svd` = False, `sv` is not referenced.

    cj : float, shape (`p`), modified in place
        `On exit`: if `job` = 0, `cj` returns the diagonal elements of
        C.

        If `job` = j > 0, `cj` returns the jth column of C.

        If `job` = -1, `cj` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lda = *<value>* and `p` = *<value>*.

            Constraint: if `svd` = False, lda >= `p`.

        (`errno` 1)
            On entry, lda = *<value>* and `p` = *<value>*.

            Constraint: if `svd` = True and `job` = -1, lda >= `p`.

        (`errno` 1)
            On entry, `job` = *<value>*, lda = *<value>* and `irank` =
            *<value>*.

            Constraint: if `svd` = True and `job` >= 0, lda >= max(1,
            `irank`).

        (`errno` 1)
            On entry, `irank` = *<value>* and `p` = *<value>*.

            Constraint: if `svd` = True, `irank` >= 0 and `irank` <=
            `p`.

        (`errno` 1)
            On entry, `job` = *<value>* and `p` = *<value>*.

            Constraint: `job` >= -1 and `job` <= `p`.

        (`errno` 1)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` >= 0.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 1.

        (`errno` 2)
            On entry, `irank` = *<value>*.

            Constraint: if `svd` = True, `irank` > 0.

        (`errno` 3)
            On entry, `svd` = False and overflow would occur in
            calculating an element of C. The upper triangular matrix U
            must be very nearly singular.

        (`errno` 4)
            On entry, `svd` = False and one of the first `irank`
            singular values is zero. Either the first `irank` singular
            values or `irank` must be incorrect: `irank` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The estimated variance-covariance matrix C of the sample regression
    coefficients is given by

        C = sigma^2(X^TX)^(-1), X^TX nonsingular,

    where X^TX is the normal matrix for the linear least squares
    regression problem

        min:||y-Xb||_2,

    sigma^2 is the estimated variance of the residual vector r = y-Xb,
    and X is an n*p observation matrix.

    When X^TX is singular, C is taken to be

        C = sigma^2(X^TX)^†,

    where (X^TX)^† is the pseudo-inverse of X^TX; this assumes that the
    minimal least squares solution of (1) has been found.

    The diagonal elements of C are the estimated variances of the sample
    regression coefficients, b.

    The function can be used to find either the diagonal elements of C,
    or the elements of the jth column of C, or the upper triangular part
    of C.

    This function must be preceded by a function that returns either the
    upper triangular matrix U of the QU factorization of X or of the
    Cholesky factorization of X^TX, or the singular values and right
    singular vectors of X.
    In particular this function can be preceded by one of the functions
    ``lapackeig.dgelss`` or ``real_gen_solve``, which return the
    arguments `irank`, `sigma`, `a` and `sv` in the required form.
    ``real_gen_solve`` returns the argument `svd`, but when this
    function is used following function ``lapackeig.dgelss`` the
    argument `svd` should be set to True.
    The argument `p` of this function corresponds to the argument n in
    functions ``lapackeig.dgelss`` and ``real_gen_solve``.

    References
    ----------
    Anderson, T W, 1958, `An Introduction to Multivariate Statistical
    Analysis`, Wiley

    Lawson, C L and Hanson, R J, 1974, `Solving Least Squares Problems`,
    Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_gen_lsq_covmat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04yaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, job.
        utils._BYREF_INT, # Input, p.
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, p].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_LOGICAL, # Input, svd.
        utils._BYREF_INT, # Input, irank.
        utils._EngineFloat64ArrayType.c_type, # Input, sv[p].
        utils._EngineFloat64ArrayType.c_type, # Output, cj[p].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _job = utils._EngineIntScalarType(
        job, {'fun_name': fun_name, 'entity': 'job'},
    ).to_ctypes()
    _p = utils._EngineIntScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _svd = utils._EngineLogicalScalarType(
        svd, {'fun_name': fun_name, 'entity': 'svd'},
    ).to_ctypes()
    _svd_py = utils._EngineLogicalScalarType.to_py(_svd)
    _job_py = utils._EngineIntScalarType.to_py(_job)
    _p_py = utils._EngineIntScalarType.to_py(_p)
    _irank = utils._EngineIntScalarType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
    ).to_ctypes()
    _irank_py = utils._EngineIntScalarType.to_py(_irank)
    if not _svd_py or _job_py == -1:
        _a_dim1 = _p_py
    elif _svd_py and _job_py >= 0:
        _a_dim1 = _irank_py
    else:
        _a_dim1 = 0
    exp_shape_a = (_a_dim1, _p_py)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _sv = utils._EngineFloat64ArrayType(
        sv, {'fun_name': fun_name, 'entity': 'sv'},
        exp_shape=(_p_py,),
    ).to_ctypes()
    _cj_cls = utils._EngineFloat64ArrayType(
        cj, {'fun_name': fun_name, 'entity': 'cj'},
        exp_shape=(_p_py,),
        check_contents=False,
    )
    _cj = _cj_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _job,
        _p,
        _sigma,
        _a,
        _lda,
        _svd,
        _irank,
        _sv,
        _cj,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lda = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if svd = False, lda >= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lda = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if svd = True and job = -1, lda >= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, job = {:s}, lda = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if svd = True and job >= 0, lda >= max(1, irank).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if svd = True, irank >= 0 and irank <= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, job = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: job >= -1 and job <= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if svd = True, irank > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, svd = False and overflow would occur in calculating',
                    'an element of C. The upper triangular matrix U must be very',
                    'nearly singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, svd = False and one of the first irank singular',
                    'values is zero. Either the first irank singular values or',
                    'irank must be incorrect:',
                    'irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _cj_cls.refresh_py(_cj)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_norm_rcomm(irevcm, m, n, x, y, estnrm, t, seed, comm):
    u"""
    Norm estimation (for use in condition estimation), real rectangular
    matrix.

    ``real_gen_norm_rcomm`` estimates the 1-norm of a real rectangular
    matrix without accessing the matrix explicitly.
    It uses reverse communication for evaluating matrix products.
    The function may be used for estimating condition numbers of square
    matrices.

    For full information please refer to the NAG Library document for
    f04yd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04ydf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

        `On intermediate entry`: `irevcm` must be unchanged.

    m : int
        The number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    x : float, shape (`n`, `t`), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 1, contains the current
        matrix X.

        `On intermediate entry`: if `irevcm` = 2, must contain A^TY.

        `On final exit`: the array is undefined.

    y : float, shape (`m`, `t`), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 2, contains the current
        matrix Y.

        `On intermediate entry`: if `irevcm` = 1, must contain AX.

        `On final exit`: the array is undefined.

    estnrm : float
        `On initial entry`: need not be set.

        `On intermediate entry`: must not be changed.

    t : int
        The number of columns t of the matrices X and Y. This is an
        argument that can be used to control the accuracy and
        reliability of the estimate and corresponds roughly to the
        number of columns of A that are visited during each iteration of
        the algorithm.

        If `t` >= 2 then a partly random starting matrix is used in the
        algorithm.

        Suggested value: `t` = 2.

    seed : int
        The seed used for random number generation.

        If `t` = 1, `seed` is not used.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: `irevcm` = 1 or 2, and `x` contains the
        n*t matrix X and `y` contains the m*t matrix Y. The calling
        program must

        (a) if `irevcm` = 1, evaluate AX and store the result in `y`

            or

            if `irevcm` = 2, evaluate A^TY and store the result in `x`,

        (#) call ``real_gen_norm_rcomm`` once again, with all the other
            arguments unchanged.

        `On final exit`: `irevcm` = 0.

    estnrm : float
        `On final exit`: an estimate (a lower bound) for ||A||_1.

    Raises
    ------
    NagValueError
        (`errno` -10)
            On entry, `t` = *<value>* and `seed` = *<value>*.

            Constraint: if `t` > 1, `seed` >= 1.

        (`errno` -9)
            On entry, `m` = *<value>* and `t` = *<value>*.

            Constraint: 1 <= `t` <= `m`.

        (`errno` -3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` -1)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1 or 2.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 1)
            Internal error; please contact NAG.

    Notes
    -----
    ``real_gen_norm_rcomm`` computes an estimate (a lower bound) for the
    1-norm

        ||A||_1 = max_{1 <= j <= n}sum_i = 1^m|a_{ij}|

    of an m*n real matrix A = (a_{ij}).
    The function regards the matrix A as being defined by a
    user-supplied 'Black Box' which, given an n*t matrix X (with t≪n) or
    an m*t matrix Y, can return AX or A^TY.
    A reverse communication interface is used; thus control is returned
    to the calling program whenever a matrix product is required.

    **Note:** this function is **not recommended** for use when the
    elements of A are known explicitly; it is then more efficient to
    compute the 1-norm directly from formula (1) above.

    The **main use** of the function is for estimating ||B^-1||_1 for a
    square matrix, B, and hence the **condition number** kappa_1(B) =
    ||B||_1||B^-1||_1, without forming B^-1 explicitly (A = B^-1 above).

    If, for example, an LU factorization of B is available, the matrix
    products B^-1X and B^-TY required by ``real_gen_norm_rcomm`` may be
    computed by back - and forward-substitutions, without computing
    B^-1.

    The function can also be used to estimate 1-norms of matrix products
    such as A^-1B and ABC, without forming the products explicitly.
    Further applications are described by Higham (1988).

    Since ||A||_infinity = ||A^T||_1, ``real_gen_norm_rcomm`` can be
    used to estimate the infinity-norm of A by working with A^T instead
    of A.

    The algorithm used is described in Higham and Tisseur (2000).

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Higham, N J and Tisseur, F, 2000, `A block algorithm for matrix
    1-norm estimation, with an application to 1-norm pseudospectra`,
    SIAM J. Matrix. Anal. Appl. (21), 1185--1201
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.real_gen_norm_rcomm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04ydft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, t].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[ldy, t].
        utils._BYREF_INT, # Input, ldy.
        utils._BYREF_FLOAT64, # Input/Output, estnrm.
        utils._BYREF_INT, # Input, t.
        utils._BYREF_INT, # Input, seed.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[m*t].
        utils._EngineIntArrayType.c_type, # Input/Output, iwork[2*n+5*t+20].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (y, 'y')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _t = utils._EngineIntScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t_py = utils._EngineIntScalarType.to_py(_t)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _t_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ldy_py = utils._EngineFloat64ArrayType.get_nd(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldy = utils._EngineIntScalarType(
        _ldy_py, {'fun_name': fun_name, 'entity': '_ldy'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py, _t_py),
        mutual_sorder=_m_sorder,
    )
    _y = _y_cls.to_ctypes()
    _estnrm = utils._EngineFloat64ScalarType(
        estnrm, {'fun_name': fun_name, 'entity': 'estnrm'},
    ).to_ctypes()
    _seed = utils._EngineIntScalarType(
        seed, {'fun_name': fun_name, 'entity': 'seed'},
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('work', 'iwork'),
    )
    if _comm_needs_init:
        _work = utils._EngineFloat64ArrayType.empty_ctypes(_m_py*_t_py)
        comm['work'] = _work
    else:
        _work = comm['work']
    if _comm_needs_init:
        _iwork = utils._EngineIntArrayType.empty_ctypes(2*_n_py+5*_t_py+20)
        comm['iwork'] = _iwork
    else:
        _iwork = comm['iwork']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _m,
        _n,
        _x,
        _ldx,
        _y,
        _ldy,
        _estnrm,
        _t,
        _seed,
        _work,
        _iwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 0 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t = {:s} and seed = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if t > 1, seed >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 1239:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and t = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= t <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 1236:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 1235:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1234:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1235:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1234:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Internal error; please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        _estnrm_py = utils._EngineFloat64ScalarType.to_py(_estnrm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['irevcm', 'estnrm'],
        )(
            _irevcm_py,
            _estnrm_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_norm_rcomm(irevcm, m, n, x, y, estnrm, t, seed, comm):
    u"""
    Norm estimation (for use in condition estimation), complex
    rectangular matrix.

    ``complex_gen_norm_rcomm`` estimates the 1-norm of a complex
    rectangular matrix without accessing the matrix explicitly.
    It uses reverse communication for evaluating matrix products.
    The function may be used for estimating condition numbers of square
    matrices.

    For full information please refer to the NAG Library document for
    f04zd

    https://www.nag.com/numeric/nl/nagdoc_28.6/flhtml/f04/f04zdf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

        `On intermediate entry`: `irevcm` must be unchanged.

    m : int
        The number of rows of the matrix A.

    n : int
        `On initial entry`: n, the number of columns of the matrix A.

    x : complex, shape (`n`, `t`), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 1, contains the current
        matrix X.

        `On intermediate entry`: if `irevcm` = 2, must contain A^HY.

        `On final exit`: the array is undefined.

    y : complex, shape (`m`, `t`), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 2, contains the current
        matrix Y.

        `On intermediate entry`: if `irevcm` = 1, must contain AX.

        `On final exit`: the array is undefined.

    estnrm : float
        `On initial entry`: need not be set.

        `On intermediate entry`: must not be changed.

    t : int
        The number of columns t of the matrices X and Y. This is an
        argument that can be used to control the accuracy and
        reliability of the estimate and corresponds roughly to the
        number of columns of A that are visited during each iteration of
        the algorithm.

        If `t` >= 2 then a partly random starting matrix is used in the
        algorithm.

        Suggested value: `t` = 2.

    seed : int
        The seed used for random number generation.

        If `t` = 1, `seed` is not used.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: `irevcm` = 1 or 2, and `x` contains the
        n*t matrix X and `y` contains the m*t matrix Y. The calling
        program must

        (a) if `irevcm` = 1, evaluate AX and store the result in `y`

            or

            if `irevcm` = 2, evaluate A^HY and store the result in `x`,
            where A^H is the complex conjugate transpose;

        (#) call ``complex_gen_norm_rcomm`` once again, with all the
            arguments unchanged.

        `On final exit`: `irevcm` = 0.

    estnrm : float
        `On final exit`: an estimate (a lower bound) for ||A||_1.

    Raises
    ------
    NagValueError
        (`errno` -10)
            On entry, `t` = *<value>* and `seed` = *<value>*.

            Constraint: if `t` > 1, `seed` >= 1.

        (`errno` -9)
            On entry, `m` = *<value>* and `t` = *<value>*.

            Constraint: 1 <= `t` <= `m`.

        (`errno` -3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` -2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` -1)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1 or 2.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 1)
            Internal error; please contact NAG.

    Notes
    -----
    ``complex_gen_norm_rcomm`` computes an estimate (a lower bound) for
    the 1-norm

        ||A||_1 = max_{1 <= j <= n}sum_i = 1^m|a_{ij}|

    of an m*n complex matrix A = (a_{ij}).
    The function regards the matrix A as being defined by a
    user-supplied 'Black Box' which, given an n*t matrix X (with t≪n) or
    an m*t matrix Y, can return AX or A^HY, where A^H is the complex
    conjugate transpose.
    A reverse communication interface is used; thus control is returned
    to the calling program whenever a matrix product is required.

    **Note:** this function is **not** **recommended** for use when the
    elements of A are known explicitly; it is then more efficient to
    compute the 1-norm directly from the formula (1) above.

    The **main use** of the function is for estimating ||B^-1||_1 for a
    square matrix B, and hence the **condition number** kappa_1(B) =
    ||B||_1||B^-1||_1, without forming B^-1 explicitly (A = B^-1 above).

    If, for example, an LU factorization of B is available, the matrix
    products B^-1X and B^-HY required by ``complex_gen_norm_rcomm`` may
    be computed by back - and forward-substitutions, without computing
    B^-1.

    The function can also be used to estimate 1-norms of matrix products
    such as A^-1B and ABC, without forming the products explicitly.
    Further applications are described in Higham (1988).

    Since ||A||_infinity = ||A^H||_1, ``complex_gen_norm_rcomm`` can be
    used to estimate the infinity-norm of A by working with A^H instead
    of A.

    The algorithm used is described in Higham and Tisseur (2000).

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396

    Higham, N J and Tisseur, F, 2000, `A block algorithm for matrix
    1-norm estimation, with an application to 1-norm pseudospectra`,
    SIAM J. Matrix. Anal. Appl. (21), 1185--1201
    """

    excs = []
    fun_name = 'naginterfaces.base.linsys.complex_gen_norm_rcomm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f04zdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[ldx, t].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[ldy, t].
        utils._BYREF_INT, # Input, ldy.
        utils._BYREF_FLOAT64, # Input/Output, estnrm.
        utils._BYREF_INT, # Input, t.
        utils._BYREF_INT, # Input, seed.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[m*t].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwork[2*n].
        utils._EngineIntArrayType.c_type, # Input/Output, iwork[2*n+5*t+20].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (y, 'y')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineComplex128ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _t = utils._EngineIntScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t_py = utils._EngineIntScalarType.to_py(_t)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _t_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ldy_py = utils._EngineComplex128ArrayType.get_nd(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldy = utils._EngineIntScalarType(
        _ldy_py, {'fun_name': fun_name, 'entity': '_ldy'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py, _t_py),
        mutual_sorder=_m_sorder,
    )
    _y = _y_cls.to_ctypes()
    _estnrm = utils._EngineFloat64ScalarType(
        estnrm, {'fun_name': fun_name, 'entity': 'estnrm'},
    ).to_ctypes()
    _seed = utils._EngineIntScalarType(
        seed, {'fun_name': fun_name, 'entity': 'seed'},
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('work', 'rwork', 'iwork'),
    )
    if _comm_needs_init:
        _work = utils._EngineComplex128ArrayType.empty_ctypes(_m_py*_t_py)
        comm['work'] = _work
    else:
        _work = comm['work']
    if _comm_needs_init:
        _rwork = utils._EngineFloat64ArrayType.empty_ctypes(2*_n_py)
        comm['rwork'] = _rwork
    else:
        _rwork = comm['rwork']
    if _comm_needs_init:
        _iwork = utils._EngineIntArrayType.empty_ctypes(2*_n_py+5*_t_py+20)
        comm['iwork'] = _iwork
    else:
        _iwork = comm['iwork']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _m,
        _n,
        _x,
        _ldx,
        _y,
        _ldy,
        _estnrm,
        _t,
        _seed,
        _work,
        _rwork,
        _iwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 0 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t = {:s} and seed = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if t > 1, seed >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 1239:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and t = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= t <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 1236:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 1235:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1234:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 1235:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1234:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Internal error; please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        _estnrm_py = utils._EngineFloat64ScalarType.to_py(_estnrm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['irevcm', 'estnrm'],
        )(
            _irevcm_py,
            _estnrm_py,
        )
    _err_class.manage()
    return return_data
